program csv2ssv(input, output);
(* csv2ssv: convert a comma delimited csv file to a space delimited file

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  https://alum.mit.edu/www/toms 

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.10; (* of csv2ssv.p 2018 Feb 22
2018 Feb 22, 1.10: fix bug: files that start with ',' are really '-,'!!
2016 Mar 23, 1.09: remove extra CR before eof for p2c compile.
2016 Mar 22, 1.08: remove control-Ms from DOS files
2015 Mar 25, 1.07: fix handling of comma at start of line
2014 Sep 26, 1.06: don't rewrite output!
2014 Sep 26, 1.05: backup
2014 May 07, 1.04: rename: csvtospace -> csv2ssv
2012 Jun 11, 1.03: backup
2012 Jun 11, 1.01: add description
2012 Jun 11, 1.01: handle ",,"
2012 Jun 11, 1.00: origin
*)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.csv2ssv *)
(*
name
   csv2ssv: convert a comma delimited csv file to a space delimited file

synopsis
   csv2ssv(input: in, output: out)

files

   input:  a csv file

   output: the csv file converted to space delimited file

description

   In the csv 'comma separated variable' file format, columns of a
   table are separated by commas.

   Delila programs usually use space separation.  This routine
   converts the csv to a space separated file.

   To convert to the Delila-style space delimited table, the following
   rules apply:

   Spaces are converted to underscores.
   Double quotes (") spaces are removed.
   Commas are replaced by spaces.
   Commas inside quotes are ignored!

   Two successive commas meaning 'nothing' are replaced by a ',_,'
   so that '_' represents 'nothing' in the output file.

examples

"First","Second title",1
"Fourth","Third line",2
"Fifth","Sixth",3
"Fifth",,"Sixth"
"ignore this, comma",,"Sixth"

result:

First Second_title 1
Fourth Third_line 2
Fifth Sixth 3
double-comma _ Sixth
ignore_this,_comma _ Sixth

documentation

see also

author

   Thomas Dana Schneider

bugs

technical notes

   The csv file format uses commas to separate columns.  The unmkcsv
   script can handle this except when there is a comma inside a quote.
   In that case the program has to be smart to know where quotes are.

*)
(* end module describe.csv2ssv *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.16 84 mar 12 tds/gds'; *)

(* begin module csv2ssv.themain *)
procedure themain(var infile, outfile: text);
(* the main procedure of the program *)
var
   cin: char; (* a character in the input file *)
   cout: char; (* a character in the input file *)
   previouscin: char; (* the previous value of cin *)
   inquote: boolean; (* true if inside a quote string *)
   beginline: boolean; (* true only at the start of a line *)
begin
   reset(infile);
   (* rewrite(outfile); this wipes prior data *)
   inquote := false;
   previouscin := ' ';

   (* Special case: file begins with a ','!
      This is where the upper left cell is empty.  Stupid, I know! *)
   if not eof(infile) then begin
      if not eoln(infile) then begin
         if (infile^ = ',') then begin
            write(outfile,'- ');
         end
      end
   end;

   while not eof (infile) do begin
if eof(infile) then begin
write(outfile,'EOF');
end;

      if eoln(infile) then begin
         readln(infile);
{
write(outfile,'OFFICIAL');
if eof(infile) then begin
write(outfile,'WOAH');
end;
}
         if not eof(infile) then begin
            (* When compiled by p2c, - but not by the old gpc -
               an extra CR is written here.
               Block it if it is actually EOF *)
            writeln(outfile);
         end;
         previouscin := ' ';
         inquote := false; (* restart quotes on every line *)
         beginline := true;
      end
      else begin
         read(infile, cin);
         if ord(cin) <> 13 then begin (* completely ignore control Ms *)
            if cin = '"' then inquote := not inquote;

            if (cin = ',') and beginline then begin
               (* handle line starting with a comma! *)
               write(outfile,'- ');
            end
            else begin
               (* Spaces are converted to underscores. *)
               cout := cin; (* output is tenatatively input ... *)
               if cin = ' '
               then cout := '_'
               else begin
                  if not inquote then begin
                     if cin = ',' then begin
                       (* handle double comma: *)
                       if previouscin = ',' then write(outfile,'-');
                       cout := ' ';
                       (* Commas outside quotes are replaced by spaces. *)
                     end;
                  end;
                  (* Commas inside quotes are ignored! *)
               end;

               (* Two Double quotes ("") are retained to show empty column! *)
               if (cin = '"') and (previouscin = '"') then write(outfile,'--');

               (* Double quotes (") are removed. *)
               if cin <> '"' then write(outfile,cout);
            end;

            previouscin := cin;
            beginline := false;
         end (* finish ignoring CRs *)
         else begin
            (* this would show the control-Ms:
                write(outfile,'M');
            *)
            (* this leaves an extra line on the end of the file!! *)
            (* beginline := true; no help *)
         end;

      end
   end;

end;
(* end module csv2ssv.themain *)

begin
   themain(input, output);
1: end. (* csvtospace *)
