program dashseq(dashedseq, dashseqp, sequ, output);
(* dashseq: convert aligned sequences with dashes to pure sequences

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  toms@alum.mit.edu (permanent)
  https://alum.mit.edu/www/toms (permanent)

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.00; (* of dashseq.p 2010 Nov 21
2010 Nov 21, 1.00: origin
*)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.dashseq *)
(*
name
   dashseq: convert aligned sequences with dashes to pure sequences

synopsis
   dashseq(dashedseq: in, dashseqp: in, sequ: out, output: out)

files

   dashedseq:  Aligned DNA sequences, one per line.
      The first one is the 'cononical sequence and it may not contain
      any dashes.
      The following lines are copied to the sequ output except when
      there is a dash, it is replaced by the corresponding letter
      in the canonical sequence.

   sequ:  pure sequence without dashes.

   dashseqp:  parameters to control the program.  The file must contain the
      following parameters, one per line:

      parameterversion: The version number of the program.  This allows the
         user to be warned if an old parameter file is used.

      not used yet.

   output: messages to the user

description

   Sometimes sequnces are given with dashes to indicate that the letter
   is the same as the letter in a cononical

examples

documentation

Examples are in figures 2 and 3 of:

@article{Argast.Monnat1998,
author = "G. M. Argast
 and K. M. Stephens
 and M. J. Emond
 and R. J. {Monnat Jr}",
title = "{I-\emph{PpoI} and I-\emph{CreI} homing site sequence 
degeneracy determined by random mutagenesis and sequential \emph{in
vitro} enrichment}",
journal = "J Mol Biol",
volume = "280",
pages = "345--353",
pmid = "9665841",
year = "1998"}

see also

author

   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.dashseq *)

(* begin module string.const *)
      maxstring = 2000; (* the maximum string *)
(* end module string.const version = 5.43; (@ of prgmod.p 2010 Oct 20 *)

type

(* begin module string.type *)
   stringptr = ^string; (* pointer to a string *)
   string = record (* a string of characters *)
      letters: array[1..maxstring] of char; (* the letters in the string *)
      length: integer; (* the number of characters in the string *)
      current: integer; (* the letter we are working on *)
      next: stringptr; (* the next string in a series *)
   end;
(* end module string.type version = 5.43; (@ of prgmod.p 2010 Oct 20 *)

var
   dashedseq, (* file used by this program *)
   dashseqp, (* file used by this program *)
   sequ: text; (* file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 5.43; (@ of prgmod.p 2010 Oct 20 *)

(* begin module interact.clearstring *)
(* begin module clearstring *)
(* These modules clear strings in various ways *)
(* ---- *)
procedure emptystring(var ribbon: string);
(* empty the contents of the string but do NOT
remove the pointer.  This is useful for clearing
one string within a linked list of them. *)
var   index: integer; (* to the ribbon *)
begin (* clearstring *)
   with ribbon do begin
      for index := 1 to maxstring do letters[index] := ' ';
      length := 0;
      current := 0;
   end
end; (* emptystring *)
(* ---- *)
procedure clearstring(var ribbon: string);
(* empty the string and remove the pointer *)
begin (* clearstring *)
   with ribbon do begin
      emptystring(ribbon);
      next := nil;
   end
end; (* clearstring *)
(* ---- *)
procedure initializestring(var ribbon: string);
(* start the string with a nil pointer.   This routine
should be called before doing linked list work.  This allows
the standard string routines to clear the string without
killing the pointer.  This is now deprecated, do not use it
since clearstring still clears the next pointer. *)
begin (* initializestring *)
   writeln(output,'remove initializestring routine!');
   writeln(output,'replace it with clearstring routine!');
   halt; (* to force deprecation *)
   clearstring(ribbon);
   ribbon.next := nil;
end; (* initializestring *)
(* end module clearstring version = 5.43; (@ of prgmod.p 2010 Oct 20 *)
(* end module interact.clearstring version = 5.43; (@ of prgmod.p 2010 Oct 20 *)

(* begin module interact.getstring *)
procedure getstring(var afile: text; var buffer: string;
                    var gotten: boolean);
(* get a line (as a string) from a file not using string calls.  this lets
one obtain lines from a file without interactive prompts *)
var   index: integer; (* of buffer *)
begin (* getstring *)
      clearstring(buffer);
      if eof(afile)
      then gotten := false
      else begin
         index := 0;
         while (not eoln(afile)) and (index < maxstring) do begin
            index := succ(index);
            read(afile, buffer.letters[index])
         end;
         if not eoln(afile) then begin
            writeln(output, ' getstring: a line exceeds maximum string size (',
                            maxstring:1,')');
            halt
         end;
         buffer.length := index;
         buffer.current := 1;
         readln(afile);
         gotten := true
      end
end; (* getstring *)
(* end module interact.getstring version = 5.43; (@ of prgmod.p 2010 Oct 20 *)

(* begin module interact.writestring *)
(* begin module writestring *)
procedure writestring(var tofile: text; var s: string);
(* write the string s to file tofile, no writeln *)
var   i: integer; (* index to s *)
begin (* writestring *)
      with s do for i := 1 to length do write(tofile, letters[i])
end; (* writestring *)
(* end module writestring version = 5.43; (@ of prgmod.p 2010 Oct 20 *)
(* end module interact.writestring version = 5.43; (@ of prgmod.p 2010 Oct 20 *)

(* begin module dashseq.themain *)
procedure themain(var dashedseq, dashseqp, sequ: text);
(* the main procedure of the program *)
var
   parameterversion: real; (* parameter version number *)
   canonical: string; (* the first string from dashedseq *)
   buffer: string; (* a string from dashedseq *)
   i: integer; (* index to buffer *)
   gotten: boolean; (* whether buffer was gotten *)
begin
   writeln(output,'dashseq ',version:4:2);
{
   reset(dashseqp);
   readln(dashseqp, parameterversion);
   if round(100*parameterversion) < round(100*updateversion) then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;
}

   reset(dashedseq);
   getstring(dashedseq, canonical, gotten);
   if not gotten then begin
      writeln(output,'canonical sequence not found');
      halt;
   end;

   rewrite(sequ);
   writestring(sequ, canonical);
   writeln(sequ,'.');
   while not eof (dashedseq) do begin
      getstring(dashedseq, buffer, gotten);
      for i := 1 to buffer.length do begin
         if buffer.letters[i] = '-' then begin
            buffer.letters[i] := canonical.letters[i];
         end;
      end;
      
      if gotten then begin
         writestring(sequ, buffer);
         writeln(sequ,'.');
      end;
   end;

end;
(* end module dashseq.themain *)

begin
   themain(dashedseq, dashseqp, sequ);
1: end.
