program diri(ari, bri, histog, dcout, output);
(* diri: finds the information for all sites in a flex model

  Ryan Shultzaberger

in the lab of:

  Dr. Thomas D. Schneider
  toms@alum.mit.edu
  permanent email: toms@alum.mit.edu (use only if first address fails)
  https://alum.mit.edu/www/toms/

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.11; (* of diri.p 2005 Oct 20
2005 Oct 20, 1.10 modified histog reading routine and changed
                  identifier of "not site"
2005 Oct 18, 1.09 diri gives its name in dcout
2005 Oct 13, 1.08 point to makeinst
2005 Oct 07, 1.07 made coords integers
2005 Sep 30, 1.06 pointer to flexrbs
2005 May 23, 1.05 upgrade for gpc, add patent message TDS
2001 Nov 12, 1.04 diclean renamed diri
2000 Jun  9, 1.03 last changes by Ryan
2000 Feb 19, 1.00 origin as 'diclean'
*)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.diri *)
(*
name
   diri: finds the information for all sites in a flex model

synopsis
   diri(ari: in, bri: in, histog, in: dcout: out, output: out)

files
   ari:  rixyin from ri file for model A

   bri:  rixyin from ri file for model B

   histog: output of the genhis file containing the binding distribution
           between model A and B, the user should use diffinst then genhis
           to generate this file.

  dcout: reports the coordinates and flexible information for all sites.

description

   Diri takes the rixyin output of ri for two models and uses the binding
   distribution, from histog to determine the flexible information for all
   sites in a variable model.  This is the same value as discan gives.

examples

documentation

see also

   {The flexrbs paper uses this program:}
   https://alum.mit.edu/www/toms/papers/flexrbs/

   {Instructions for using this program:}
   https://alum.mit.edu/www/toms/paper/flexrbs/refineinst.html

   {The makeinst program will convert dcout to Delila instructions:}
   https://alum.mit.edu/www/toms/delila/makeinst.html

   {Related programs:} discan.p, diffinst.p, ri.p

author
   Rye Shultzaberger

bugs
   Diri checks to see if the absolute value of the difference between the A
   coordinate and B coordinate is equal to any absolute value of a distance
   in the histog.  If your model can have a positive or negative spacing,
   this program will hit both spacings and will report the total information
   for both spacings.  This is not an issue for most of the applications of
   this, where the one component is either always upstream or downstream of
   the other component.

technical notes

*)
(* end module describe.diri *)

   maxarray = 1000; (* max size of array *)

type

(* begin module diri.type *)
   box = record (* the spacing distribution information *)
      range : real;       (* the numbers that make up the range of the dist *)
      occur : integer;    (* the number of occurrences at each position *)
      probdist : real;    (* the log2 (prob for each position) *)
      isitzero : boolean;    (* checks to see if the occurrence is 0 *)
   end;

   (* an array of all the distributrion info *)
   map = array[1..maxarray] of box;

(* end module diri.type *)

var
   ari, (* file used by this program *)
   histog, (* file used by this program *)
   dcout, (* file used by this program *)
   bri: text; (* file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

(* begin module skipblanks *)

(* 2003 July 31: tab is considered a blank character *)

function isblank(c: char): boolean;
(* is the character c blank or tab? *)
const
   tab = 9; (* tab character *)
begin
   isblank := (c = ' ') or (ord(c) = tab)
end;

procedure skipblanks(var thefile: text);
(* skip over blanks until a non-blank, or end of line, is found *)
begin
   while isblank(thefile^) and not eoln(thefile) do get(thefile);
end;

procedure skipnonblanks(var thefile: text);
(* skip over nonblanks until a blank, or end of line, is found *)
begin
   while (not isblank(thefile^)) and not eoln(thefile) do get(thefile);
end;

procedure skipcolumn(var thefile: text);
(* skip over a data column *)
begin
   skipblanks(thefile); skipnonblanks(thefile)
end;
(* end module skipblanks version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

(* begin module PatentMessage *)
procedure PatentMessage(var f: text);
(* Write the patent message to file f
version = 3.02 of PatentMessage_module 2017 Jul 31
*)
procedure makebar;
begin
writeln(f,
'**********************************************************************');
end;
begin
makebar;
writeln(f,
'* By downloading this code you agree to the Source Code Use License: *');
writeln(f,
'* https://alum.mit.edu/www/toms/Source_Code_Use_License_Schneider.txt *');
writeln(f,
'* Contact: https://alum.mit.edu/www/toms/contacts.html                *');
makebar;
end;
(* end module PatentMessage version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

(* begin module distribution *)
procedure distribution(var thefile : text;
                       var distsize: integer;
                       var correction: real;
                       var  store : map);
(* reads in distribution information from the file and stores it in
   an array *)
var
   x: integer; (* loop index *)
   normvalue: real; (* the occurrence / sum *)
   sum: integer; (* number of sites in distribution *)
   ln2: real; (* the natural log of 2 *)
   c: char;
begin
   reset(thefile);
   read(thefile,c);
   while c = '*' do begin
      readln(thefile);
      read(thefile,c);
   end;
 
   distsize := 0;
   while not eof(thefile) do begin
      distsize := distsize + 1;
      (* read in the positions that make up range *)
      read(thefile,store[distsize].range);
      readln(thefile,store[distsize].occur);
      (* read in the # of occurrences at each position *)
      if round(store[distsize].range) <> store[distsize].range then begin
         writeln(output, 'distribution values must be increments of 1');
         halt;
      end;
   end;
 
   sum := 0;
   for x:= 1 to distsize do begin
      sum := sum + store[x].occur; (* the sum of occurrences at all positions *)
   end;
 
   ln2 := ln(2);
   for x := 1 to distsize do begin
      store[x].isitzero := false;
      normvalue := store[x].occur / sum; (* normalize distribution values *)
      if (normvalue = 0) then begin
          store[x].isitzero := true;
      end
      else store[x].probdist := -1 * ln(normvalue) / ln2;
   end;  
 
   (* calculate sample correction *)
   correction := (distsize - 1 ) / ( 2 * ln( 2 ) * sum );
end;
 
(* end module distribution *)

(*****************************************************************************)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

(* begin module diloop *)
procedure diloop(correction: real; distsize: integer; store: map);
(* primary loop of the program *)
var
   c: char; (* generic character *)
   sum: real; (* flexible information *)
   coorda: integer; (* coord for site in model A *)
   coordb: integer; (* coord for site in model B *)
   ria: real; (* the ri value for the site in model A *)
   rib: real; (* the ri value for the site in model B *)
   z: integer; (* counter *)
   x: integer; (* counter *)
   difference: integer; (* difference in distance between A and B *)
   foundsite: boolean; (* True if site space matches a histog value *)

begin
   reset(ari);  
   reset(bri);  
   rewrite(dcout);  
   writeln(output,'* diri ',version:4:2);
   writeln(dcout, '* diri ',version:4:2);
   writeln(dcout,'*  loc. a    loc. b        dist.',
                 '  Ri A        Ri B        log2(p.d.) tot info');
   writeln(dcout,'*  Sample correction value = ',correction:1:6);
   while not eof(ari) and not eof(bri) do begin
      (* Danger: this code assumes that ari comments MATCH those
         in bri.  But this may not be true if, for example
         one has a wave and the other does not. *)
      read(ari,c);
      read(bri,c);
      if c ='*' then begin
{ original method did not record in dcout the source files:
         readln(ari);
         readln(bri);
}
         write(dcout,'*'); copyaline(ari, dcout);
         write(dcout,'*'); copyaline(bri, dcout);
      end
      else begin
         for x:= 1 to 4 do skipcolumn(ari);
         for x:= 1 to 4 do skipcolumn(bri);
         read(ari,coorda);
         read(bri,coordb);
         read(ari,ria);
         read(bri,rib);
         difference:= coorda - coordb;
         foundsite:= false;
         for z:= 1 to distsize do begin
            if (abs(difference) = abs(store[z].range)) and
               (store[z].occur > 0) then begin
               sum:= ria + rib - store[z].probdist + correction;
               foundsite:= true;

               write(dcout, coorda:10);
               write(dcout, coordb:10);
               write(dcout,-abs(difference):10);
               write(dcout,ria:12:6);
               write(dcout,rib:12:6);
               write(dcout,store[z].probdist:12:6);
               writeln(dcout, sum:12:6);

            end; 
         end; 
         if foundsite = false then begin
            write(dcout,'* not site');
            write(dcout, coorda:10);
            writeln(dcout, coordb:10);
         end;
         readln(ari);
         readln(bri);
      end;
   end;
end;
(* end module diloop *)

(* begin module diri.themain *)
procedure themain;
(* the main procedure of the program *)
var
   distsize: integer; (* for dist procedure *)
   correction: real; (* for dist procedure *)
   store: map; (* for dist procedure *)
begin
   writeln(output,'diri ',version:4:2);

   PatentMessage(output);

   distribution(histog, distsize, correction, store);
writeln(output,store[20].occur:2);
writeln(output,'here');
   diloop(correction, distsize, store);

end;
(* end module diri.themain *)

begin
   themain;
   writeln(output,'done');
1: end.
