program epsclean (input, output);
(* epsclean: clean an eps file

  Tom Schneider
  network address: toms@alum.mit.edu


  1993
 *)

label 1; (* end of program *)

const
(* begin module version *)
version = 2.09; (* of epsclean.p 1993 January 26
origin 1992 July 23 *)
(* end module version *)

(* begin module describe.epsclean *)
(*
name
   epsclean: clean an eps file

synopsis
   epsclean(input: in, output: out)

files
   input: eps file from microtex scanner
   output: cleaned eps file ready to print

description

   1. On the Mac:  scan an image into the mac using the microtex B&W scanner
   software.  Use:
      EPS format
      600 dpi
      2x2 high contrast
      line art

   2. Use the fetch program to move it to Unix with parameters:
      binary
      rawdata

   3. On Unix, the file comes over with all control-M's instead of newlines (ie
   instead of ascii 012).  However, if all controm-M's were converted to
   newlines, the image would be wrecked because it apparently contains control
   M's!

   4. Run the file through this program to make it useable.  This program works
   by finding the second occurance of the word 'dopic'.  Before this point,
   control-M's are converted to ascii 012, after this point they are left alone.
   The first occurance of 'dopic' is the definition of the "do picture" routine
   that defines the image.  The second one calls the routine, so the raw
   data follow that point.

examples

documentation

see also

author
   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.epsclean *)

(* begin module interact.const *)
      maxstring = 150; (* the maximum string *)
(* end module interact.const *)

(* begin module filler.const *)
      fillermax = 50; (* the size of the filler array for a string *)
(* end module filler.const *)

type
(* begin module interact.type *)
      string = record (* a string of characters *)
         letters: array[1..maxstring] of char; (* the letters in the string *)
         length: integer; (* the number of characters in the string *)
         current: integer; (* the letter we are working on *)
      end;
(* end module interact.type *)
 
(* begin module filler.type *)
      (* the following is an array used to fill a string.
      it is convenient to have it much shorter than the maxstring, so that
      it is easy to fill the string using procedure fillstring.
      the user must declare the value of constant fillermax. *)
      filler = packed array[1..fillermax] of char;
(* end module filler.type *)
 
(* begin module trigger.type *)
      trigger = record (* an object to be searched for *)
         seek: string; (* the characters looked for *)
         state: integer; (* how close to triggering we are *)
         skip: boolean; (* trigger not found- skip the line *)
         found: boolean (* the trigger was found *)
      end;
(* end module trigger.type *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.16 84 mar 12 tds/gds'; *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline *)

(* begin module interact.clearstring *)
procedure clearstring(var ribbon: string);
(* empty the string *)
var   index: integer; (* to the ribbon *)
begin (* clearstring *)
      with ribbon do begin
         for index := 1 to maxstring do letters[index] := ' ';
         length := 0;
         current := 0;
      end
end; (* clearstring *)
(* end module interact.clearstring *)


(* begin module filler.fillstring *)
procedure fillstring(var s: string; a: filler);
(* this procedure makes it reasonably easy to fill the string s with
characters.  one calls the procedure as: *)
(*                           1         2         3         4         5 *)
(*                  12345678901234567890123456789012345678901234567890 *)
(*   fillstring(s, 'this-is-the-string                                ');
the two comments make it easy to line the characters up. also, for this
example, it was assumed that the length of filler as defined by the
constant fillermax was 50. *)
var
      length: integer; (* of the string without trailing blanks *)
      index: integer; (* of s *)
begin (* fillstring *)
      clearstring(s);
      length := fillermax;
      while (length > 1) and (a[length] = ' ') do length := pred(length);
      if (length = 1) and (a[length] = ' ') then begin
         writeln(output, 'fillstring: the string is empty');
         halt
      end;

      for index := 1 to length do s.letters[index] := a[index];
      s.length := length;
      s.current := 1
end; (* fillstring *)
(* end module filler.fillstring *)

(* begin module filler.filltrigger *)
procedure filltrigger(var t: trigger;
                      a: filler);
(* fill the trigger t *)
begin (* filltrigger *)
         fillstring(t.seek,a)
end; (* fillstring *)
(* end module filler.filltrigger *)


(* begin module trigger.proc *)
(* this module allows one to scan a series of characters, as from
an array or a file, and to "trigger" or detect a simple string
in the series.  the advantage of the trigger is that several triggers
can "observe" a stream of characters at once, each looking for a
different thing.
some other modules required: interact.const, interact.type *)

procedure resettrigger(var t: trigger);
(* reset the trigger to ground state *)
begin (* resettrigger *)
      with t do begin
         state := 0;
         skip := false;
         found := false
      end
end; (* resettrigger *)
 
procedure testfortrigger(ch: char; var t: trigger);
(* look at the character ch.
   if it is part of the trigger (at the current trigger state),
       then the trigger state goes higher.
   if it is not part of the trigger then the trigger state is reset,
      skip is true and one should skip onward to find the trigger.
   if the trigger is found, found is true. *)
begin (* testfortrigger *)
      with t do begin
         state := succ(state);
(*         if debugging then begin
            writestring(list,seek);
            writeln(list,'testfortrigger seek.letters[',state:1,']:',
                           seek.letters[state],' ch:',ch);
         end;*)
         if seek.letters[state] = ch
         then begin
            skip := false;
            if state = seek.length then found := true
                                   else found := false
         end
         else begin (* reset trigger *)
            state := 0;
            skip := true;
            found := false
         end
      end
end; (* testfortrigger *)
(* end module trigger.proc *)

procedure copydrop(trig: trigger; var f, t: text);
(* copy from file f to file t 
to second occurance of string 'dopic', then
drop the file like a hot potato! *)
var
   count: integer; (* number of trigger strings found *)
begin
   reset(f);
   resettrigger(trig);

   count := 0;
   while count < 2 do begin
      testfortrigger(f^,trig);
      if trig.found then begin
         count := count + 1;
         resettrigger(trig);
      end;
      if eoln(f)
      then begin
         readln(f);
         writeln(t)
      end
      else begin
(*
         writeln(t,'|',f^,'|', ord(f^):3);
 this showed that the ord of control-M is 13, or octal 15.
*)
         if ord(f^) = 13 (* which is \015, the control-M carriage return *)
         then t^ := chr(10) (* which is \012, the newline character *)
         else t^ := f^; (* transfer *)
         put(t);
         get(f);
      end;
   end;
   t^ := chr(10); (* which is \012, the newline character *)
   put(t);
  
end;

procedure skipcopy(   {trig: trigger;}
                   var f, t: text);
(* skip from file f to file t until after 
to second occurance of string 'dopic', then
copy to the end of the file! *)
{
var
   count: integer; (* number of trigger strings found *)
}
begin
(*
   reset(f);
   resettrigger(trig);

   count := 0;
   while count < 2 do begin
      testfortrigger(f^,trig);
      if trig.found then begin
         count := count + 1;
         resettrigger(trig);
      end;
      if eoln(f)
      then begin
         readln(f);
      end
      else begin
         get(f);
      end;
   end;
*)
  
   while not eof(f) do begin
      if eoln(f)
      then begin
         readln(f);
         writeln(t)
      end
      else begin
         t^ := f^; (* transfer *)
         put(t);
         get(f);
      end;
   end;

end;

(* begin module epsclean.themain *)
procedure themain;
(* the main procedure of the program *)
var
  trig: trigger; (* contains the search word *)
begin
(*
   writeln(output,'epsclean ',version:4:2);
*)
(*                           1         2         3         4         5 *)
(*                  12345678901234567890123456789012345678901234567890 *)
   fillstring(trig.seek, 'dopic                                             ');

   copydrop(trig, input, output);
   skipcopy({trig,} input, output);
end;
(* end module epsclean.themain *)

begin
   themain;
1: end.
