program exponential(exponentialp, exponentialdata, output);
(* exponential: create an exponential curve for binding site separtions

  Dr. Thomas D. Schneider
  toms@alum.mit.edu
  permanent email: toms@alum.mit.edu
  https://alum.mit.edu/www/toms/


 *)

label 1; (* end of program *)

const
(* begin module version *)
version = 1.03; (* of exponential.p 2003 Aug 22
2003 Aug 22, 1.03: clean
2003 Aug 22, 1.02: got the logic finally
2003 Aug 22, 1.01: clean
2003 Aug 22, 1.00: origin
*)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.exponential *)
(*
name
   exponential: create an exponential curve for binding site separtions

synopsis
   exponential(exponentialp: in, exponentialdata: out, output: out)

files

   exponentialp:  parameters to control the program.  The file must contain the
      following parameters, one per line:

      parameterversion: The version number of the program.  This allows the
         user to be warned if an old parameter file is used.

      G (integer): size of the genome.

      n (integer): number of sites on the genome.

      maxdist (integer): largest distance to compute

   exponentialdata:  The number of sites predicted at spacings from 1
      to maxdist.

description

   For a given genome of size G and a set of sites n,
   give the distribution of spacings between sites.

   let p = n/G be the probability of a Fis site at a base.
   let q = (1-p) be the probability that a Fis site is not at a base

   Then the probability of having two fis sites separated by
   a distance d is p(q^d)p = p^2(q^d).

   Therefore the number of fis sites separated by exactly
   a distance d with no others between is Gp^2((1-p)^d).

   This is equivalent to dropping n sites randomly into the range
   between 1 and G.  Sort these and compute the distances between
   them.  Tabulate the distribution of these distances.

examples

documentation

see also

   {Use random numbers to simulate this process: } spray.p

author

   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.exponential *)

var
   exponentialp, (* file used by this program *)
   exponentialdata: text; (* file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.16 84 mar 12 tds/gds'; *)

(* begin module exponential.themain *)
procedure themain(var exponentialp, exponentialdata: text);
(* the main procedure of the program *)
var
   c: real; (* Gp^2(1-p)^d, predicted count at a particular distance *)
   d: integer; (* index to maxdist *)
   G: integer; (* genome size *)
   maxdist: integer; (* largest distance to compute *)
   n: integer; (* number of sites *)
   p: real; (* probability of having d sites without one inbetween *)
   parameterversion: real; (* parameter version number *)
   q: real; (* 1-p *)

begin
   writeln(output,'exponential ',version:4:2);
   reset(exponentialp);
   readln(exponentialp, parameterversion);
   if round(100*parameterversion) < round(100*updateversion) then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;

   readln(exponentialp, G);
   readln(exponentialp, n);
   readln(exponentialp, maxdist);

   rewrite(exponentialdata);
   writeln(exponentialdata,'* exponential ',version:4:2);
   writeln(exponentialdata,'* ',G:4,' size of genome:  G');
   writeln(exponentialdata,'* ',n:4,' number of sites: n');
   writeln(exponentialdata,'* ',maxdist:4,' maximum distance computed: maxdist');

   p := n/G;
   q := 1-p;
   c := p*p * G;

   writeln(output, '***************');
   writeln(output, 'n     ',n    :8  );
   writeln(output, 'G     ',G    :8  );
   writeln(output, 'p     ',p    :8:5);
   writeln(output, 'q     ',    q:8:5);
   writeln(output, 'q*p*G ',q*p*G:8:5);
   writeln(output, 'c     ',c    :8:5);
   writeln(output, 'p*p*G ',p*p*G:8:5);
   writeln(output, '***************');

   for d := 0 to maxdist do begin
      if d > 0 then c := c * q;
      writeln(output, 'for d = ',d:2,' c = (1-p)^d: ',c:8:5);
      writeln(exponentialdata, d:5,' ',round(c):5);
   end;

end;
(* end module exponential.themain *)

begin
   themain(exponentialp, exponentialdata);
1: end.
