program freb(frebp, book, output);
(* freb: frequency table to Delila book

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  toms@alum.mit.edu (permanent)
  https://alum.mit.edu/www/toms (permanent)

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.18; (* of freb.p 2010 Nov 15
2010 Nov 15, 1.18: don't stop because of up/down, just warn!!
2010 Nov 02, 1.17: not computing number sequences, maxmultiplier
2008 May 21, 1.16: sum is wrong
2008 May 21, 1.15: multiplier is wrong
2007 Nov 20, 1.14: point to transpose
2007 Jan 10, 1.13: allow real numbers
1999 Dec 13, 1.12: functional
1997 Apr 22, 1.00: origin from frese 1.01 1991 November 30 *)
(* end module version *)

(* begin module describe.freb *)
(*
name
   freb: frequency table to Delila book

synopsis
   freb(frebp: in, book: out, output: out)

files

   frebp: input frequency table (parameters to the program)

     The first line is the name of the site.

     The following lines each have a set of numbers, 5 per line,
     representing first the coordinate and then the numbers of a,c,g
     and t to use.

     If the numbers are integers, then those integers are used to
     create the sequences.  The numbers can be real in which case the
     smallest power of 10 multiplier is used to make the numbers
     integer.

   book: sequences which could have produced the frebp frequencies,
     in Delila book format.

   output: messages to the user

description

   The freb program converts a table of frequencies to a set of
   aligned sequences so they may be analyzed.  The raw sequences have
   the same frequencies, but, of course, are not the same as the
   original sequences.

examples

documentation

see also
   {program to create a sequence set from a table:} frese.p
   {program to create a book:} makebk.p
   {program to transpose the table:} transpose.p

   {example parameter file:} frebp

author
   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.freb *)

(* begin module freb.const *)
   maxarray = 100; (* largest possible array that can be stored *)
   maxmultiplier = 1000; (* largest number of sequences *)
(* end module freb.const *)

(* begin module interact.const *)
(* begin module string.const *)
      maxstring = 2000; (* the maximum string *)
(* end module string.const version = 4.86; (@ of prgmod.p 2004 Sep 8 *)
(* end module interact.const version = 7.70; {of delmod.p 2005 Sep 15} *)

(* begin module book.const ***************************************************)
(* constants needed for book manipulations *)
      dnamax = 1024;  (* length of dna arrays *)
      namelength = 100; (* maximum key name length *)
      linelength = 200; (* maximum line readable in book *)
(* end module book.const version = 7.70; {of delmod.p 2005 Sep 15} *)

(* begin module datetime.const *)

   datetimearraylength = 19; (* length of dataarray for dates,
       It is just long enough to include the 4 digit year - solving the
       year 2000 problem:
1980/06/09 18:49:11
123456789 123456789 
         1         2
*)

(* end module datetime.const version = 7.70; {of delmod.p 2005 Sep 15} *)
 
type
(* begin module interact.type *)
(* begin module string.type *)
   stringptr = ^string; (* pointer to a string *)
   string = record (* a string of characters *)
      letters: array[1..maxstring] of char; (* the letters in the string *)
      length: integer; (* the number of characters in the string *)
      current: integer; (* the letter we are working on *)
      next: stringptr; (* the next string in a series *)
   end;
(* end module string.type version = 4.86; (@ of prgmod.p 2004 Sep 8 *)
(* end module interact.type version = 7.70; {of delmod.p 2005 Sep 15} *)

(* begin module book.type *)
(* types needed for book manipulations *)

      chset = set of 'a'..'z';

      (* types defined in book definition *)

      alpha = packed array[1..namelength] of char; (* this is not alfa *)

      (* name is a left justified string with blanks following the
         characters *)
      name = record
         letters: alpha;
         length: 0..namelength (* zero means an unspecified structure *)
      end;

      lineptr = ^line;
      line = record (* a line of characters *)
         letters: packed array [1..linelength] of char;
         length: 0..linelength;
         next: lineptr
      end;

      direction = (plus, minus, dircomplement, dirhomologous);
      configuration = (linear, circular);
      state = (on, off);

      header = record (* header of key *)
         keynam: name; (* key name of structure *)
         fulnam: lineptr; (* full name of structure *)
         note:   lineptr (* note key *)
      end;

(* begin module base.type *)
   (* define the four nucleotide bases *)
   base = (a,c,g,t);
(* end module base.type version = 7.70; {of delmod.p 2005 Sep 15} *)

      (* sequence types *)
      dnaptr = ^dnastring;
      dnarange = 0..dnamax;
      seq = packed array[1..dnamax] of base;
      dnastring = record
         part: seq;
         length: dnarange;
         next: dnaptr
      end;


      orgkey = record (* organism key *)
         hea: header;
         mapunit: lineptr (* genetic map units *)
      end;

      chrkey = record (* chromosome key *)
         hea: header;
         mapbeg: real; (* number of genetic map beginning *)
         mapend: real (* number of genetic map ending *)
      end;

      pieceptr = ^piece;
      piekey = record (* piece key *)
         hea: header;
         mapbeg: real; (* genetic map beginning *)
         coocon: configuration; (* configruation (circular/linear) *)
         coodir: direction; (* direction (+/-) relative to genetic map *)
         coobeg: integer; (* beginning nucleotide *)
         cooend: integer; (* ending nucleotide *)
         piecon: configuration; (* configruation (circular/linear) *)
         piedir: direction; (* direction (+/-) relative to coordinates *)
         piebeg: integer; (* beginning nucleotide *)
         pieend: integer; (* ending nucleotide *)
      end;


      piece = record
         key: piekey;
         dna: dnaptr
      end;

      reference = record
         pienam : name;  (* name of piece referred to *)
         mapbeg : real;  (* genetic map beginning *)
         refdir : direction;  (* direction relative to coordinates *)
         refbeg : integer;  (* beginning nucleotide *)
         refend : integer;  (* ending nucleotide *)
      end;

      genkey = record  (* gene key *)
         hea : header;
         ref : reference;
      end;

      trakey = record  (* transcript key *)
         hea : header;
         ref : reference;
      end;

      markerptr = ^marker;
      markey = record (* marker key *)
         hea : header;
         ref : reference;
         sta : state;
         phenotype : lineptr;
         next : markerptr;
      end;

      marker = record
         key : markey;
         dna : dnaptr;
      end;

(* end module book.type version = 7.70; {of delmod.p 2005 Sep 15} *)

(* begin module datetime.type *)
   (* array for dates *)
   datetimearray = packed array[1..datetimearraylength] of char;
(* end module datetime.type version = 7.70; {of delmod.p 2005 Sep 15} *)
 
var
   frebp, book: text; (* files used by this program *)

(* begin module book.var *)
(* ************************************************************************ *)
(* global variables needed for book manipulations *)

      (* free storage: *)
      freeline: lineptr; (* unused lines *)
      freedna: dnaptr; (* unused dnas *)

      readnumber: boolean; (* whether to read a number from the notes, or
                              to read in the notes *)
      number: integer; (* the number of the item just read *)
      numbered: boolean; (* true when the item just read is numbered *)
      skipunnum: boolean; (* a control variable to allow skipping of
                             un-numbered items in the book *)

(* ************************************************************************ *)
(* end module book.var version = 7.70; {of delmod.p 2005 Sep 15} *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 7.70; {of delmod.p 2005 Sep 15} *)

(* begin module book.basis *)
(* procedures needed for book manipulations *)

(* get procedures should be used for all linked lists of records *)

procedure getline(var l: lineptr);
(* obtain a line from the free line list or by making a new one *)
begin
      if freeline<>nil
         then begin
            l:=freeline;
            freeline:=freeline^.next
         end
         else new(l);
      l^.length:=0;
      l^.next:=nil
end;

procedure getdna(var l: dnaptr);
begin
      if freedna<>nil
         then begin
            l:=freedna;
            freedna:=freedna^.next
         end
         else new(l);
      l^.length:=0;
      l^.next:=nil
end;

(* clear procedures should be called each time the records are no longer needed
         failure to do this may result in a stack overflow. *)

procedure clearline(var l: lineptr);
(* return a line to the free line list *)
var   lptr: lineptr;
begin
      if l<>nil then begin
         lptr:=l;
         l:=l^.next;
         lptr^.next:=freeline;
         freeline:=lptr
      end
end;

procedure writeline(var afile: text; l: lineptr;
                        carriagereturn: boolean);
(* write a line to a file, with carriage return if
carriagereturn is true. *)
var
   index: integer; (* index to characters in l *)
begin
   with l^ do begin
      for index := 1 to length
      do write(afile, letters[index]);
   end;
   if carriagereturn then writeln(afile);
end;

procedure showfreedna;
(* show the freedna list *)
var
   counter: integer; (* count of freedna list *)
   l: dnaptr; (* pointer into freedna list *)
begin
   l := freedna;
   counter := 0;
   while l <> nil do begin
      counter := succ(counter);
      write(output,counter:1);
      write(output, ', length = ',l^.length:1);
{
This is illegal according to gpc because one cannot
write a pointer to a text file.  It can be unearthed
for debugging.
      write(output, ', pointer id: ',l:1);
}
      writeln(output);
      l := l^.next
   end;
end;

procedure cleardna(var l: dnaptr);
(* clear the dna strutures to the free list *)
var   lptr: dnaptr;
begin
      if l<>nil then begin
         lptr:=l;
         l:=l^.next;
         lptr^.next:=freedna;
         freedna:=lptr
      end
end;

procedure clearheader(var h: header);
(* clear the header h (remove lines to free storage) *)
begin
      with h do begin
         clearline(fulnam);
         while note<>nil do clearline(note)
      end
end;

procedure clearpiece(var p: pieceptr);
(* clear the dna of the piece *)
begin
      while p^.dna<>nil do cleardna(p^.dna);
      clearheader(p^.key.hea)
end;

function chartobase(ch:char):base;
(* convert a character into a base *)
begin
      case ch of
         'a': chartobase:=a;
         'c': chartobase:=c;
         'g': chartobase:=g;
         't': chartobase:=t
      end
end;

function basetochar(ba:base):char;
(* convert a base into a character *)
begin
      case ba of
         a: basetochar:='a';
         c: basetochar:='c';
         g: basetochar:='g';
         t: basetochar:='t';
      end
end;

function complement(ba:base):base;
(* take the complement of ba *)
begin
      case ba of
         a: complement:=t;
         c: complement:=g;
         g: complement:=c;
         t: complement:=a;
      end
end;

function chomplement(b: char): char;
(* create the character complement of base b.  I must be getting hungry! *)
begin
  chomplement := basetochar(complement(chartobase(b)));
end;

function pietoint(p: integer; pie: pieceptr): integer;
(* p is a coordinate on the piece.
   we want to transform p into a number
   from 1 to n: an internal coordinate system for
   easy manipulation of piece coordinates *)
(* Note: the dirhomologous and dircomplement are treated as
plus and minus directions, which MIGHT NOT BE RIGHT! *)
var   i: integer; (* an intermediate value *)
begin
      with pie^.key do begin
         case piedir of
            dirhomologous,
            plus:  if p>=piebeg
                   then i:=p-piebeg+1
                   else i:=(p-coobeg)+(cooend-piebeg)+2;
            dircomplement,
            minus: if p<=piebeg
                   then i:=piebeg-p+1
                   else i:=(cooend-p)+(piebeg-coobeg)+2
         end;
         pietoint:=i
      end
end;

function inttopie(i: integer; pie: pieceptr):integer;
(* i is in the range 1 to some maximum.  it is an internal coordinate
   system for the program.  we want to do a
   coordinate transformation to obtain
   a value in the range of the piece called pie:
   i=1 corresponds to piebeg and
   i=its maximum corresponds to pieend *)
(* Note: the dirhomologous and dircomplement are treated as
plus and minus directions, which MIGHT NOT BE RIGHT! *)
var   p: integer; (* an intermediate value *)
begin
      with pie^.key do begin
         case piedir of
            dirhomologous,
            plus: begin
               p:=piebeg+(i-1);
               if p>cooend
                  then if coocon=circular
                     then p:=p-(cooend-coobeg+1)
            end;
            dircomplement,
            minus: begin
               p:=piebeg-(i-1);
               if p<coobeg
                  then if coocon=circular
                     then p:=p+(cooend-coobeg+1)
            end
         end;
         inttopie:=p
      end
end;

function piecelength(pie: pieceptr): integer;
(* return the length of the dna in pie *)
begin
      piecelength:=pietoint(pie^.key.pieend,pie)
end;

(* end module book.basis version = 7.70; {of delmod.p 2005 Sep 15} *)

(* begin module package.bwrite *)
(****************************************************************************)
(* this is a package of procedures for writing books,
   by gary stormo,  aug 17, 1982 *)

(* begin module book.bwbasics *)

procedure bwstartline(var book: text);
(* start a line of output to the book *)
begin
      write(book,'* ')
end;

procedure bwline(var book: text; l: lineptr);
(* write a line to the book *)
var   i: integer;
begin
   if l<>nil then begin
      bwstartline(book);
      if l^.length<>0
      then for i:=1 to l^.length do write(book,l^.letters[i]);
      writeln(book);
   end
end;

procedure bwtext(var book: text; lines: lineptr);
(* write a set of lines to the book *)
var 
   l: lineptr;   
begin 
   l := lines; 
   while l<>nil do begin
      bwline(book,l);
      l := l^.next; 
   end; 
end;

procedure bwnote(var book: text; note: lineptr);
(* writes the notes pointed to by 'note' to 'book' *)
begin
   if note<>nil then begin
      writeln(book,'note');
      bwtext(book,note);
      writeln(book,'note');
   end
end;

procedure bwnumber(var book: text; num: integer);
(* write a number to the book *)
begin
   bwstartline(book);
   writeln(book,num:1) (* pascal will expand the field as far as needed *)
end;

procedure bwreanum(var book: text; reanum: real);
(* write a real number to the book *)
begin
   bwstartline(book);
   writeln(book,reanum:1:2) (* pascal will expand the field *)
end;

procedure bwstate(var book: text; sta: state);
(* write a state to the book *)
begin
   bwstartline(book);
   case sta of
      on:  writeln(book,'on');
      off: writeln(book,'off')
   end
end;

procedure bwname(var book: text; nam: name);
(* write a name to the book *)
var   i: integer;
begin
   with nam do begin
      bwstartline(book);
      for i:=1 to length do write(book,letters[i]);
      writeln(book)
   end
end;

procedure bwdirect(var book: text; direct: direction);
(* write a direction to the book *)
begin
   bwstartline(book);
   case direct of
                  dirhomologous, (* handle case, may not be right *)
      plus:  writeln(book,'+');
                  dircomplement, (* handle case, may not be right *)
      minus: writeln(book,'-')
   end
end;

procedure bwconfig(var book: text; config: configuration);
(* write a configuration to the book *)
begin
   bwstartline(book);
   case config of
      linear:   writeln(book,'linear');
      circular: writeln(book,'circular')
   end
end;

procedure bwheader(var book: text; hea: header);
(* write a key header to the book *)
begin
   with hea do begin
      bwname(book,keynam);
      bwline(book,fulnam);
      bwnote(book,note);
   end
end;
 
procedure bworgkey(var book: text; org: orgkey);
(* write the organism key *)
begin
   with org do begin
      bwheader(book,hea);
      bwline  (book,mapunit)
   end
end;
 
procedure bwchrkey(var book: text; chr: chrkey);
(* write the chromosome key *)
begin
   with chr do begin
      bwheader(book,hea);
      bwreanum(book,mapbeg);
      bwreanum(book,mapend)
   end
end;

(* end module book.bwbasics version = 7.70; {of delmod.p 2005 Sep 15} *)

(* begin module book.bworg *)
procedure bworg(var thefile: text; org: orgkey;
                var chropen, orgopen: boolean);
(* this writes the organism key 'org' to 'thefile', and returns 'orgopen'
   as true.  if there is already a chromosome or organism open they
   are closed. *)
begin
      if chropen then begin
         writeln(thefile,'chromosome');
         chropen := false;
      end;
      if orgopen then writeln(thefile,'organism');
      writeln(thefile,'organism');
      bworgkey(thefile,org);
      orgopen := true;
end;
(* end module book.bworg version = 7.70; {of delmod.p 2005 Sep 15} *)

(* begin module book.bwchr *)
procedure bwchr(var thefile: text; chr: chrkey; var chropen: boolean);
(* write the chromosome key 'chr' to 'thefile' and return chropen as
   true.  if a chromosome is already open it is closed. *)
begin
      if chropen then writeln(thefile,'chromosome');
      writeln(thefile,'chromosome');
      bwchrkey(thefile,chr);
      chropen := true;
end;
(* end module book.bwchr version = 7.70; {of delmod.p 2005 Sep 15} *)

(* begin module book.bwdna *)
procedure bwdna(var thefile: text; d: dnaptr);
(* write the dna pointed to by 'd' to 'thefile' *)
var i: integer; (* index to the sequence *)
    l: integer; (* index to the number of bases on the line *)
    newline: boolean; (* true when a new line should be started *)
begin
   writeln(thefile,'dna');
   newline := true;
   while (d <> nil) do begin
      for i := 1 to d^.length do begin
         if newline then begin
            bwstartline(thefile);
            l := 0;
            newline := false
         end;
         write(thefile,basetochar(d^.part[i]));
         l := succ(l);
         if (l mod 60 = 0) or  (* end of line *)
            ((i = d^.length) and (d^.next = nil)) (* last base *)
         then begin
            writeln(thefile);
            newline := true
         end;
      end;
      d := d^.next;
   end;
   if not newline then writeln(thefile); (* last base *)
   writeln(thefile,'dna');
end;  
(* end module book.bwdna version = 7.70; {of delmod.p 2005 Sep 15} *)

(* begin module book.bwpie *)
procedure bwpie(var thefile: text; pie: pieceptr);
(* writes the information pointed to by 'pie' to 'thefile' *)
begin
   writeln(thefile,'piece');
   with pie^ do begin
      bwheader(thefile,key.hea);

      bwstartline(thefile);
      writeln(thefile,key.mapbeg:1:2);

      bwstartline(thefile);
      if (key.coocon = circular) then writeln(thefile,'circular')
            else writeln(thefile,'linear');

      bwstartline(thefile);
      if (key.coodir = plus) then writeln(thefile,'+')
            else writeln(thefile,'-');

      bwstartline(thefile);
      writeln(thefile,key.coobeg:1);

      bwstartline(thefile);
      writeln(thefile,key.cooend:1);

      bwstartline(thefile);
      if (key.piecon = circular) then writeln(thefile,'circular')
            else writeln(thefile,'linear');

      bwstartline(thefile);
      if (key.piedir = plus) then writeln(thefile,'+')
            else writeln(thefile,'-');

      bwstartline(thefile);
      writeln(thefile,key.piebeg:1);

      bwstartline(thefile);
      writeln(thefile,key.pieend:1);
   end;
   bwdna(thefile,pie^.dna);
   writeln(thefile,'piece');
end;
(* end module book.bwpie version = 7.70; {of delmod.p 2005 Sep 15} *)

(* begin module book.bwref *)
procedure bwref(var book: text; ref: reference);
(* write a key reference to the book *)
begin
   with ref do begin
      bwname  (book,pienam);
      bwreanum(book,mapbeg);
      bwdirect(book,refdir);
      bwnumber(book,refbeg);
      bwnumber(book,refend)
   end
end;
(* end module book.bwref version = 7.70; {of delmod.p 2005 Sep 15} *)

(* begin module book.bwgen *)
procedure bwgen(var thefile: text; gene: genkey);
(* this proecdure writes to 'thefile' the information in 'gene',
   properly formatted; *)
begin
      writeln(thefile,'gene');
      with gene do
      begin
         bwheader(thefile,hea);
         bwref(thefile,ref);
      end;
      writeln(thefile,'gene');
end;
(* end module book.bwgen version = 7.70; {of delmod.p 2005 Sep 15} *)

(* begin module book.bwtra *)
procedure bwtra(var thefile: text; trans: trakey);
(* this proecdure writes to 'thefile' the information in 'trans',
   properly formatted; *)
begin
      writeln(thefile,'transcript');
      with trans do
      begin
         bwheader(thefile,hea);
         bwref(thefile,ref);
      end;
      writeln(thefile,'transcript');
end;
(* end module book.bwtra version = 7.70; {of delmod.p 2005 Sep 15} *)

(* begin module book.bwmar *)
procedure bwmar(var thefile: text; mark: marker);
(* this proecdure writes to 'thefile' the information in 'mark',
   properly formatted; *)
var i : integer;
begin
      writeln(thefile,'marker');
      with mark.key do
      begin
         bwheader(thefile,hea);
         bwref(thefile,ref);
         bwstate(thefile,sta);
         bwstartline(thefile);
         for i := 1 to phenotype^.length do
            write(thefile,phenotype^.letters[i]);
         writeln(thefile);
      end;
      bwdna(thefile,mark.dna);
      writeln(thefile,'marker');
end;
(* end module book.bwmar version = 7.70; {of delmod.p 2005 Sep 15} *)

(****************************************************************************)
(* end module package.bwrite version = 7.70; {of delmod.p 2005 Sep 15} *)

(* begin module getdatetime *)
procedure getdatetime(var adatetime: datetimearray);
(* get the date and time into a single array from the system clock.
   adatetime contains the date:
       1980/06/09 18:49:11
         ye mo da ho mi se
(year, month, day, hour, minute, second).
As of 2000 February 18, the Sun Pascal compiler requires a formatting
statement.  This statement allows the date to be generated in this
standard Delila format in a single call.  Information about the
formatting statement is available on the manual page for date in Unix.
If a computer does not have this method, see the 'oldgetdatetime' routine
in delmod.p (https://alum.mit.edu/www/toms/delila/delmod.html)
for some conversion code.

GPC Functions:
function  GetUnixTime (var MicroSecond : Integer) : UnixTimeType;

http://agnes.dida.physik.uni-essen.de/~gnu-pascal/gpc_109.html#SEC109

7.10.8 Date And Time Routines 

procedure GetTimeStamp (var t : TimeStamp); 
function Date (t : TimeStamp) : packed array [1 .. DateLength] of Char; 
function Time (t : TimeStamp) : packed array [1 .. TimeLength] of Char; 

DateLength and TimeLength are implementation dependent constants. 

GetTimeStamp (t) fills the record `t' with values. If they are valid, the Boolean
flags are set to True. 

TimeStamp is a predefined type in the Extended Pascal standard. It may be
extended in an implementation, and is indeed extended in GPC. For the full
definition of `TimeStamp', see section 8.255 TimeStamp. 

*)
var
    t: TimeStamp; 

(* begin module pluckdigit *)
function pluckdigit(number, logplace:integer): char;
(* return the digit at the place value ('logplace') position of number.
example:
      pluckdigit(13625, 3) = 3
      pluckdigit(13625, 4) = 1
This routine was taken from module numberdigit in prgmod.p, but is
modified so as not to give the sign.  Instead it gives zeros
above the digits.  'myabsolute' replaced 'absolute', which
is apparently a keyword for GPC.  The name is kept for
to keep the code looking similar to its origin.
*)
var
      place: integer; (* the exponent of logplace *)
      count: integer; (* used to make place *)
      myabsolute: integer; (* the absolute value of number *)
      acharacter: char; (* the character to be returned *)
procedure digit;
(* extract a digit at the place position *)
var
      tenplace: integer; (* ten times place *)
      z: integer; (* an intermediate value *)
      d: integer; (* the digit extracted *)
begin (* digit *)
      tenplace:=10*place;
      z:=myabsolute-((myabsolute div tenplace)*tenplace);
      if place = 1
         then d:=z
         else d:= z div place;
      case d of
         0: acharacter:='0';
         1: acharacter:='1';
         2: acharacter:='2';
         3: acharacter:='3';
         4: acharacter:='4';
         5: acharacter:='5';
         6: acharacter:='6';
         7: acharacter:='7';
         8: acharacter:='8';
         9: acharacter:='9';
      end
end; (* digit *)
begin (* pluckdigit *)
      place:=1;
      for count:=1 to logplace do place:=10*place;
      if number=0 then begin
         acharacter:='0'
      end
      else begin
         myabsolute:=number;
         if myabsolute >= place
         then digit
         else acharacter := '0'
      end;
      pluckdigit:=acharacter
end; (* pluckdigit *)
(* end module pluckdigit *)

begin

 (* according to:
http://agnes.dida.physik.uni-essen.de/~gnu-pascal/gpc_109.html#SEC109
*)
   GetTimeStamp(t); 

(* Predefined time stamp:
http://agnes.dida.physik.uni-essen.de/~gnu-pascal/gpc_389.html#SEC389
  TimeStamp = {@@packed} record
      DateValid,
      TimeValid   : Boolean;
      Year        : Integer;
      Month       : 1 .. 12;
      Day         : 1 .. 31;
      DayOfWeek   : 0 .. 6;  { 0 means Sunday }
      Hour        : 0 .. 23;
      Minute      : 0 .. 59;
      Second      : 0 .. 61; { to allow for leap seconds }
      MicroSecond : 0 .. 999999
    end;
*)

  with t do begin
    if TimeValid then begin
{
       writeln(output,'valid time');
       writeln(output,'year   =',year:4);
       writeln(output,'month  =',month:2);
       writeln(output,'day    =',day:2);
       writeln(output,'hour   =',hour:2);
       writeln(output,'minute =',minute:2);
       writeln(output,'second =',second:2);
}
       adatetime := 'year/mm/dd hh:mm:ss';
       adatetime[ 1] := pluckdigit(Year,3);
       adatetime[ 2] := pluckdigit(Year,2);
       adatetime[ 3] := pluckdigit(Year,1);
       adatetime[ 4] := pluckdigit(Year,0);
       adatetime[ 6] := pluckdigit(Month,1);
       adatetime[ 7] := pluckdigit(Month,0);
       adatetime[ 9] := pluckdigit(Day,1);
       adatetime[10] := pluckdigit(Day,0);
       adatetime[12] := pluckdigit(Hour,1);
       adatetime[13] := pluckdigit(Hour,0);
       adatetime[15] := pluckdigit(Minute,1);
       adatetime[16] := pluckdigit(Minute,0);
       adatetime[18] := pluckdigit(Second,1);
       adatetime[19] := pluckdigit(Second,0);
    end
    else begin
       writeln(output,'getdatetime: invalid time!');
       halt;
    end
  end;
{
   Sun compiler method:
   date(adatetime,'%Y/%m/%d %H:%M:%S');
}
end;
(* end module getdatetime version = 7.70; {of delmod.p 2005 Sep 15} *)
 
(* begin module writedatetime *)
procedure writedatetime(var thefile: text; adatetime: datetimearray);
(* expand the date and time out and print in the file *)
var
      index: integer; (* index of datetime *)
begin
      for index:=1 to datetimearraylength do write(thefile,adatetime[index])
end;
(* end module writedatetime version = 7.70; {of delmod.p 2005 Sep 15} *)

(* begin module interact.clearstring *)
(* begin module clearstring *)
procedure clearstring(var ribbon: string);
(* empty the string *)
var   index: integer; (* to the ribbon *)
begin (* clearstring *)
      with ribbon do begin
         for index := 1 to maxstring do letters[index] := ' ';
         length := 0;
         current := 0;
      end
end; (* clearstring *)
procedure initializestring(var ribbon: string);
(* start the string with a nil pointer.   This routine
should be called before doing linked list work.  This allows
the standard string routines to clear the string without
killing the pointer. *)
begin (* initializestring *)
   clearstring(ribbon);
   ribbon.next := nil;
end; (* initializestring *)
(* end module clearstring version = 4.86; (@ of prgmod.p 2004 Sep 8 *)
(* end module interact.clearstring version = 7.70; {of delmod.p 2005 Sep 15} *)

(* begin module interact.getstring *)
procedure getstring(var afile: text; var buffer: string;
                    var gotten: boolean);
(* get a string from a file not using string calls.  this lets one
obtain lines from a file without interactive prompts *)
var   index: integer; (* of buffer *)
begin (* getstring *)
      clearstring(buffer);
      if eof(afile)
      then gotten := false
      else begin
         index := 0;
         while (not eoln(afile)) and (index < maxstring) do begin
            index := succ(index);
            read(afile, buffer.letters[index])
         end;

         if not eoln(afile) then begin
            writeln(output, ' getstring: a line exceeds maximum string size (',
                            maxstring:1,')');
            halt
         end;

         buffer.length := index;
         buffer.current := 1;
         readln(afile);
         gotten := true
      end
end; (* getstring *)
(* end module interact.getstring version = 4.20; (@ of prgmod.p 1997 March 15 *)

(* begin module interact.writestring *)
procedure writestring(var tofile: text; var s: string);
(* write the string s to file tofile, no writeln *)
var   i: integer; (* index to s *)
begin (* writestring *)
      with s do for i := 1 to length do write(tofile, letters[i])
end; (* writestring *)
(* end module interact.writestring version = 4.20; (@ of prgmod.p 1997 March 15 *)

(* begin module fillname *)
procedure fillname(thename: string; var n: name);
(* fill the name into the name type *)
var   i: integer; (* index to s *)
begin (* fillname *)
   for i := 1 to thename.length do n.letters[i] := thename.letters[i];
   n.length := thename.length;
{
      name = record
         letters: alpha;
         length: 0..namelength (* zero means an unspecified structure *)
      end;
}
end; (* fillname *)
(* end module fillname *)

(* begin module fillline *)
procedure fillline(thename: string; var l: lineptr);
(* fill the name into the line *)
var   i: integer; (* index to s *)
begin (* fillline *)
   for i := 1 to thename.length do l^.letters[i] := thename.letters[i];
   l^.length := thename.length;
   l^.next := nil;
{
      lineptr = ^line;
      line = record (* a line of characters *)
         letters: packed array [1..linelength] of char;
         length: 0..linelength;
         next: lineptr
      end;
}
end; (* fillline *)
(* end module fillline *)

(* begin module fillhead *)
procedure fillhead(thename: string; var head: header);
(* fill the header information *)
begin
   with head do begin
      fillname(thename,keynam);
      new(fulnam);
      fillline(thename,fulnam);
      note :=nil
   end;
end;
(* end module fillhead *)

(* begin module numberdigit *)
function numberdigit(number, logplace:integer): char;
(* return the digit at the place value ('logplace') position of number.
example:
      numberdigit(13625, 3) = 3
      numberdigit(13625, 4) = 1
*)
var
      place: integer; (* the exponent of logplace *)
      count: integer; (* used to make place *)
      absolute: integer; (* the absolute value of number *)
      acharacter: char; (* the character to be returned *)
procedure digit;
(* extract a digit at the place position *)
var
      tenplace: integer; (* ten times place *)
      z: integer; (* an intermediate value *)
      d: integer; (* the digit extracted *)
begin (* digit *)
      tenplace:=10*place;
      z:=absolute-((absolute div tenplace)*tenplace);
      if place = 1
         then d:=z
         else d:= z div place;
      case d of
         0: acharacter:='0';
         1: acharacter:='1';
         2: acharacter:='2';
         3: acharacter:='3';
         4: acharacter:='4';
         5: acharacter:='5';
         6: acharacter:='6';
         7: acharacter:='7';
         8: acharacter:='8';
         9: acharacter:='9';
      end
end; (* digit *)
procedure sign;
(* put a negative sign out or a positive sign *)
begin (* sign *)
      if number <0 then acharacter:='-'
                   else acharacter:='+'
end; (* sign *)
begin (* numberdigit *)
      place:=1;
      for count:=1 to logplace do place:=10*place;

      if number=0 then begin
         if place=1 then acharacter:='0'
                    else acharacter:=' '
      end
      else begin
         absolute:=abs(number);
         if absolute < (place div 10)
            then acharacter:=' '
            else if absolute >= place
                    then digit
                    else sign
      end;
      numberdigit:=acharacter
end; (* numberdigit *)
(* end module numberdigit version = 4.20; (@ of prgmod.p 1997 March 15 *)

(* begin module mknumber *)
procedure mknumber(n:integer; var l: lineptr);
(* make the number n into the line l *)
const
   basis = 3; (* location of zero digit of number *)
   numberlength = 10; (* length of number *)
var
   c: char; (* possible character of the number *)
   digit: integer; (* index to the number *)
   fillspot: integer; (* spot in the letters to fill *)
begin
   fillspot := basis;
   with l^ do begin
      for digit := numberlength downto 0 do begin
         c := numberdigit(n, digit);
         letters[fillspot] := c;
         if (c <> ' ') and (c <> '+')
         then fillspot := fillspot + 1;
      end;
      length := basis + numberlength;
   end;
end;
(* end module mknumber *)

(* begin module findmultiplier *)
procedure findmultiplier(x: real; var multiplier: integer);
(* find the multiplier to multiply the real number x by
so that there are only zeros after the decimal place.
Use multiples of 10 and use the previous multiplier value
to start with.
2010 Nov 02:  This failed with x = 0.06870 because
round of that equals trunc of that!
*)
const
  multmax = 1000000; (* maximum multiplier *)
  debugging = false; (* set true to watch the process *)
var
  done: boolean; (* done searching for the multiplier? *)
begin
  if debugging then writeln(output,'findmultiplier');
  if debugging then writeln(output, 'findmultiplier: x = ',x:10:5);
  done := false;
  while not done do begin
     if debugging then writeln(output, 'findmultiplier: multiplier = ',
        multiplier:1);
     if debugging then writeln(output, 'round(multiplier*x) = ',
        round(multiplier*x):1);
     if debugging then writeln(output, 'trunc(multiplier*x) = ',
        trunc(multiplier*x):1);
     if round(multiplier*x) = trunc(multiplier*x)
     then if multiplier*x > trunc(multiplier*x) then begin
             if debugging then writeln(output,'keep on going!');
             multiplier := multiplier * 10;
          end
          else begin
              done := true
          end
     else multiplier := multiplier * 10;
     if multiplier > multmax then done := true
  end;
end; (* findmultiplier *)
(* end module findmultiplier *)

(* begin module freb.themain *)
procedure themain(var frebp, book: text);
(* the main procedure of the program *)
var
   na,nc,ng,nt: integer; (* numbers of bases *)
   naReal,ncReal,ngReal,ntReal: real; (* real numbers of bases *)
   apiece: pieceptr;
   chr: chrkey; (* chromosome info *)
   chropen, orgopen: boolean; (* open or closed data types *)
   DNAstore: integer; (* given coordinate as stored in DNA data structure *)
   dummycounter: integer; (* ignore the counts given *)
   dummycounterprevious: integer; (* previous dummycounter *)
   gotten: boolean; (* got the name? *)
   i: integer; (* index to note numbers *)
   l: integer; (* index to storage *)
   lstore: integer; (* coordinate for storage *)
   kill: boolean; (* kill the book: bad table *)
   n: integer; (* count of the sequences produced *)
   thename: string; (* the name of the matrix *)
   maxlength: integer; (* longest sequence length coordinate *)
   maxtotal: integer; (* the highest total value of a,c,g,t *)
   multiplier: integer; (* number to multiply input numbers by *)
   previousmultiplier: integer; (* previous multiplier *)
   newtotal: integer; (* the current total value of a,c,g,t *)
   org: orgkey; (* organism info *)
   q: integer; (* position across the table *)
   spat: boolean; (* have we spat out a base? *)
   store: array[1..maxarray,0..4] of integer; (* for storing the table *)
   nl: array[1..maxarray] of integer; (* number of bases at position l *)
   nowfrom: integer; (* range from, decreasing from thefrom *)
   nowto: integer; (* range to, decreasing from theto *)
   requestedfrom: integer; (* from value requested in file *)
   thedate: datetimearray; (* now *)
   thefrom: integer; (* range from *)
   theto: integer; (* range to *)
   up: boolean; (* the n(l) is going up *)
   zerobase: integer; (* the zero coordinate in the range 1..n *)

begin
   writeln(output,'freb ',version:4:2);

   reset(frebp);

   getstring(frebp, thename, gotten);
   if not gotten then begin
      writeln(output,'empty frebp file');
      halt;
   end;

   writestring(output,thename);
   writeln(output);

   up := true;
   kill := false;
   maxlength := 0;
   maxtotal := 0;
   dummycounter := 1;
   dummycounterprevious := -maxint;
   previousmultiplier := -maxint;
   writeln(output,'  l  n(l)');
   while not eof(frebp) do begin
      readln(frebp,dummycounter,naReal,ncReal,ngReal,ntReal);
{
      write  (output,dummycounter:5,
                 ' ',naReal:5:1,
                 ' ',ncReal:5:1,
                 ' ',ngReal:5:1,
                 ' ',ntReal:5:1);
}

      if dummycounterprevious = -maxint then begin
         (* allow the initial counter to be anything *)
         dummycounterprevious := dummycounter -1;
         requestedfrom := dummycounter;
         l := requestedfrom;
      end;
      if dummycounter <> dummycounterprevious + 1 then begin
         writeln(output,' WARNING: skipped numbers in original numbering!');
         (* If we skipped a position, add one to the requested from.
            so -3 -2 -1 1 2 3 is set to -2 -1 0 1 2 3 *)
         requestedfrom := requestedfrom
                        + dummycounter - dummycounterprevious - 1;
      end;
      dummycounterprevious := dummycounter;
      l := l + 1;  (* l just keeps marching along *)

      multiplier := 1;
      findmultiplier(naReal, multiplier);
      findmultiplier(ncReal, multiplier);
      findmultiplier(ngReal, multiplier);
      findmultiplier(ntReal, multiplier);

      if multiplier > maxmultiplier then begin
         multiplier := maxmultiplier;
      end;

      if previousmultiplier = -maxint then begin
{
         writeln(output, ' using multiplier ', multiplier:1);
}
         previousmultiplier := multiplier;
      end
      else begin
         if previousmultiplier <> multiplier then begin
{
            write  (output, ' multiplier changed to ', multiplier:1);
            writeln(output, ' - using ', previousmultiplier:1);
}
            multiplier := previousmultiplier;
         end
         else begin
{
            write  (output, '-');
}
         end;
      end;

      na := round(multiplier * naReal);
      nc := round(multiplier * ncReal);
      ng := round(multiplier * ngReal);
      nt := round(multiplier * ntReal);

      maxlength := maxlength + 1;
      if maxlength > maxarray then begin
         writeln(output,'the table is too wide, increase constant maxarry');
         halt
      end;
{
      store[maxlength,0] := maxlength;
}
      store[maxlength,0] := l;
      store[maxlength,1] := na;
      store[maxlength,2] := nc;
      store[maxlength,3] := ng;
      store[maxlength,4] := nt;
      newtotal := na+nc+ng+nt;
{
      writeln(output,maxlength:3,' ',newtotal:5);
}
      writeln(output,l:3,' ',newtotal:5);
      nl[maxlength] := newtotal;
      if newtotal > maxtotal then begin
         maxtotal := newtotal;
         zerobase := maxlength;
         if not up then begin
(* this mechanism can be dropped probably because rounded
numbers can be smaller *)
            writeln(output, 'n(l) goes up then down then up again!');
            up := true;
            kill := true
         end;
      end
      else begin
         if newtotal < maxtotal then up := not up;
      end
   end;
{
   thefrom := 1 - zerobase;
   theto := maxlength - zerobase;
}
   thefrom := requestedfrom;
   theto := maxlength + requestedfrom - 1;

   writeln(output,'total number of sequences to make: ',maxtotal:1);
   writeln(output,'maximum length of sequences to make: ',maxlength:1);
   writeln(output,'from ',thefrom:1,' to ',theto:1);

   (* begin to write the book *)
   rewrite(book);
   getdatetime(thedate);
   bwstartline(book);
   writedatetime(book,thedate);
   write(book,', ');
   writedatetime(book,thedate);
   write(book,', ');
   writestring(book,thename);
   writeln(book);

   with org do begin
      fillhead(thename,hea);
      new(mapunit);
      fillline(thename,mapunit);
   end;
   with chr do begin
      fillhead(thename,hea);
      mapbeg := thefrom;
      mapend := theto;
   end;

   chropen := false; orgopen := false;

   bworg(book, org, chropen, orgopen);
   bwchr(book, chr, chropen);

   new(apiece);
   with apiece^.key do begin
      fillhead(thename,hea);
      new(hea.note);
      with hea.note^ do begin
         for i := 1 to linelength do letters[i] := ' ';
         length := 5;
         letters[1] := '#';
         next := nil
      end;
      mapbeg := 0;
      coocon := linear;
      coodir := plus;
      coobeg := thefrom;
      cooend := theto;
      piecon := linear;
      piedir := plus;
      piebeg := thefrom;
      pieend := theto;
   end;
   with apiece^ do begin
      new(dna);
      with dna^ do begin
         length := maxlength;
         next := nil
      end;
   end;

   (* for testing, output should be similar to frebp: *)
   for l := 1 to maxlength do begin
     for q := 0 to 4 do begin
        write(output,' ',store[l,q]:4);
        if q = 0 then write(output,' | ');
     end;
     writeln(output);
   end;

   nowfrom := thefrom;
   nowto   := theto;
   for n := 1 to maxtotal do begin (* create the total number of sequences *)

{
     (* determine current range *)

     while nl[nowfrom-thefrom+1] <= 0 do begin
        nowfrom := nowfrom + 1;
        if nowfrom > nowto then begin
           writeln(output,'program error, nowfrom > nowto');
           writeln(output,'nowfrom = ',nowfrom:1);
           writeln(output,'nowto   = ',nowto  :1);
           halt;
        end;
     end;

     while nl[nowto-theto+1] <= 0 do begin
        nowto := nowto - 1;
        if nowfrom > nowto then begin
           writeln(output,'program error, nowfrom > nowto');
           writeln(output,'nowfrom = ',nowfrom:1);
           writeln(output,'nowto   = ',nowto  :1);
           halt;
        end;
     end;

     apiece^.key.piebeg := nowfrom;
     apiece^.key.pieend := nowto;
     apiece^.dna^.length := nowto - nowfrom + 1;
}

     nowfrom := thefrom;
     nowto   := theto  ;
     apiece^.key.piebeg := nowfrom;
     apiece^.key.pieend := nowto;
     apiece^.dna^.length := nowto - nowfrom + 1;

     lstore := nowfrom-thefrom;
     for l := nowfrom to nowto do begin (* create one sequence *)
        DNAstore := l - nowfrom;
        lstore := lstore + 1;
        q := 1;
        spat := false;
        while not spat do begin
           if store[lstore,q] > 0
           then begin
              case q of
                 1: apiece^.dna^.part[DNAstore+1] := a;
                 2: apiece^.dna^.part[DNAstore+1] := c;
                 3: apiece^.dna^.part[DNAstore+1] := g;
                 4: apiece^.dna^.part[DNAstore+1] := t;
              end;
              store[lstore,q] := store[lstore,q] - 1;
              nl[lstore] := nl[lstore] - 1;
              spat := true
           end
           else begin
              q := q + 1;
              if q > 4 then q := 1;
              if nl[lstore] <= 0 then spat := true;
           end
        end;
     end;
     mknumber(n,apiece^.key.hea.note);
     bwpie(book, apiece);
   end;

   writeln(book, 'chromosome');
   writeln(book, 'organism');

   if kill then begin
      writeln(output,'WARNING: up and down of sums!');
{
      writeln(output,'BOOK IS KILLED because of up and down');
      rewrite(book);
      writeln(book,'halt: bad book');
}
   end;
end;
(* end module freb.themain *)

begin
   themain(frebp, book);
1: end.





























