program markov(cmp,mkvseqs,listing,markovp,output);
(* markov: random generation of DNA sequences from a composition
            based on probability of occurance of previous oligomer.
   by John Eberwein, Gary Stormo, Tom Schneider

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  https://alum.mit.edu/www/toms (permanent)

module libraries needed: delman, delmods, auxmods  *)

label 1;  (* end of program *)

const
(* begin module version *)
   version = 3.77; (* of markov.p 2012 May 31
2012 May 31, 3.77: improve documentation
2004 Sep  9, 3.76: GPC upgrade
1999 dec 13, 3.75: y2k upgrade
1994 sep 5: previous changes
date of origin: feb 14,1984 *)
(* end module version *)

(* begin module describe.markov *)
(*
name
      markov: markov chain generation of a DNA sequence from composition.

synopsis
      markov(cmp: in, mkvseqs: out, listing: out, markovp: in, output: out)

files
      cmp: the input composition, which is the output of program comp.
           If the file is empty, equiprobable mononucleotide
           composition is used.

      mkvseqs: the output dna sequences of this program.

      listing: contains the following information about program execution:
         program and version number.
         first three lines of the input composition file used to
            generate the sequences with.
         the four input parameters - number of sequences, length of sequences,
            the seed number, and the depth of sequence generation.
         a listing of any sequence that could not be generated from the
            prior length specified in the markovp file.
            the listing contains the sequence number, the sequence position
            and the depth of restart of sequence generation.

      markovp: for parameters; markovp must contain four numbers
         each on separate lines:
         1. number of sequences desired - an integer
         2. the length of each sequence - an integer
         3. a seed number between zero and one or outside this range
         if a computer date and time seed is desired.  the seed
         is used to start the random number generator. - a real
         4. the number of bases prior to the one about to be inserted
         which are to influence the choice of the base to be inserted.
         zero means equiprobable random sequences are desired.
         example:   20 number of sequences desired
                    100 length of sequences desired
                    2.0 for a computer date and time seed
                    3 composition depth used to generate next base

      output: for user messages.

description

      Markov generates a set of random dna sequences which have
      approximately the same composition as the one in the composition
      file supplied to the program.  The user chooses the depth of the
      composition to be used.  For example, if trinucleotides
      (composition depth = 3) are used, the previous two bases
      determine the probability of the next base in the sequence. 
      This is called a markov chain.

      Sometimes the program will work itself into a corner, when no
      composition exists for the previous few bases.  In these cases,
      the program restarts with the longest possible oligonucleotide
      that does exist in the composition.  These cases are recorded in
      the listing.

see also
      comp.p, compan.p, rndseq.p

author
      John Eberwein, Gary Stormo, Tom Schneider

bugs
      none known
*)
(* end module describe.markov *)

(* begin module markov.contents *)
(* the following shows the relationships of the program modules
                               **basetochar
                               **thefirstbackup
                               **makespiders
                               **resetspiders
               **makesequences***rtrv
               *               **backup***rtrv
               *               **advance
               *               **slither
               *               **makeabase***random
               *
               *       **halt
               **header**copylines**copyaline
markov***main***
               *                          **halt
               *              **timeseed****getdatetime
               **mkvgetparams***halt
               *
               *                     **halt
               *                     **skipoligo
               **getacomp**readcomp****getcolon
                                     **clearitem
                                     **getitem  *)
(* end module markov.contents *)

(* begin module markov.const *)
      endsymbol = '.'; (* the symbol to end each sequence *)
      linelength = 60; (* maximum number of bases written to one line. *)
      namelength = 20;  (* for timeseed *)
(* end module markov.const *)

(* begin module datetime.const *)

   datetimearraylength = 19; (* length of dataarray for dates,
       It is just long enough to include the 4 digit year - solving the
       year 2000 problem:
1980/06/09 18:49:11
123456789 123456789 
         1         2
*)

(* end module datetime.const version = 7.67; {of delmod.p 2004 Sep 8} *)

type

(* begin module datetime.type *)
   (* array for dates *)
   datetimearray = packed array[1..datetimearraylength] of char;
(* end module datetime.type version = 7.67; {of delmod.p 2004 Sep 8} *)

(* begin module markov.type *)
   base = (a,c,g,t);

(* for timeseed *)
   alpha=array[1..namelength] of char;
(* end module markov.type *)

(* begin module comp.type *)
            (* the composition is stored in a tree of these nodes *)
      compnodeptr = ^compnode; (* points to a node of the tree *)
      compnode = record    (* a node of the composition tree *)
         count : integer; (* the number of oligos for this node *)
         son : array[base] of compnodeptr; (* the pointers to 'descendants'
                                           of this node of the tree *)
      end;

            (* spiders are used to make the composition tree *)
      spiderptr = ^spider; (* points to a 'spider' *)
      spider = record  (* a spider climbs the composition tree, its path
                        determined by the sequences, and increments 'count' at
                        all the nodes it passes, thereby determining the
                        composition *)
         depth : integer; (* the level of the node now at *)
         place : compnodeptr; (* a pointer to the current node *)
         next : spiderptr; (* the next spider in the collection *)
      end;

            (* the total number of composition entries at a given level
               is stored in the linked list of type comptotal *)
      compptr = ^comptotal;
      comptotal = record
         count: integer;  (* the number at a given level *)
         next: compptr;  (* pointer to the next level totals *)
      end;

           (* 'path' is used in printing the tree *)
      pathptr = ^path; (* pointer into a path on the tree *)
      path = record  (* the path of bases to get to a particular node *)
         bas : base;
         next : pathptr;
      end;
(* end module comp.type version = 'auxmod 1.37 85 apr 4 gds/tds'; *)

var   cmp,mkvseqs,listing,markovp : text;

(* begin module package.primitive *)
(* ************************************************************************ *)
(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 7.67; {of delmod.p 2004 Sep 8} *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 7.67; {of delmod.p 2004 Sep 8} *)

(* begin module copylines *)
function copylines(var fin, fout: text; n: integer): integer;
(* copy n lines of file fin to file fout.
the actual number of lines copied is returned. *)
var
      index: integer; (* the current line number *)
begin (* copylines *)
      index := 0;
      while (not eof(fin)) and (index < n) do begin
         copyaline(fin, fout);
         index := succ(index)
      end;

      copylines := index
end; (* copylines *)
(* end module copylines version = 7.67; {of delmod.p 2004 Sep 8} *)

(* ************************************************************************ *)
(* end module package.primitive version = 7.67; {of delmod.p 2004 Sep 8} *)

(* begin module basetochar *)
function basetochar(ba:base):char;
(* convert a base into a character *)
begin
      case ba of
         a: basetochar:='a';
         c: basetochar:='c';
         g: basetochar:='g';
         t: basetochar:='t';
      end
end;
(* end module basetochar version = 'auxmod 1.37 85 apr 4 gds/tds'; *)

(* begin module comp.readcomp *)

(* begin module skipoligo *)
procedure skipoligo(var thefile: text);
(* this procedure is used to skip over an oligonucleotide string.
   it reads until it comes to a base, and then continues to read until
   it comes to a blank. no checking is done for non-base characters
   in between. *)
var ch : char;
begin
      repeat read(thefile,ch)
      until (ch in ['a','c','g','t']);
      repeat read(thefile,ch)
      until (ch = ' ');
end;
(* end module skipoligo version = 'auxmod 1.37 85 apr 4 gds/tds'; *)

procedure readcomp(var comp: text; var compmax: integer; readmax: integer;
                   var root: compnodeptr; var monocomptotal: compptr);
(* this procedure requires modules: comp.type, skipoligo, halt;
   this procedure reads from file 'comp' a composition and puts it into
   the tree pointed to by the 'root' pointer.  'compmax' is the depth
   of the composition tree which is stored.  it is the minimum of
   'readmax', the requested depth, and 'detcomp', the depth for which
   the input file composition was determined.  'monocomptotal' points
   to the beginning (for the monos) of a linked list which gives the
   totals for each level of the composition. *)
type  (* for efficient reading  of the information into the tree, each
         node that has a successor is stored in a queue as well as put
         into the tree.  the variables of the type list are the queue
         elements. *)
    listptr = ^list;
    list = record
      item: compnodeptr;  (* points to the composition tree node *)
      next: listptr;  (* points to the next list item in the queue *)
    end;

var  freeitem,  (* the list of unused list pointers *)
     listitem, (* an item in the queue *)
     first,  (* the first item in the queue *)
     last: listptr;  (* the last item in the queue *)
     comptot,  (* the comp total for a given level *)
     newcomptot: compptr;  (* for adding to the string of comptot"s *)
     detcomp,  (* the level to which the input composition was determined *)
     level,  (* of the composition being read, i.e., monos, dis, ... *)
     number: integer;  (* read from the 'comp' file *)
     ch: char;  (* for reading from 'comp' *)
     ba: base;  (* an index *)

(* getitem and clearitem provide efficient use of linked list storage by
   keeping a list of unused pointers that can be allocated instead of
   always creating 'new' ones  *)
procedure getitem(var l: listptr);
(* obtain a listitem from the free list or by making a new one *)
begin
      if freeitem<>nil
         then begin
            l:=freeitem;
            freeitem:=freeitem^.next
         end
         else new(l);
      l^.next:=nil
end;


procedure clearitem(var l: listptr);
(* return a listitem to the free list *)
var   lptr: listptr;
begin
      if l<>nil then begin
         lptr:=l;
         l:=l^.next;
         lptr^.next:=freeitem;
         freeitem:=lptr
      end
end;

begin  (* readcomp *)
      reset(comp);
      if eof(comp) then begin
         writeln(output,' error: no composition file provided');
         halt;
      end;
      readln(comp); (* skip the program identification *)
      readln(comp); (* skip the book identification *)
      readln(comp,detcomp); (* obtain the determined composition *)
      readln(comp); (* skip the blank line *)

            (* determine the level of composition to be stored *)
      if (readmax < 1) then begin
         writeln(output);
         writeln(output,' warning: 0 or negative oligo length requested');
         writeln(output,' composition used is depth ',detcomp:1);
         writeln(output);
         compmax := detcomp
      end
      else if (readmax > detcomp) then begin
         writeln(output);
         writeln(output,' warning: requested composition oligo',
                        ' length (',readmax:1,')');
         writeln(output,' is larger than the determined composition',
                        ' oligo length (',detcomp:1,').');
         writeln(output,' composition used is to depth ',detcomp:1);
         writeln(output);
         compmax := detcomp
      end
      else compmax := readmax;

      new(root);
      new(monocomptotal);
      new(comptot);
      new(first);
      new(last);
      first^.item := root;
      first^.next := nil;
      last := first;
      new(freeitem);
      freeitem^.next := nil;

         (* read in the total number of bases from the composition *)
      readln(comp); (* skip the * *)
      readln(comp); (* skip the information line *)
      readln(comp,number);
      root^.count := number;

      repeat
         read(comp,ch); (* skip the space *)
         read(comp,ch); (* the ch determines what to do next *)
         if (ch = '*') then begin  (* determine the level about to be read *)
            readln(comp);
            readln(comp,level);
            if (level = 1) then monocomptotal := comptot
            else begin
               new(newcomptot);
               comptot^.next := newcomptot;
               comptot := newcomptot;
            end;
            comptot^.count := 0;
            comptot^.next := nil
         end
         else begin  (* read the composition values *)
            for ba := a to t do begin
               skipoligo(comp);
               read(comp,number);
               if (number <> 0) then begin
                  new(first^.item^.son[ba]);
                  first^.item^.son[ba]^.count := number;
                  comptot^.count := comptot^.count + number;
                  getitem(listitem);
                  last^.next := listitem;
                  last := listitem;
                  last^.next := nil;
                  last^.item := first^.item^.son[ba]
               end
               else first^.item^.son[ba] := nil;
            end;
            clearitem(first);
            readln(comp);
         end;
      until level = compmax;
         (* for the last level of compositions to be read we don"t
            need to store the nodes in the queue because their successors
            will not be read in. *)
      repeat
         for ba := a to t do begin
            skipoligo(comp);
            read(comp,number);
            if (number <> 0) then begin
               new(first^.item^.son[ba]);
               first^.item^.son[ba]^.count := number;
               comptot^.count := comptot^.count + number;
            end
            else first^.item^.son[ba] := nil;
         end;
         clearitem(first);
         readln(comp);
      until (first = nil);
end;  (* readcomp *)

function getcount(root: compnodeptr; start: pathptr): integer;
(* this function follows the tree from the node 'root' (which
   may be the root of a subtree) along the path initiated with
   'start', and returns the count of the resulting node.  if the
   path through the tree ever hits a null node in the process
   the count is returned to be zero. *)
var place: compnodeptr;  (* a place in the composition tree *)
    point: pathptr;  (* a point in the path through the tree *)
begin
      point:= start;
      place := root;
      while ((place <> nil) and (point <> nil)) do begin
         place := place^.son[point^.bas];
         point := point^.next;
      end;
      if (place = nil) then getcount := 0
      else getcount := place^.count;
end;

(* end module comp.readcomp version = 'auxmod 1.37 85 apr 4 gds/tds'; *)

  (***********************************)

(* begin module random *)
procedure random(var seed: real);
(* the seed value is used to generate a random number in the range 0 to 1.
the seed is then altered, to provide the seed of the next random number.
in other words, the seed is the random number.  to obtain the next random
number in the pseudo-random series, simply call random again with the same
seed.  note that the series will not be truely random, since the previous
value determines the next value.  this is "pseudo-random".

an initial seed of 4.0 is suggested.  note that this is not the first
random number.  you can use different seeds to get different pseudo-random
number series.

the expected mean and standard deviation of a flat distribution is
calculated using simple integrals:

       1
mean = / x dx = 0.5
       0
           1           2
variance = / (x - mean)  dx = 1/12;  stdev = sqrt(variance) = .2886751
           0

(the integral for the variance is solved easily by substituting y = x - mean.)

   the algorithm is derived from:
pascal programs for scientists and engineers, alan r. miller
sybex, berkeley ca (1981).
which in turn was an adaptation from hp-35 applications programs *)
const
      pi = 3.141592653589; (* a value used to help mix up the result *)
var
      intermediate: real; (* an intermediate result *)
begin (* random *)
      intermediate := seed + pi;
      intermediate := exp(5.0*ln(intermediate));
      seed := intermediate - trunc(intermediate)
end; (* random *)
(* end module random version = 1.06; (@ of rndseq 1983 oct 17 *)

(* begin module getdatetime *)
procedure getdatetime(var adatetime: datetimearray);
(* get the date and time into a single array from the system clock.
   adatetime contains the date:
       1980/06/09 18:49:11
         ye mo da ho mi se
(year, month, day, hour, minute, second).
As of 2000 February 18, the Sun Pascal compiler requires a formatting
statement.  This statement allows the date to be generated in this
standard Delila format in a single call.  Information about the
formatting statement is available on the manual page for date in Unix.
If a computer does not have this method, see the 'oldgetdatetime' routine
in delmod.p (https://alum.mit.edu/www/toms/delila/delmod.html)
for some conversion code.

GPC Functions:
function  GetUnixTime (var MicroSecond : Integer) : UnixTimeType;

http://agnes.dida.physik.uni-essen.de/~gnu-pascal/gpc_109.html#SEC109

7.10.8 Date And Time Routines 

procedure GetTimeStamp (var t : TimeStamp); 
function Date (t : TimeStamp) : packed array [1 .. DateLength] of Char; 
function Time (t : TimeStamp) : packed array [1 .. TimeLength] of Char; 

DateLength and TimeLength are implementation dependent constants. 

GetTimeStamp (t) fills the record `t' with values. If they are valid, the Boolean
flags are set to True. 

TimeStamp is a predefined type in the Extended Pascal standard. It may be
extended in an implementation, and is indeed extended in GPC. For the full
definition of `TimeStamp', see section 8.255 TimeStamp. 

*)
var
    t: TimeStamp; 

(* begin module pluckdigit *)
function pluckdigit(number, logplace:integer): char;
(* return the digit at the place value ('logplace') position of number.
example:
      pluckdigit(13625, 3) = 3
      pluckdigit(13625, 4) = 1
This routine was taken from module numberdigit in prgmod.p, but is
modified so as not to give the sign.  Instead it gives zeros
above the digits.  'myabsolute' replaced 'absolute', which
is apparently a keyword for GPC.  The name is kept for
to keep the code looking similar to its origin.
*)
var
      place: integer; (* the exponent of logplace *)
      count: integer; (* used to make place *)
      myabsolute: integer; (* the absolute value of number *)
      acharacter: char; (* the character to be returned *)
procedure digit;
(* extract a digit at the place position *)
var
      tenplace: integer; (* ten times place *)
      z: integer; (* an intermediate value *)
      d: integer; (* the digit extracted *)
begin (* digit *)
      tenplace:=10*place;
      z:=myabsolute-((myabsolute div tenplace)*tenplace);
      if place = 1
         then d:=z
         else d:= z div place;
      case d of
         0: acharacter:='0';
         1: acharacter:='1';
         2: acharacter:='2';
         3: acharacter:='3';
         4: acharacter:='4';
         5: acharacter:='5';
         6: acharacter:='6';
         7: acharacter:='7';
         8: acharacter:='8';
         9: acharacter:='9';
      end
end; (* digit *)
begin (* pluckdigit *)
      place:=1;
      for count:=1 to logplace do place:=10*place;
      if number=0 then begin
         acharacter:='0'
      end
      else begin
         myabsolute:=number;
         if myabsolute >= place
         then digit
         else acharacter := '0'
      end;
      pluckdigit:=acharacter
end; (* pluckdigit *)
(* end module pluckdigit *)

begin

 (* according to:
http://agnes.dida.physik.uni-essen.de/~gnu-pascal/gpc_109.html#SEC109
*)
   GetTimeStamp(t); 

(* Predefined time stamp:
http://agnes.dida.physik.uni-essen.de/~gnu-pascal/gpc_389.html#SEC389
  TimeStamp = {@@packed} record
      DateValid,
      TimeValid   : Boolean;
      Year        : Integer;
      Month       : 1 .. 12;
      Day         : 1 .. 31;
      DayOfWeek   : 0 .. 6;  { 0 means Sunday }
      Hour        : 0 .. 23;
      Minute      : 0 .. 59;
      Second      : 0 .. 61; { to allow for leap seconds }
      MicroSecond : 0 .. 999999
    end;
*)

  with t do begin
    if TimeValid then begin
{
       writeln(output,'valid time');
       writeln(output,'year   =',year:4);
       writeln(output,'month  =',month:2);
       writeln(output,'day    =',day:2);
       writeln(output,'hour   =',hour:2);
       writeln(output,'minute =',minute:2);
       writeln(output,'second =',second:2);
}
       adatetime := 'year/mm/dd hh:mm:ss';
       adatetime[ 1] := pluckdigit(Year,3);
       adatetime[ 2] := pluckdigit(Year,2);
       adatetime[ 3] := pluckdigit(Year,1);
       adatetime[ 4] := pluckdigit(Year,0);
       adatetime[ 6] := pluckdigit(Month,1);
       adatetime[ 7] := pluckdigit(Month,0);
       adatetime[ 9] := pluckdigit(Day,1);
       adatetime[10] := pluckdigit(Day,0);
       adatetime[12] := pluckdigit(Hour,1);
       adatetime[13] := pluckdigit(Hour,0);
       adatetime[15] := pluckdigit(Minute,1);
       adatetime[16] := pluckdigit(Minute,0);
       adatetime[18] := pluckdigit(Second,1);
       adatetime[19] := pluckdigit(Second,0);
    end
    else begin
       writeln(output,'getdatetime: invalid time!');
       halt;
    end
  end;
{
   Sun compiler method:
   date(adatetime,'%Y/%m/%d %H:%M:%S');
}
end;
(* end module getdatetime version = 7.67; {of delmod.p 2004 Sep 8} *)

(* begin module timeseed *)
(* Read the computer date and time.  Reverse the order of the digits and put
a decimal point in front.  This gives a fraction between zero and one that
varies quite quickly, and is always unique (if the computer has sufficient
accuracy).  It is to be used as a seed to a random number generator.  This
has the nice property that the seed changes every second and does not repeat
for thousands of years!  *)
procedure addtoseed(var seed, power: real; c: char);
(* add the digit represented by c to the seed at the power position *)
var
      n: integer; (* the character represented by c *)
begin (* addtoseed *)
      power := power/10;
{
writeln(output,'addtoseed, c = ',c);
writeln(output,'addtoseed, ord(c) = ',ord(c));
}
   n := ord(c) - ord('0');
   if (n < 0) or (n > 9) then begin
      writeln(output,'timeseed: error in datetime');
      writeln(output,'it contains "',c,'" which is not a number.');
      writeln(output,'The getdatetime routine must be fixed.');
      halt;
   end;
   seed := seed + power*n
end; (* addtoseed *)

procedure makeseed(adatetime: datetimearray; var seed: real);
(* convert adatetime to a real number in seed, reversed order
Here is the standard adatetime format:
123456789 123456789
         1         
1980/06/09 18:49:11
*)
var
   power: real; (* a digit of the seed such as 0.01 *)
begin
   seed := 0.0;
   power := 1.0;
   addtoseed(seed, power, adatetime[19]);
   addtoseed(seed, power, adatetime[18]);
   (* : *)
   addtoseed(seed, power, adatetime[16]);
   addtoseed(seed, power, adatetime[15]);
   (* : *)
   addtoseed(seed, power, adatetime[13]);
   addtoseed(seed, power, adatetime[12]);
   (*   *)
   addtoseed(seed, power, adatetime[10]);
   addtoseed(seed, power, adatetime[ 9]);
   (* / *)
   addtoseed(seed, power, adatetime[ 7]);
   addtoseed(seed, power, adatetime[ 6]);
   (* / *)
   addtoseed(seed, power, adatetime[ 4]);
   addtoseed(seed, power, adatetime[ 3]);
   addtoseed(seed, power, adatetime[ 2]);
   addtoseed(seed, power, adatetime[ 1]);
end;

procedure orderseed(adatetime: datetimearray; var seed: real);
(* convert adatetime to a real number in seed, normal order *)
var
   power: real; (* a digit of the seed such as 0.01 *)
begin
   seed := 0.0;
   power := 1.0;
   addtoseed(seed, power, adatetime[ 3]);
   addtoseed(seed, power, adatetime[ 4]);
   addtoseed(seed, power, adatetime[ 6]);
   addtoseed(seed, power, adatetime[ 7]);
   (* / *)
   addtoseed(seed, power, adatetime[ 9]);
   addtoseed(seed, power, adatetime[10]);
   (* / *)
   addtoseed(seed, power, adatetime[12]);
   addtoseed(seed, power, adatetime[13]);
   (*   *)
   addtoseed(seed, power, adatetime[15]);
   addtoseed(seed, power, adatetime[16]);
   (* : *)
   addtoseed(seed, power, adatetime[18]);
   addtoseed(seed, power, adatetime[19]);
end;

procedure timeseed(var seed: real);
(* read the computer date and time.  reverse the order of the digits
and put a decimal point in front.  this gives a fraction between
zero and one that varies quite quickly, and is always unique (if the
computer has sufficient accuracy).  it is to be used as a seed to
a random number generator. *)
var
   adatetime: datetimearray; (* a date and time *)
begin (* timeseed *)
   getdatetime(adatetime);
{
writeln(output,'timeseed: adatetime: ',adatetime);
}
   makeseed(adatetime, seed);
end; (* timeseed *)
(* end module timeseed version = 7.67; {of delmod.p 2004 Sep 8} *)

(********************************************)

(* begin module markov.thefirstbackup *)
procedure thefirstbackup(var listing:text);
  (* writes a heading line to the listing file only in the event a deadnode
     was reached during spider advancing.  also, writes a message to the
     terminal indicating there were dead nodes reached. *)

begin
   writeln(listing);
   writeln(listing,' listing of bases not produced from compmax level:');
   writeln(listing,'     sequence   position   depth of restart');
   writeln(listing);
   writeln(output,' see ''listing'' file for a list of backups')
end;  (* procedure thefirstbackup *)
(* end module markov.thefirstbackup *)

(* begin module markov.makespiders *)
procedure makespiders(var firstspider:spiderptr;var root:compnodeptr;
                          compmax:integer);
(* make as many spiders as there are levels to the composition tree,
which is 'compmax', and leave firstspider pointing to the beginning
of the list *)
var
   i: integer;   (* an index of the 'for' loop *)
   aspider : spiderptr; (* current spider being used. *)
begin
   new(firstspider);
   firstspider^.depth :=0;
   firstspider^.place := root;
   firstspider^.next := nil;
   if compmax > 1 then
      for i:=1 to (compmax-1) do begin
         new(aspider);
         aspider^.next :=firstspider;
         firstspider:=aspider
      end;  (* for loop *)
end; (* procedure makespiders *)
(* end module markov.makespiders *)

(* begin module markov.resetspiders *)
procedure resetspiders(var firstspider:spiderptr;var root:compnodeptr;
                           compmax:integer);
(* restarts all spiders at the root to start the generation of the next
markov sequence. *)

var
   active : spiderptr;  (* current spider being used *)
   i      : integer; (* loop index *)

begin
   active:=firstspider;
   for i:=1 to compmax do begin
      active^.depth:=0;
      active^.place:=root;
      active:=active^.next;
   end; (* for loop *)
end;  (* procedure resetspiders *)
(* end module markov.resetspiders *)

(* begin module markov.rtrv *)
function rtrv(var anode:compnodeptr; ba:base):integer;
(* retrieves the next count down the ba branch of the tree.  if the ba
branch is a nil pointer then a zero is returned. *)

begin
   if anode^.son[ba]=nil
      then rtrv:=0
      else rtrv:= anode^.son[ba]^.count;
end;  (* function rtrv *)
(* end module markov.rtrv *)

 (* the following procedure, backup, is a data exception
    checking procedure that is rarely executed but will keep the program
    from crashing if either of two rare conditions occured in the sequence(s)
    that were used to make the composition file.  these conditions can
    occur as follows:
       if a unique oligomer occured at the end of the sequence(s) of compmax-1
    size, this will result in at least one unbranchable node occuring in the
    tree.  this situation can cause procedure backup to be executed.
        if the unique sequence was a poly sequence and the sequence is unique
    at the compmax-2 level, this will result in only one path to a deadnode.
    when backup occurs from this node, the firstspider will occupy the node
    above the deadnode with only one path available which is to the deadnode.
    this would result in a see-saw situation on the tree and result in the
    creation of nothing but a poly sequence for the rest of the length of
    the sequence.  this condition is checked for in the last half of
    procedure backup and trips a boolean which will cause backup to
    occur again. *)

(* begin module markov.backup *)
procedure backup(var firstspider:spiderptr);
  (* this procedure invoked if there is no path firstspider can advance
     or slither to.  procedure itself moves each spider up the tree one
     node. *)

var
   aspider : spiderptr; (* current spider being backed up. *)
   prevplace : compnodeptr; (*previous node firstspider occupied before backup*)
   zeroes : integer; (* number of zero counts retrieved from next lower nodes.*)
   ba,  (* loop index *)
   temp   : base;  (* holds only path from node if 3 zeroes were retrieved. *)
   onepath : boolean; (* used to test if only onepath back to the node you
                          just came from *)

begin
   repeat
      aspider:=firstspider;
      prevplace:=firstspider^.place;
      while (aspider^.depth > 0) do begin
         aspider^.place:=aspider^.next^.place;
         aspider^.depth:=pred(aspider^.depth);
         aspider:=aspider^.next
      end;  (* while loop *)
      onepath:=false;
      zeroes:=0;
      for ba:=a to t do
         if rtrv(firstspider^.place,ba)=0 then
            zeroes:=zeroes+1
         else
            temp:=ba;
      if (zeroes=3) and (firstspider^.place^.son[temp]=prevplace) then
         onepath:=true;
   until onepath = false;
end;  (* procedure backup *)
(* end module markov.backup *)

(* begin module markov.advance *)
procedure advance(var firstspider:spiderptr;ba:base);
(* the spider s is advanced down the ba branch of the tree.  the depth
of the spider is incremented by 1. *)

var
   aspider : spiderptr; (* current spider being advanced down the tree. *)

begin
   aspider:=firstspider;
   while (aspider^.depth > 0) do begin
      aspider^.place:=aspider^.place^.son[ba];
      aspider^.depth:=succ(aspider^.depth);
      aspider:=aspider^.next
   end;  (* while loop *)
   aspider^.place:=aspider^.place^.son[ba];
   aspider^.depth:=succ(aspider^.depth)
end;  (* procedure advance *)
(* end module markov.advance *)

(* begin module markov.slither *)
procedure slither(var firstspider:spiderptr;ba:base);
 (* moves all the spiders sideways on the tree since the firstspider is at
    the compmax minus one level.  each spider is moved to the ba branch
    from the node the next spider above occupies. *)

var
   aspider : spiderptr; (* current spider being slithered. *)

begin
   aspider:=firstspider;
   while (aspider^.depth > 0) do begin
      aspider^.place:=aspider^.next^.place^.son[ba];
      aspider:=aspider^.next
   end;  (* while loop *)
end;  (* procedure slither *)
(* end module markov.slither *)

(* begin module markov.makeabase *)
function makeabase(n,ai,ci,gi,ti:integer;var seed:real):base;
(* sets boundaries between zero and one for each of the four bases.
   n is the total of all four counts retrieved from each node.
   a random number is fetched and a base is returned depending on which
   interval the random number falls in. *)

var
   sa,sac,sacg:real;   (* boudnaries between 0 and 1 that determine
                           which base will be generated. *)
begin
   sa:=ai/n;
   sac:=(ai+ci)/n;
   sacg:=(ai+ci+gi)/n;
   random(seed);
   if seed <=sa then
      makeabase:=a
   else
      if seed <=sac then
         makeabase:=c
      else
         if seed <=sacg then
            makeabase:=g
         else
            makeabase:=t
end;  (* makeabase *)
(* end module markov.makeabase *)

(* begin module markov.makesequences *)
procedure makesequences(var mkvseqs,listing:text;var root:compnodeptr;
                        var seed:real;numberofsequences,lengthofsequences,
                            compmax:integer);
(* makes specified number of dna sequences of specified length. *)

var
   firstbackup : boolean; (* used to write message to listing file and to
                            terminal only if a dead node is encountered. *)
   firstspider : spiderptr;   (* first spider of list of spiders. *)
   ba  : base;   (* stores generated base. *)
   i,h,  (* loop indices *)
   n,   (* total number of counts retrieved from each of the four nodes. *)
   ai,ci,gi,ti : integer;  (* counts from nodes below current node. *)

begin
   firstbackup:=true;
   rewrite(mkvseqs);
   makespiders(firstspider,root,compmax);

   for h:=1 to numberofsequences do begin
      resetspiders(firstspider,root,compmax);
      for i:=1 to lengthofsequences do begin
         repeat
            ai:=rtrv(firstspider^.place,a);
            ci:=rtrv(firstspider^.place,c);
            gi:=rtrv(firstspider^.place,g);
            ti:=rtrv(firstspider^.place,t);
            n:=ai+ci+gi+ti;

            (* following two if statements are only executed if a deadnode
                 is reached. *)

            if n=0 then begin
               backup(firstspider);
               if firstbackup then begin
                  thefirstbackup(listing);
                  firstbackup:=false
               end;   (* if statement *)
               writeln(listing,h:10,i:12,firstspider^.depth + 1:12)
            end; (* if statement *)

         until n > 0;  (* repeat *)
         ba:=makeabase(n,ai,ci,gi,ti,seed);
         write(mkvseqs,basetochar(ba));
         if (i mod linelength) = 0 then
            writeln(mkvseqs);

         (* if compmax equals one, the firstspider never leaves the root.
          the spiders always found in the 0 to compmax -1 levels of the tree. *)
         if compmax > 1 then
            if firstspider^.depth < compmax-1 then
               advance(firstspider,ba)
            else
               slither(firstspider,ba)
      end; (* inner for loop *)
      writeln(mkvseqs,endsymbol);
   end; (* outer for loop *)
end; (* procedure makesequences *)
(* end module markov.makesequences *)

(****************************)

(* begin module markov.header *)
procedure header(var cmp,listing:text;
                       version,seed:real;
                       numberofsequences,lengthofsequences,
                       compmax:integer);
 (* writes the program name and version to the terminal and to the listing
    file.  also, writes the first three lines of the composition file to
    listing as well as the four input parameters. *)

begin
   reset(cmp);
   rewrite(listing);
   writeln(listing,' markov version ',version:4:2,
                   '; markov chain generation from composition');
   writeln(listing);
   if eof(cmp) then
      writeln(listing,' sequences generated from equiprobable mononucleotide',
                      ' composition')
   else
      if copylines(cmp,listing,3)<>3 then begin
         writeln(output,'bad composition file');
         halt
      end;
   writeln(listing);
   writeln(listing,numberofsequences:10,' sequences generated.');
   writeln(listing,lengthofsequences:10,' length of sequences.');
   writeln(listing,seed:23:12,' seed used.');
   writeln(listing,compmax:10,' depth of sequence generation.')
end; (* procedure header *)
(* end module markov.header *)

(* begin module markov.mkvgetparams *)
procedure mkvgetparams(var markovp:text;
                       var numberofsequences,lengthofsequences:integer;
                       var seed:real;
                       var readmax:integer);
(* gets the 4 parameters from an input  file-markovp.  checks for negative
    input parameters and timeseeds if seed is not between zero and one. *)

begin
   reset(markovp);
   if eof(markovp) then begin
      writeln(output,'markovp is empty');
      halt
   end; (* if statement *)
   readln(markovp,numberofsequences);
   readln(markovp,lengthofsequences);
   readln(markovp,seed);
   readln(markovp,readmax);

   if numberofsequences < 1 then begin
      writeln(output,' number of sequences must be positive');
      halt
   end;
   if lengthofsequences < 1 then begin
      writeln(output,' sequence lengths must be positive');
      halt
   end;
   if readmax < 0 then begin
      writeln(output,' composition depth must be zero or positive');
      halt
   end;

   if (seed <= 0.0) or (seed >=1.0) then timeseed(seed);
end; (* procedure mkvgetparams *)
(* end module markov.mkvgetparams *)

(* begin module markov.getacomp *)
procedure getacomp(var cmp:text;var compmax:integer;readmax:integer;
                   var root:compnodeptr;var monocomptotal:compptr);
(* if the composition file is empty then a mono tree is made with
equiprobable counts at each node. *)

var
   ba : base;

begin
   reset(cmp);
   if eof(cmp) or (readmax = 0) then begin
      if eof(cmp) then writeln(output,' composition file is empty.');
      writeln(output,' using equiprobable mononucleotide composition.');
      new(root);
      for ba:=a to t do begin
         new(root^.son[ba]);
         root^.son[ba]^.count:=1
      end; (* for loop *)
      compmax:=1
   end (* first half of if condition *)
   else
      readcomp(cmp,compmax,readmax,root,monocomptotal)
end; (* procedure getacomp *)
(* end module markov.getacomp *)

(* begin module markov.themain *)
procedure themain(var cmp,markovp,mkvseqs,listing:text);
 (* executes the  main procedures of the program. *)

var
   root: compnodeptr;  (* of the composition tree *)
   monocomptotal: compptr; (* beginning of comp totals list *)
   numberofsequences, (* number of sequences to be generated. *)
   lengthofsequences, (* length of sequences to be generated. *)
   compmax,  (* maximum length oligo for which comp was determined *)
   readmax: integer;  (* maximum amount of the comp used *)
   seed: real; (* seed for the random number generator. *)
begin
   writeln(output,'markov ',version:4:2);
   mkvgetparams(markovp,numberofsequences,lengthofsequences,seed,readmax);
   getacomp(cmp,compmax,readmax,root,monocomptotal);
   header(cmp,listing,version,seed,numberofsequences,
                                              lengthofsequences,compmax);
   makesequences(mkvseqs,listing,root,seed,numberofsequences,
                                              lengthofsequences,compmax)
end;  (* procedure themain *)
(* end module markov.themain *)

(* begin module markov.mainprogram *)
begin (*********markov********)
   themain(cmp,markovp,mkvseqs,listing);
1: end.
(* end module markov.mainprogram *)
