program modin(fin,inst,output);
(* generate modularized delila instructions for absolute sites
     John Hoffhines
     module libraries: delman, prgmods *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.45; (* of modin.p 2018 Nov 22
2018 Nov 22, 1.45: make compilable by pc/fpc/tc
1994 Sep 05, 1.44: functional
1983 Sep 22, 1.00: origin
origin 22 september 1983 *)
(* end module version *)

(* begin module describe.modin *)
(*
name
      modin: generate modularized delila instructions for absolute sites

synopsis
      modin(fin: in, inst: out, output: out)

files
      fin: sequence site positions in a special format.
      inst:  modularized delila instructions
      output: messages to the user

description
      The existence of a file containing modularized delila instructions
      allows one to pull, from the file, instructions for generating
      specific sequence sites using the module program.  For instance,
      using modin, one may make a file containing delila instructions for
      all the laci amber mutation sites, one site per module.  Then, using
      the module program, one could pull from the file instructions for
      sites a9, a16, a19, and a21 by using the module program.  This
      would be useful if one had several different sets of amber mutations
      to analyse separately.

see also
      module.p, delila.p, describe.modin.use

author
      John Hoffhines

bugs
      The program has not been used much, so its usefulness is not known.

*)
(* end module describe.modin *)
(* begin module version *) (* neutralizes the version for compiling
version = 1.43; of describe.modin.use 1993 Jan 27 *)
(* end module version *)
(* begin module describe.modin.use *)
(*
name
      modin.use: more information on using the modin program

     MODIN FIN FILE FORMAT (BNF)

<BOOK TITLE><SETS>
<SETS>::=<SET>|<SET><SETS>
<SET>::=SET <KEY> <KEY> <KEY> <KEY> <SIGNED NUMBER> <SIGNED NUMBER>
            <MODULE GROUP>
<MODULE GROUP>::=<MODULE PARAMETERS>|<MODULE GROUP> <MODULE PARAMETERS>
<MODULE PARAMETERS>::=M <NUMBER> <IDENTIFIER>|<NUMBER> <IDENTIFIER>

     NOTES ON MODIN FIN FILE FORMAT

1) Any terms undefined here are defined in LIBDEF.

2) Keys designate, respectively, the words "organism", "chromosome",
    and "p" (piece), "g" (gene), or "t" (transcript).

3) The module parameter "m" sets delila instructions direction
    to - for one module only.  Default is +.

4) Within the module parameters section, "number" is DNA base position,
   and "identifier" immediately following it is that which will be
   used in that position's module name.

5) Only one set per line is allowed, without its module group.  Module
   groups follow on subsequent lines; more than one per line is allowed,
   but they may not be truncated by the end of a line.

*)
(* end module describe.modin.use *)

(* more constants *)

      fillermax = 20; (* length of the filler array *)

(* begin module interact.const *)
      maxstring = 150; (* the maximum string *)
(* end module interact.const version = 'prgmod 3.89  82 dec  3 tds'; *)

type
(* begin module interact.type *)
      string = record (* a string of characters *)
         letters: array[1..maxstring] of char; (* the letters in the string *)
         length: integer; (* the number of characters in the string *)
         current: integer; (* the letter we are working on *)
      end;
(* end module interact.type version = 'prgmod 3.89  82 dec  3 tds'; *)

(* begin module filler.type *)
      (* the following is an array used to fill a string.
      it is convenient to have it much shorter than the maxstring, so that
      it is easy to fill the string using procedure fillstring.
      the user must declare the value of constant fillermax. *)
      filler = packed array[1..fillermax] of char;
(* end module filler.type version = 'prgmod 3.89  82 dec  3 tds'; *)

var fin, (* sequence site positions in a special format *)
    inst: (* modularized delila instructions *)
          text;
    got: (* for use with the get procedures *)
         boolean;
    buff: (* buffer for holding strings *)
           string;
    tok: (* for holding tokens *)
         string;

(* ************************************************************************ *)
(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.16 84 mar 12 tds/gds'; *)

(* begin module interact.prompt *)
procedure prompt(var afile: text);
(* prompt a file.
   the prompt is sent to the output file, and
   a line is read into the pascal line buffer.
   (for the cyber system this means to readln afile.)
   guarantee no bomb *)
begin (* prompt *)
      if eof(afile) then reset(afile);
      readln(afile);
end; (* prompt *)
(* end module interact.prompt version = 'prgmod 3.89  82 dec  3 tds'; *)

(* begin module interact.clearstring *)
procedure clearstring(var ribbon: string);
(* empty the string *)
var   index: integer; (* to the ribbon *)
begin (* clearstring *)
      with ribbon do begin
         for index := 1 to maxstring do letters[index] := ' ';
         length := 0;
         current := 0;
      end
end; (* clearstring *)
(* end module interact.clearstring version = 'prgmod 3.89  82 dec  3 tds'; *)

(* begin module interact.readchar *)
procedure readchar(var afile:text; var ch: char);
(* read a character from afile, guarantee no bomb *)
begin (* readchar *)
      if eof(afile) then prompt(afile);
      read(afile,ch);
(*writeln(output,'"',ch,'"') *)
end; (* readchar *)
(* end module interact.readchar version = 'prgmod 3.89  82 dec  3 tds'; *)

(* begin module interact.nostring *)
function nostring(var buffer: string): boolean;
(* true if there are no characters in the rest of the buffer;
   false if there are characters.
   also, if there is no buffer, then buffer.length is set to 0 *)
var   answer: boolean; (* the answer returned *)
procedure kill;
(* destroy the line *)
begin (* kill *)
      answer := true; (* blood and gore *)
      clearstring(buffer) (* total death *)
end; (* kill *)
begin (* nostring *)
      with buffer do begin
         if length > 0 then begin
            if length < maxstring
            then while (letters[current] = ' ') and (current < length)
               do current := succ(current);

            if current <= maxstring
            then if letters[current] = ' '
                 then kill
                 else answer := false
            else kill
         end
         else kill
      end;
      nostring := answer
end; (* nostring *)
(* end module interact.nostring version = 'prgmod 3.89  82 dec  3 tds'; *)

(* begin module filler.fillstring *)
procedure fillstring(var s: string; a: filler);
(* this procedure makes it reasonably easy to fill the string s with
characters.  one calls the procedure as: *)
(*                           1         2         3         4         5 *)
(*                  12345678901234567890123456789012345678901234567890 *)
(*   fillstring(s, 'this-is-the-string                                ');
the two comments make it easy to line the characters up. also, for this
example, it was assumed that the length of filler as defined by the
constant fillermax was 50. *)
var
      length: integer; (* of the string without trailing blanks *)
      index: integer; (* of s *)
begin (* fillstring *)
      clearstring(s);
      length := fillermax;
      while (length > 1) and (a[length] = ' ') do length := pred(length);
      if (length = 1) and (a[length] = ' ') then begin
         writeln(output, 'fillstring: the string is empty');
         halt
      end;

      for index := 1 to length do s.letters[index] := a[index];
      s.length := length;
      s.current := 1
end; (* fillstring *)
(* end module filler.fillstring version = 'prgmod 3.89  82 dec  3 tds'; *)

(* begin module interact.readstring *)
procedure readstring(var afile: text; var line: string);
(* read in a string from afile, protect against bombing *)
var
      index: integer; (* for line *)
      cha: char; (* a character read in *)
      done: boolean; (* used for removing trailing blanks from the line *)
      acceptable: boolean; (* was the line typed short enough? *)
begin (* readstring *)
      with line do begin
         repeat
            clearstring(line);
            prompt(afile);
            index := 0; (* we now count characters *)
            while (not eoln(afile)) and (index < maxstring) do begin
               index:=succ(index);
               readchar(afile,cha);
               line.letters[index]:=cha
            end;
            if not eoln(afile)
            then begin
               writeln(output, 'type lines shorter than ', (maxstring+1): 1,
                               ' characters.  please retype the line...');
               acceptable := false
            end
            else acceptable := true
         until acceptable;

         length := index;
         if length > 0 then begin
            done := false;
            repeat (* remove blanks from the line.  note that a while loop
                  can not be used because one must avoid letters[0], since that
                  position does not exist... *)
               if letters[length] = ' '
               then length := pred(length)
               else done := true;

               if length = 0 then done := true
            until done
         end;

         if length > 0 then current := 1
                       else current := 0
      end
end; (* readstring *)
(* end module interact.readstring version = 'prgmod 3.89  82 dec  3 tds'; *)

(* begin module interact.writestring *)
procedure writestring(var tofile: text; var s: string);
(* write the string s to file tofile, no writeln *)
var   i: integer; (* index to s *)
begin (* writestring *)
      with s do for i := 1 to length do write(tofile, letters[i])
end; (* writestring *)
(* end module interact.writestring version = 'prgmod 3.89  82 dec  3 tds'; *)

(* begin module interact.figurestring *)
procedure figurestring(
      var line: string; (* a string of characters to figure out *)
      var first: integer; (* first found non-blank character in the line *)
      var last:  integer; (* last character before a blank after first *)
      var whzat: char; (* what the token is *)
      var c:     char;  (* the first character of the token *)
      var i:     integer; (* integer value of token if it is integer; or 0 *)
      var r:     real); (* the real value if it is real; or 0.0 *)
(* figurestring figures out the tokens in a string.
 it recognizes words, integers, reals and poorly formed numbers.
 you can easily use it to parse lines.

our goal is to figure out what thing is on a string.
start looking at the current place on the line.
first and last are the first 'token' in line after start.
the current place is updated to the letter after last.
the thing found is described by the value of whzat:
      'c': character (when the token does not begin with a digit, '+', or '-')
      'i': integer
      'r': real
      ' ': blank line
      'g': garbage, cannot figure it out
and the value of the thing found is the appropriate variable *)
const
      fsversion = 2.01; (* version of figurestring *)
var
      numbers: set of '0'..'9';
      sign: integer; (* sign of a number *)
      numberstart: integer; (* the point a number starts,
                               beyond its sign, if any *)
      point: integer; (* location of decimal point *)
      power: integer; (* of 10 that a number is *)

      l: integer; (* an index for dissecting numbers *)

function figureinteger(first,last:integer):integer;
(* figure the integer in the token *)
var
      i: integer; (* index *)
      sum, increment: integer;
begin (* figureinteger *)
      power:=1; (* start at one"s place *)
      sum:=0; (* start sum at zero *)
      for i:=last downto first do begin
         case line.letters[i] of
            '0': increment:=0;
            '1': increment:=1;
            '2': increment:=2;
            '3': increment:=3;
            '4': increment:=4;
            '5': increment:=5;
            '6': increment:=6;
            '7': increment:=7;
            '8': increment:=8;
            '9': increment:=9
         end;
         sum:=sum+power*increment;
         power:=power*10
      end;
      figureinteger:=sum
end; (* figureinteger *)

begin (* figurestring *)
      numbers:=['0','1','2','3','4','5','6','7','8','9'];
(* c:=' '; i:=0; r:=0.0; do not affect these variables unless necessary *)
      point:=0;
      whzat := '.'; (* assume that we have someting to work on *)
      (* now to see if that is true: *)

      with line do if (length = 0) or (current < 1) or (current > length)
      then whzat := ' '
      else begin
         (* figure out where the first token is in the line *)
         first:=line.current;
         while (line.letters[first]=' ') and (first < line.length)
            do first:=succ(first);
         if (first = line.length) and (line.letters[first] = ' ')
            then whzat := ' ';
      end;

      if whzat <> ' ' then begin
         last:=first;
         while (line.letters[last]<>' ') and (last < line.length)
            do last:=succ(last);
         if line.letters[last] = ' ' then last := pred(last);

         (* the token is between inclusive first and last *)

         c:=line.letters[first];
         if (c in numbers) or (c in ['+','-']) then begin
            if c in ['+','-'] then begin
               case c of
                  '+': sign:=+1;
                  '-': sign:=-1;
               end;
               numberstart:=succ(first)
            end
            else begin
               sign:=+1;
               numberstart:=first
            end;

            whzat:='i';
            for l:=numberstart to last do begin
               if not(line.letters[l] in numbers)
               then if line.letters[l]='.' (* we found a period *)
                  then if whzat='i' (* if so far it is numbers *)
                     then begin
                        whzat:='r'; (* it is actually real *)
                        point:=l
                     end
                     else whzat:='g' (* it is a second '.', ie garbage *)
                  else whzat:='g' (* it is garbage *)
            end;
            (* if it is only numbers, it is integer *)

            (* build number *)
            (* if it ends in a period, it is integer *)
            if (whzat = 'r') and (point = last) then whzat:='i';

            if whzat = 'i' then begin
               if point = last (* had an ending decimal point *)
               then i:=sign * figureinteger(numberstart,pred(last))
               else i:=sign * figureinteger(numberstart,last);
               r:=i
            end
            else if whzat = 'r' then begin
               i:=figureinteger(numberstart,point-1);
               r:=sign * (i+figureinteger(point+1,last)/power);
               i:=sign * i
            end
         end
         else begin
            whzat:='c';
         end;

         (* move the start to just beyond the last character of the token *)
         line.current:=succ(last)
      end
end; (* figurestring *)

(* end module interact.figurestring version = 'prgmod 3.89  82 dec  3 tds'; *)

(* ************************************************************************ *)

(* begin module interact.getstring *)
procedure getstring(var afile: text; var buffer: string;
                    var gotten: boolean);
(* get a string from a file not using string calls.  this lets one
obtain lines from a file without interactive prompts *)
var   index: integer; (* of buffer *)
begin (* getstring *)
      clearstring(buffer);
      if eof(afile)
      then gotten := false
      else begin
         index := 0;
         while (not eoln(afile)) and (index < maxstring) do begin
            index := succ(index);
            read(afile, buffer.letters[index])
         end;

         if not eoln(afile) then begin
            writeln(output, ' getstring: a line exceeds maximum string size (',
                            maxstring:1,')');
            halt
         end;

         buffer.length := index;
         buffer.current := 1;
         readln(afile);
         gotten := true
      end
end; (* getstring *)
(* end module interact.getstring version = 'prgmod 3.89  82 dec  3 tds'; *)

(* begin module interact.gettoken *)
procedure gettoken(var afile: text; var buffer: string;
                     var token: string; var gotten: boolean);
(* get a token from the buffer or interactive file afile *)
var
      (* variables for calling figurestring: *)
      first: integer;
      last: integer;
      what: char;
      cha: char;
      int: integer;
      rea: real;
      index: integer; (* to the buffer *)
begin (* gettoken *)
      if buffer.length = 0
      then begin
         gotten:=false;
         readstring(afile,buffer)
      end
      else begin
         figurestring(buffer,first,last,what,cha,int,rea);

         if what = ' '
         then gotten := false
         else begin
            clearstring(token);
            for index := first to last
               do token.letters[index-first+1] := buffer.letters[index];
            token.length := last - first + 1;
            token.current := 1;

            gotten:=true
         end
      end
end; (* gettoken *)
(* end module interact.gettoken version = 'prgmod 3.89  82 dec  3 tds'; *)

(* begin module interact.getinteger *)
procedure getinteger(var afile: text; var buffer: string;
                     var int: integer; var gotten: boolean);
(* get the integer int from the buffer or interactive file afile *)
var
      (* variables for calling figurestring: *)
      first: integer;
      last: integer;
      what: char;
      cha: char;
      rea: real;
begin (* getinteger *)
      if buffer.length = 0
      then begin
         gotten:=false;
         readstring(afile,buffer)
      end
      else begin
         figurestring(buffer,first,last,what,cha,int,rea);

         if what <> 'i'
         then (* begin
            flagstring(output,buffer);
            writeln(output,' please type an integer'); *)
            gotten:=false
      (* end *)

      (* the two preceding sets of comment delimiters are to remove unwanted
      code *)

         else gotten:=true
      end
end; (* getinteger *)
(* end module interact.getinteger version = 'prgmod 3.89  82 dec  3 tds'; *)

(* begin module modin.isstring *)
function isstring(var s: string; a: filler): boolean;
(* is the string s the same as a? *)

const maxcompare = 150; (* length of array compare *)

var compare: array[1..maxcompare] of char; (* the array to compare to s *)
    counter: integer; (* a for-loop counting integer *)

begin (* isstring *)

      unpack(a,compare,1);
      isstring := true;
      counter := 1;
      while (s.letters[counter] <> ' ') or (compare[counter] <>' ')
      do begin
         if ord(s.letters[counter]) <> ord(compare[counter])
         then isstring := false;
         counter := counter + 1;
      end
end; (* isstring *)
(* end module modin.isstring *)

(* begin module modin.setprocess *)
procedure setprocess(var infile: text; (* input file *)
                     var buffer: string; (* array with instruction tokens *)
                     var outfile: text; (* file to which delila instruction
                                           are sent *)
                     var from: (* number of bases before site of interest *)
                               integer;
                     var upto: (* number of bases after site of interest *)
                               integer;
                     var setdone: (* is the set 'done'; was it processed? *)
                                  boolean);
(* first, write one line of delila instructions to 'outfile' of the form,
   'organism <key name>; chromosome <key name>; piece <key name>;'.
   secondly, give values to 'from' and 'upto'.  key names and values for
   'from' and 'upto' are the respective tokens in 'buffer'. *)

(* setprocess calls procedures gettoken,fillstring,writestring, and
   getinteger, and functions nostring and isstring. *)

var counter, (* counts 'for' loops *)
    int: integer; (* for giving values to from and upto *)
    organismstructure, (* part of delila instructions to be sent to outfile *)
    token: string; (* each token is an instruction to the program *)
    gotten: boolean; (* for calling 'get' procedures *)

begin (* setprocess *)


      if nostring(buffer)
      then setdone := false

      else begin

         gettoken(infile,buffer,token,gotten);

         if not isstring(token,'set                 ')
         then setdone := false

         else begin

            for counter := 1 to 3 do begin

               if nostring(buffer)
               then setdone := false

               else begin

                  gettoken(infile,buffer,token,gotten);

                  case counter of
                     1: fillstring(organismstructure,'organism            ');
                     2: fillstring(organismstructure,' chromosome         ');
                     3: begin
                           if token.letters[1] = 'p'
                           then fillstring(organismstructure,
                                           ' piece              ')
                           else if token.letters[1] = 'g'
                           then fillstring(organismstructure,
                                           ' gene               ')
                           else if token.letters[1] = 't'
                           then fillstring(organismstructure,
                                           ' transcript         ')
                           else halt;
                           gettoken(infile,buffer,token,gotten)
                        end
                  end;

                  writestring(outfile,organismstructure);
                  write(outfile,' ');
                  writestring(outfile,token);
                  write(outfile,';');
               end
            end;

            writeln(outfile);
            writeln(outfile);

            if nostring(buffer)
            then setdone := false

            else begin

               getinteger(infile,buffer,int,gotten);

               if not gotten
               then setdone := false

               else begin

                  from := int;

                  if nostring(buffer)
                  then setdone := false

                  else begin
                     getinteger(infile,buffer,int,gotten);
                     if not gotten
                     then setdone := false
                     else upto := int
                  end

               end
            end
         end
      end
end; (* setprocess *)
(* end module modin.setprocess *)

(* begin module modin.modparprocess *)
procedure modparprocess(var infile: text; (* input file *)
                        var buffer: string; (* contains tokens *)
                            from,  (* how many bases before site? *)
                            upto:  (* how many bases after site? *)
                                  integer;
                        var outfile: text; (* file to which delila
                                              instructions are sent *)
                        var modpardone: (* were the module parameters
                                          processed? *)
                                          boolean);
(* process the module parameters *)

(* modparprocess calls procedures getinteger, gettoken, and writestring,
   and functions isstring and nostring. *)

var position: (* absolute site for which delila looks *)
              integer;
    token:  (* for calling gettoken *)
           string;
    minus, (* for negative direction in delila instructions *)
    gotint,gottok: boolean; (* for getting integers and
                               strings, respectively *)

begin (* modparprocess *)

      modpardone := true;
      minus := false;

      repeat
         if not nostring(buffer)
         then begin
            getinteger(infile,buffer,position,gotint);
            if not gotint
            then begin
               buffer.current := 1;
               gettoken(infile,buffer,token,gottok);
               if isstring(token,'m                   ')
               then minus := true
               else modpardone := false;
               if nostring(buffer)
               then modpardone := false
               else getinteger(infile,buffer,position,gotint);
            end;

            if gotint
            then begin

               if nostring(buffer)
               then modpardone := false
               else gettoken(infile,buffer,token,gottok);
               write(outfile,'(* begin module ');
               writestring(outfile,token);
               writeln(outfile,' *)');

               write(outfile,'get from ',position:1,' ');

               if from >= 0
               then write(outfile,'+');
               write(outfile,from:1);
               write(outfile,' to ',position:1,' ');

               if upto >=0
               then write(outfile,'+');
               write(outfile,upto:1);
               write(outfile,' direction ');

               if minus
               then writeln(outfile,'-;')
               else writeln(outfile,'+;');

               (* if nostring(buffer) *)
               (* then modpardone := false; *)
               write(outfile,'(* end module ');
               writestring(outfile,token);
               writeln(outfile,' *)');

            end
            else modpardone := false
         end
         else modpardone := false
      until nostring(buffer)
end; (* modparprocess *)
(* end module modin.modparprocess *)

(* begin module modin.order *)
procedure modinorder(var infile: text; var outfile: text);
(* take properly formatted information from infile, send delila instructions
   to outfile. *)
(* procedure modinorder calls procedures getstring, halt, gettoken,
   writestring, setprocess, modparprocess and functions isstring
   and nostring. *)

var buffer, (* array for holding lines from infile *)
    token, (* array for holding strings from buffer *)
    name: (* name of module appearing in outfile *)
          string;
    linenumber, (* for counting lines of infile *)
    from, (* how many bases before absolute site in delila instructions? *)
    upto: (* how many bases after absolute site in delila instructions? *)
          integer;
    gotten, (* for calling get procedures *)
    setdone, (* was the set processed? *)
    modpardone: (* were the module parameters processed? *)
                boolean;

begin (* modinorder *)

      reset(infile);
      rewrite(outfile); (* files ready *)

      (* initializations *)
      linenumber := 0; (* we have counted zero lines so far *)
      from := 0; (* default number of bases before 'position.' *)
      upto := 0; (* default number of bases after 'position.' *)
      gotten := false;
      setdone := true;
      modpardone := true;

      (* begin section 1 *)
      (* in case infile is empty *)
      if eof(infile)
      then begin
         writeln(output,' input file is empty.');
         halt
      end;
      (* end section 1 *)

      (* begin section 2 *)
      (* first line of infile must be title line *)
      getstring(infile,buffer,gotten);
      linenumber := linenumber + 1; (* count the line *)
      gettoken(infile,buffer,token,gotten);

      if isstring(token,'title               ')
      then begin
         buffer.current := 1; (* go to beginning of line *)
         writestring(outfile,buffer);
         writeln(outfile,';');
      end

      else begin
         writeln(output,' title line needed in input file.');
         halt
      end;
      (* end section 2 *)

      writeln(outfile,'default out-of-range reduce-range;');

      getstring(infile,buffer,gotten);

      (* begin section 3 *)
      repeat
         linenumber := linenumber + 1;
         setprocess(infile,buffer,outfile,from,upto,setdone);
         if setdone
         then
         repeat
            getstring(infile,buffer,gotten);
            linenumber := linenumber + 1;
            modparprocess(infile,buffer,from,upto,outfile,modpardone);
            writeln(outfile);
            setdone := false;
         until not modpardone

         else begin
            writeln(output,'format error, line ', linenumber:1,
                           ' of input file.');
            rewrite(outfile);
            halt
         end;
      until eof(infile);

end; (* modinorder *)
(* end module modin.order *)

begin (* modin *)
      writeln(output,' modin ',version:4:2);
      modinorder(fin,inst);
      writeln(output,' instructions have been written.');
1: end. (* modin *)
