program mutseq(sequseed, mutseqp, output);
(* mutseq: mutate a sequence

  Dr. Thomas D. Schneider
  toms@alum.mit.edu
  permanent email: toms@alum.mit.edu
  https://alum.mit.edu/www/toms/


 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.03; (* of mutseq.p 2004 Jun 29
2004 Jun 29, 1.03: functional and works correctly
2004 Jun 29, 1.02: random procedure inserted
2004 Jun 29, 1.01: documented, starting coding
2004 Jun 29, 1.00: origin
*)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.mutseq *)
(*
name
   mutseq: mutate a sequence

synopsis
   mutseq(sequseed: in, mutseqp: in, output: out)

files

   sequseed:  DNA sequence and seed.  The first line contains
      a sequence of only the characters [acgt].  The second line
      contains a random number seed between 0 and 1.
      The sequence may be of any length less than constant maxsequ.

   mutseqp:  parameters to control the program.  The file must contain the
      following parameters, one per line:

      parameterversion: The version number of the program.  This allows the
         user to be warned if an old parameter file is used.

      If the first line contains 'v' then the program will
      show the steps it used, verbose mode.

   output: messages to the user

description

   The program reads in a DNA sequence and a random number seed (a
   real number between 0 and 1) from the sequseed file.  It makes a
   new random number seed and then uses that to select a base to
   modify and a base to modify it to.  The result is written back out
   to the sequseed file.

examples

   An example mutseqp parameter file:

1.00  version of mutseq that this parameter file is designed for.
verbose mode

   An example sequseed file:

agttatactc
 0.82911751563182178870

documentation

see also

   {Another random sequence generator program:} markov.p

author

   Thomas Dana Schneider

bugs

technical notes

   maxsequ is a constant in the program.

*)
(* end module describe.mutseq *)

var
   sequseed, (* file used by this program *)
   mutseqp: (* file used by this program *)
     text;

const

(* begin module const.mutseq *)
   seedwid = 20; (* seed number width *)
   seeddec = 20; (* seed number decimals *)
   maxsequ = 100; (* maximum sequence length *)
(* end module const.mutseq *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'matmod 2.06 94 sep 5 tds/gds'; *)

(* begin module random *)
procedure random(var seed: real);
(* the seed value is used to generate a random number in the range 0 to 1.
the seed is then altered, to provide the seed of the next random number.
in other words, the seed is the random number.  to obtain the next random
number in the pseudo-random series, simply call random again with the same
seed.  note that the series will not be truely random, since the previous
value determines the next value.  this is "pseudo-random".

an initial seed of 4.0 is suggested.  note that this is not the first
random number.  you can use different seeds to get different pseudo-random
number series.

the expected mean and standard deviation of a flat distribution is
calculated using simple integrals:

       1
mean = / x dx = 0.5
       0
           1           2
variance = / (x - mean)  dx = 1/12;  stdev = sqrt(variance) = .2886751
           0

(the integral for the variance is solved easily by substituting y = x - mean.)

   the algorithm is derived from:
pascal programs for scientists and engineers, alan r. miller
sybex, berkeley ca (1981).
which in turn was an adaptation from hp-35 applications programs *)
const
      pi = 3.141592653589; (* a value used to help mix up the result *)
var
      intermediate: real; (* an intermediate result *)
begin (* random *)
      intermediate := seed + pi;
      intermediate := exp(5.0*ln(intermediate));
      seed := intermediate - trunc(intermediate)
end; (* random *)
(* end module random version = 'matmod 2.06 94 sep 5 tds/gds'; *)

(* begin module mutseq.themain *)
procedure themain(var sequseed, mutseqp: text);
(* the main procedure of the program *)
var
   b: char; (* a base in the sequence *)
   newb: char; (* a new base in the sequence *)
   verbose: boolean; (* if true, show lots of information *)
   parameterversion: real; (* parameter version number *)
   length: integer; (* length of the sequence *)
   l: integer; (* index for the sequence *)
   seed: real; (* a random number seed *)
   sequ: array[1..maxsequ] of char; (* a DNA sequence *)
begin
   writeln(output,'mutseq ',version:4:2);
   reset(mutseqp);
   readln(mutseqp, parameterversion);
   if round(100*parameterversion) < round(100*updateversion) then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;

   if eoln(mutseqp) then begin
      verbose := false;
   end
   else begin
      verbose := mutseqp^= 'v';
      readln(mutseqp);
   end;

   (* read in the sequseed *)

   reset(sequseed);

   length := 0;
   while (length <= maxsequ) and (not eoln(sequseed)) do begin
      length := succ(length);
      read(sequseed, sequ[length]);
      if not (sequ[length] in ['a','c','g','t']) then begin
         if verbose then writeln(output,'bad letter at position ',length:1);
         sequ[length] := 'a';
         if verbose then writeln(output,'using ',sequ[length], ' instead'); 
      end;
   end;

   readln(sequseed, seed);
   if (seed < 0) or (seed > 1) then begin
      if verbose then writeln(output,' WARNING: the seed was out of range, '
                                   , seed:3:1);
      seed := 0.5;
      if verbose then writeln(output,' WARNING: setting seed to ', seed:3:1);
   end;

   for l := 1 to length do if verbose then write(output,sequ[l]);
   if verbose then writeln(output);
   if verbose then writeln(output, ' ',seed:seedwid:seeddec,' ');

   if verbose then writeln(output, ' ',length:1, ' bp');

   (* pick a base to mutate in the sequence *)
   random(seed);
   if verbose then writeln (output, ' ',seed:seedwid:seeddec, ' new seed');

   l := round(seed * length)+1;
   if verbose then writeln(output, ' ',l:1, ' position to change');

   b := sequ[l];
   if verbose then writeln(output, ' "',b, '" old base');

   newb := b;
   while (newb = b) do begin

      (* mutate the sequence *)
      random(seed);
      if verbose then writeln(output, ' ',seed:seedwid:seeddec, ' new seed');

      if seed < 0.25
      then newb := 'a'
      else if seed < 0.5
           then newb := 'c'
           else if seed < 0.75
           then newb := 'g'
           else newb := 't'
   end;
   if verbose then writeln(output, ' "',newb, '" new base at position ',l:1);

   (* do the mutation *)
   sequ[l] := newb;
   
   rewrite(sequseed);
   for l := 1 to length do write(sequseed,sequ[l]);
   writeln(sequseed);
   writeln(sequseed, ' ',seed:seedwid:seeddec,' ');


end;
(* end module mutseq.themain *)

begin
   themain(sequseed, mutseqp);
1: end.
