program normal(normalp,data,genhisp,output);
(* normal: generate normally distributed random numbers
   Tom Schneider *)

label 1; (* end of program *)

const
(* begin module version *)
version = 3.17; (* of normal.p 1994 sep 5
origin 1989 Feb 13 from gentst 3.11 tds *)
(* end module version *)

(* begin module describe.normal *)
(*
name
   normal: generate normally distributed random numbers

synopsis
   normal(normalp:in, data: out, output: out) 

files 
   normalp: parameter file controlling the program.
      Two numbers, one per line:
         seed: random seed to start the process
         total: the number of numbers to generate
   data: This is a set of numbers which should have Gaussian distribution
      if the random number generator is a reasonable one.
      It will be N(0,1), a normal distribution
      with mean 0 and standard deviation 1.
   genhisp: control file for the genhis histogram plotting program.
   output: messages to the user

description 
   Test of a random number generator by creating a gaussian
   distribution of numbers for plotting by genhis.
   Method: if U is a member of the set [0..1] and Un and Un+1
   are two members, then define
      theta = Un 2 pi
      r     = sqrt(-2 ln(Un+1)) 
   then when these polar coordinates are converted to Cartesian
   coordinates, one gets two independent Normally distributed numbers,
   with mean 0 and standard deviation 1.  To get other standard deviations
   multiply by a constant, and to get other means, add a constant.

   The proof was from a friend; I only have sketch notes at the moment.
   I'm sure it is available in standard texts.  However, it works,
   as shown by the example.

example
  seed := 0.5;
  total := 10000;
     The mean was 0.00 (to two places) and the standard deviation was 1.01.

see also
   gentst.p, tstrnd.p, genhis.p

author
  Tom Schneider
  Frederick, Maryland
  toms@alum.mit.edu

bugs
   none known
  
*)
(* end module describe.normal *)

var
(* begin module normal.var *)
   normalp,data, genhisp: text; (* output files *)
(* end module normal.var *)

(* begin module random *) 
procedure random(var seed: real); 
(* random generator 2.  version = 1.00; 1986 December 31
   Test this routine with the program tstrnd.
   written by David Masternarde *)

(* This random number generator is based on a shift register with a single bit
of feedback, as described in Electronics for Neurobiologists, by Brown, Maxfield
and Moraff, MIT press 1973, referencing Random Process Simulation and
Measurement by Korn, McGraw-Hill 1966.  The random seed rand, a number between 0
and 1 exclusive, is converted to an integer between 1 and 2**23-1, inclusive.
This 23-bit number is shifted right one bit and the output of the last (23rd)
bit and the 9th bit are added modulo 2 (exclusive orred) and fed back into the
new first bit.  This is done between 4 and 11 times, depending on the last 3
bits of the original number.  The result is converted back to a real number
between 0 and 1 from which the 23 bit integer can be recovered on the next call.
The 23-bit shift register goes through all 2**23-1 values before repeating;
the repetition frequency of this algorithm could be less or greater depending
on the seed, because of the random number of multiple shifts per call. *)
(* powers of 2 *)
const pow14=16384; pow15=32768; pow22= 4194304; pow23=8388608;
var iseed, (* integer shift register *)
    i, nrep: integer;  (* index, number of times to do shift *)
begin (* random *)  
      iseed := round(seed*pow23);  (* convert to 23 bit number *)
      if (iseed<1) or (iseed>=pow23) then iseed := 1;
      nrep := 4 + iseed mod 8;  (* do it 4 to 11 times based on last 3 bits *)
      for i:= 1 to nrep do
         (* if last bit and 9th bit are equal, feed back a 0, otherwise a 1 *)
	 if( odd(iseed) = ((iseed mod pow15) >= pow14))
	    then iseed := iseed div 2
	    else iseed := (iseed div 2) + pow22;
      seed := iseed/pow23;
end; (* random *) 
(* end module random *) 

(* begin module pic.polrec *)
procedure polrec(r,theta: real; var x,y: real);
(* convert polar to rectangular coordinates,
theta is in radians *)
begin
   x := r*cos(theta);
   y := r*sin(theta)
end;
(* end module pic.polrec *)

(* begin module normal.themain *)
procedure themain(var data, genhisp: text);
(* the main procedure of normal *)
const
  pi = 3.1415926539;  (* ratio of circumfrence to diameter of circle *)
  xnumber = 20; (* number of points on x axis to plot *)
var
  total: integer; (* total numbers to make *)
  indexfortotal: integer; (* index *)
  r: real; (* random radial distance *)
  seed: real; (* the current random number seed *)
  theta: real; (* random angle *)
  x, y: real; (* Cartesian coordinates of (r, theta) *)
begin
  writeln(output,'normal ',version:4:2);

  reset(normalp);
  readln(normalp,seed);
  readln(normalp,total);

  rewrite(genhisp);
  (* set x scale to fit a page: *)
  writeln(genhisp,'x n ', xnumber:1);
  writeln(genhisp,'p g'); (* plot gaussian *)

  rewrite(data);
  writeln(data,'* normal ',version:4:2);
  writeln(data,'* ',seed:10:9, ' random seed');
  writeln(data,'* ',total:1,' numbers');

  for indexfortotal :=1 to total do begin
    repeat
      random(seed);
      theta := 2.0 * pi * seed;
      random(seed);

      (* avoid bombing if the seed happens to be zero *)
      if seed <> 0 then r := sqrt(-2 * ln(seed));
    until seed <> 0.0;

    polrec(r,theta, x, y);
    writeln(data,x:10:8);
    writeln(data,y:10:8);
  end
end;
(* end module normal.themain *)

begin
   themain(data,genhisp);
1:end.
