program patser (book,pattern,scale,patserp,values,inst,output);
(* search a book of sequences with a pattern from the pattern learning program.
   by gary stormo (modified by tom schneider), copyright 1989
modules libraries needed: auxmods, delmods, delman  *)

label 1;  (* end of program *)

const
(* begin module version *)
version = 2.32; (* of patser.p 1994 sep 5
origin july 12, 1982 *)
(* end module version *)

(* begin module describe.patser *)
(*
name
      patser: pattern searcher

synopsis
      patser(book: in, pattern: in, scale: in, patserp: in,
             values: out, inst: out, output: out)

files
      book: the book of sequences to be searched. only numbered sequences are
         searched.
      pattern: the pattern used to search with. this is the output of the
         pattern learning program, patlrn.
      scale: contains one integer, by which the values should be divided
         to bring them into the correct scale if a matrix from rseq was used.
      patserp: parameter file, to set the value of 'printmin', the minimum
         value of a site in order for it to be identified in the file 'values'.
         if this file is empty, 'printmin' is set to the functional sequence
         minimum of the pattern matrix.
      values: the sites, and their values, which are evaluated above 'printmin'.
      inst: the instructions to get the regions around the sites identified
         in the file 'values'.  the region obtained is identical to the pattern
         used in the search.
      output: for messages to the user.

description
      patser uses a pattern matrix, the output of the pattern learning program
      patlrn, to search a book of sequences.  each base in each sequence is
      used as the 'aligned base', and the sites which are evaluated above
      'printmin' are identified in the file 'values'.  instructions which can
      be used to obtain those sites, and the nucleotides around them over the
      region contained in the pattern matrix, are put into the file 'inst'.

      NOTE: if the pattern is off the end of the sequence it is
      nolonger reported.

see also
      patrln.p, patval.p

author
      gary d. stormo (modified by tom schneider)

bugs
      none known
*)
(* end module describe.patser *)

(* begin module book.const *)
(* constants needed for book manipulations *)

      dnamax = 3000;  (* length of dna arrays *)
      namelength = 20; (* maximum key name length *)
      linelength = 80; (* maximum line readable in book *)

(* end module book.const version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* begin module matrix.const *)
      maxmatrix = 120; (* the maximum size of a pattern matrix *)
(* end module matrix.const version = 'auxmod 1.37 85 apr 4 gds/tds'; *)

type

(* begin module book.type   *)
(* types needed for book manipulations *)

      chset = set of 'a'..'z';

      (* types defined in book definition *)

      alpha = packed array[1..namelength] of char; (* this is not alfa *)

      (* name is a left justified string with blanks following the
         characters *)
      name = record
         letters: alpha;
         length: 0..namelength (* zero means an unspecified structure *)
      end;

      lineptr = ^line;
      line = record (* a line of characters *)
         letters: packed array [1..linelength] of char;
         length: 0..linelength;
         next: lineptr
      end;

      direction = (plus, minus, dircomplement, dirhomologous);
      configuration = (linear, circular);
      state = (on, off);


      header = record (* header of key *)
         keynam: name; (* key name of structure *)
         fulnam: lineptr; (* full name of structure *)
         note:   lineptr (* note key *)
      end;


      (* base types *)
      base = (a,c,g,t);
      dnaptr = ^dnastring;
      dnarange = 0..dnamax;
      seq = packed array[1..dnamax] of base;
      dnastring = record
         part: seq;
         length: dnarange;
         next: dnaptr
      end;


      orgkey = record (* organism key *)
         hea: header;
         mapunit: lineptr (* genetic map units *)
      end;

      chrkey = record (* chromosome key *)
         hea: header;
         mapbeg: real; (* number of genetic map beginning *)
         mapend: real (* number of genetic map ending *)
      end;

      pieceptr = ^piece;
      piekey = record (* piece key *)
         hea: header;
         mapbeg: real; (* genetic map beginning *)
         coocon: configuration; (* configruation (circular/linear) *)
         coodir: direction; (* direction (+/-) relative to genetic map *)
         coobeg: integer; (* beginning nucleotide *)
         cooend: integer; (* ending nucleotide *)
         piecon: configuration; (* configruation (circular/linear) *)
         piedir: direction; (* direction (+/-) relative to coordinates *)
         piebeg: integer; (* beginning nucleotide *)
         pieend: integer; (* ending nucleotide *)
      end;


      piece = record
         key: piekey;
         dna: dnaptr
      end;

      reference = record
         pienam : name;  (* name of piece referred to *)
         mapbeg : real;  (* genetic map beginning *)
         refdir : direction;  (* direction relative to coordinates *)
         refbeg : integer;  (* beginning nucleotide *)
         refend : integer;  (* ending nucleotide *)
      end;

      genkey = record  (* gene key *)
         hea : header;
         ref : reference;
      end;

      trakey = record  (* transcript key *)
         hea : header;
         ref : reference;
      end;

      markerptr = ^marker;
      markey = record (* marker key *)
         hea : header;
         ref : reference;
         sta : state;
         phenotype : lineptr;
         next : markerptr;
      end;

      marker = record
         key : markey;
         dna : dnaptr;
      end;

(* end module book.type version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* begin module matrix.type *)
      matrix = array[base,1..maxmatrix] of integer; (* a pattern matrix *)
(* end module matrix.type version = 'auxmod 1.37 85 apr 4 gds/tds'; *)

var   patserp, (* file of a single integer, to set printmin - if patserp is
               empty printmin is set by default to funcmin of the pattern *)
      inst,  (* file of delila instructions which is generated to get all
               pieces identified by this program *)
      book,  (* book of sequences to be evaluated *)
      values, (* output of the sites in the book evaluated above printmin *)
      scale,  (* scaling factor file *)
      pattern : text;  (* the matrix by which to evaluate *)
      beginning,  (* of the pattern matrix *)
      width,  (* the width of wmatrix *)
      printmin,  (* the minimum 'valu' to print the position *)
      maxafter,  (* number of bases after aligned base in pattern *)
      maxaligned : integer; (* number of positions up to (and including)
                             the aligned base in the pattern matrix *)
      wmatrix : matrix;  (* read from pattern input *)
      scaleup: integer; (* scale factor from scale file *)
      thispie : pieceptr;
            (* these variables are needed by the procedure getocp *)
      org: orgkey;
      chr: chrkey;
      orgchange,
      orgopen,
      chrchange,
      chropen,
      pieopen: boolean;

(* begin module book.var *)
(* ************************************************************************ *)
(* global variables needed for book manipulations *)

      (* free storage: *)
      freeline: lineptr; (* unused lines *)
      freedna: dnaptr; (* unused dnas *)

      readnumber: boolean; (* whether to read a number from the notes, or
                              to read in the notes *)
      number: integer; (* the number of the item just read *)
      numbered: boolean; (* true when the item just read is numbered *)
      skipunnum: boolean; (* a control variable to allow skipping of
                             un-numbered items in the book *)

(* ************************************************************************ *)
(* end module book.var version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* begin module package.primitive *)
(* ************************************************************************ *)
(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* begin module unlimitln *)
procedure unlimitln(var afile: text);
(* this procedure removes a stupid system dependent limit on the number of
lines that one can write to a file.  you may remove it from the code if
your system does not want or need this.  suggested method: place comments
around the contents of the procedure. *)
begin
(*      linelimit(afile, maxint); (@ set 'infinite' lines allowed for afile *)
end;
(* end module unlimitln version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* begin module copylines *)
function copylines(var fin, fout: text; n: integer): integer;
(* copy n lines of file fin to file fout.
the actual number of lines copied is returned. *)
var
      index: integer; (* the current line number *)
begin (* copylines *)
      index := 0;
      while (not eof(fin)) and (index < n) do begin
         copyaline(fin, fout);
         index := succ(index)
      end;

      copylines := index
end; (* copylines *)
(* end module copylines version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* ************************************************************************ *)
(* end module package.primitive version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* begin module package.getocp *)
(* ************************************************************************ *)
(* begin module package.brpiece *)
(* ************************************************************************ *)
(* begin module book.basis *)
(* procedures needed for book manipulations *)

(* get procedures should be used for all linked lists of records *)

procedure getline(var l: lineptr);
(* obtain a line from the free line list or by making a new one *)
begin
      if freeline<>nil
         then begin
            l:=freeline;
            freeline:=freeline^.next
         end
         else new(l);
      l^.length:=0;
      l^.next:=nil
end;

procedure getdna(var l: dnaptr);
begin
      if freedna<>nil
         then begin
            l:=freedna;
            freedna:=freedna^.next
         end
         else new(l);
      l^.length:=0;
      l^.next:=nil
end;

(* clear procedures should be called each time the records are no longer needed
         failure to do this may result in a stack overflow. *)

procedure clearline(var l: lineptr);
(* return a line to the free line list *)
var   lptr: lineptr;
begin
      if l<>nil then begin
         lptr:=l;
         l:=l^.next;
         lptr^.next:=freeline;
         freeline:=lptr
      end
end;

procedure cleardna(var l: dnaptr);
var   lptr: dnaptr;
begin
      if l<>nil then begin
         lptr:=l;
         l:=l^.next;
         lptr^.next:=freedna;
         freedna:=lptr
      end
end;

procedure clearheader(var h: header);
(* clear the header h (remove lines to free storage) *)
begin
      with h do begin
         clearline(fulnam);
         while note<>nil do clearline(note)
      end
end;

procedure clearpiece(var p: pieceptr);
(* clear the dna of the piece *)
begin
      while p^.dna<>nil do cleardna(p^.dna);
      clearheader(p^.key.hea)
end;

function chartobase(ch:char):base;
(* convert a character into a base *)
begin
      case ch of
         'a': chartobase:=a;
         'c': chartobase:=c;
         'g': chartobase:=g;
         't': chartobase:=t
      end
end;

function basetochar(ba:base):char;
(* convert a base into a character *)
begin
      case ba of
         a: basetochar:='a';
         c: basetochar:='c';
         g: basetochar:='g';
         t: basetochar:='t';
      end
end;

function complement(ba:base):base;
(* take the complement of ba *)
begin
      case ba of
         a: complement:=t;
         c: complement:=g;
         g: complement:=c;
         t: complement:=a;
      end
end;

function pietoint(p: integer; pie: pieceptr): integer;
(* p is a coordinate on the piece.
   we want to transform p into a number
   from 1 to n: an internal coordinate system for
   easy manipulation of piece coordinates *)
var   i: integer; (* an intermediate value *)
begin
      with pie^.key do begin
         case piedir of
            plus:  if p>=piebeg
                   then i:=p-piebeg+1
                   else i:=(p-coobeg)+(cooend-piebeg)+2;
            minus: if p<=piebeg
                   then i:=piebeg-p+1
                   else i:=(cooend-p)+(piebeg-coobeg)+2
         end;
         pietoint:=i
      end
end;

function inttopie(i: integer; pie: pieceptr):integer;
(* i is in the range 1 to some maximum.  it is an internal coordinate
   system for the program.  we want to do a
   coordinate transformation to obtain
   a value in the range of the piece called pie:
   i=1 corresponds to piebeg and
   i=its maximum corresponds to pieend *)
var   p: integer; (* an intermediate value *)
begin
      with pie^.key do begin
         case piedir of
            plus: begin
               p:=piebeg+(i-1);
               if p>cooend
                  then if coocon=circular
                     then p:=p-(cooend-coobeg+1)
            end;
            minus: begin
               p:=piebeg-(i-1);
               if p<coobeg
                  then if coocon=circular
                     then p:=p+(cooend-coobeg+1)
            end
         end;
         inttopie:=p
      end
end;

function piecelength(pie: pieceptr): integer;
(* return the length of the dna in pie *)
begin
      piecelength:=pietoint(pie^.key.pieend,pie)
end;

(* end module book.basis version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* begin module book.getto *)
function getto(var thefile: text; ch: chset): char;
(* search the file for a character in the first line which is a
   member of the set ch. *)
var achar: char;
begin
      achar:=' ';
      while (not(achar in ch)) and (not eof(thefile))
         do readln(thefile,achar);
      if (achar in ch) then getto:=achar
                       else getto:=' '
end;
(* end module book.getto version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* begin module book.skipstar *)
procedure skipstar(var thefile:text);
(* skip start of line (or star = '*'). *)
begin (* skipstar *)
      if thefile^ <> '*' then begin
         writeln(output,' procedure skipstar: bad book');
         writeln(output,' "*" expected as first character on the line, but "',
                        thefile^,'" was found');
         halt
      end;
      get(thefile); (* skip the star *)

      if thefile^ <> ' ' then begin
         writeln(output,' procedure skipstar: bad book');
         writeln(output,' "* " expected on a line but "*',
                        thefile^,'" was found');
         halt
      end;
      get(thefile) (* skip the blank *)
end; (* skipstar *)
(* end module book.skipstar version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* begin module book.brreanum *)
procedure brreanum(var thefile: text; var reanum: real);
(* read a real number from the file *)
begin
      skipstar(thefile);
      readln(thefile,reanum);
end;
(* end module book.brreanum version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* begin module book.brnumber *)
procedure brnumber(var thefile: text; var num: integer);
(* read a number from the file *)
begin
      skipstar(thefile);
      readln(thefile,num)
end;
(* end module book.brnumber version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* begin module book.brname *)
procedure brname(var thefile: text; var nam: name);
(* read a name from the file *)
var   i: integer; (* an index to the name *)
      c: char; (* a character read *)
begin (* brname *)
      skipstar(thefile);
      with nam do begin
         length:=0;
         repeat
            length:=succ(length);
            read(thefile,c);
            letters[length] := c
         until (eoln(thefile)) or
               (length>=namelength) or
               (letters[length]=' ');
         if letters[length]=' ' then length:=length-1;
         if length<namelength
            then for i:=length+1 to namelength do letters[i]:=' '
      end;
      readln(thefile)
end; (* brname *)
(* end module book.brname version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* begin module book.brline *)
procedure brline(var thefile: text; var l: lineptr);
(* read a line from the file *)
var
      i,j: integer;
      acharacter: char;
begin
      skipstar(thefile);
      i:=0;
      while (not eoln(thefile)) do begin
         i:=succ(i);
         read(thefile,acharacter);
         l^.letters[i]:=acharacter
      end;
      if i<l^.length then for j:=i+1 to l^.length do l^.letters[j]:=' ';
      l^.length:=i;
      l^.next:=nil;
      readln(thefile)
end;
(* end module book.brline version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* begin module book.brdirect *)
procedure brdirect(var thefile: text; var direct: direction);
(* read a direction *)
var   ch: char;
begin
      skipstar(thefile);
      readln(thefile,ch);
      if ch='+' then direct:=plus
                else direct:=minus
end;
(* end module book.brdirect version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* begin module book.brconfig *)
procedure brconfig(var thefile: text; var config: configuration);
(* read a configuration *)
var   ch: char;
begin
      skipstar(thefile);
      readln(thefile,ch);
      if ch='l' then config:=linear
                else config:=circular
end;
(* end module book.brconfig version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* begin module book.brnotenumber *)
procedure brnotenumber(var thefile: text; var note: lineptr);
(* book note reading to obtain the number of the object.
the procedure returns the value of the number as a global.
(this is not such a good practice, but we are stuck with it for now.) *)
begin (* brnotenumber *)
      note:=nil;
      numbered := false;
      number := 0; (* force number to zero if there
         is no number at all *)
      (* the next character is n or * depending on whether there are notes *)
      if thefile^ = 'n' then begin
         readln(thefile);
         if thefile^ <> 'n' then begin
            skipstar(thefile);
            if not eoln(thefile) then begin
               if thefile^ = '#' then begin
                  numbered := true;
                  get(thefile); (* move past the number symbol *)
                  read(thefile,number);
               end
            end;
            repeat
               readln(thefile)
            until thefile^ = 'n';
            readln(thefile)
         end
         else readln(thefile)
      end
end; (* brnotenumber *)
(* end module book.brnotenumber version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* begin module book.brnote *)
procedure brnote(var thefile: text; var note: lineptr);
(* read note key *)
var
      newnote: lineptr; (* the new note *)
      previousnote: lineptr; (* the last line of the notes *)
begin (* brnote *)
      note:=nil;
      if thefile^ = 'n' then begin (* enter note *)
         readln(thefile);
         if thefile^ <> 'n' then begin (* abort null note (n/n) *)
            getline(note);
            newnote:=note;
            while thefile^ <> 'n' do begin (* wait until end of note *)
               brline(thefile,newnote);
               previousnote:=newnote;
               (* get next note *)
               getline(newnote^.next);
               newnote:=newnote^.next;
            end;
            (* last note was not used, so: *)
            clearline(newnote);
            previousnote^.next:=nil;
            readln(thefile)
         end
         else readln(thefile)
      end
end; (* brnote *)
(* end module book.brnote version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* begin module book.brheader *)
procedure brheader(var thefile: text; var hea: header);
(* read the header of a key. *)
begin
      with hea do begin
         (* read key name *)
         brname(thefile,keynam);

         (* read full name *)
         getline(fulnam);
         brline(thefile,fulnam);

         (* read note key *)
         if readnumber then brnotenumber(thefile,note)
                       else brnote(thefile,note)
      end
end;
(* end module book.brheader version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* begin module book.brpiekey *)
procedure brpiekey(var thefile: text; var pie: piekey);
(* read piece key *)
begin
      with pie do begin
         brheader(thefile,hea);
         brreanum(thefile,mapbeg);
         brconfig(thefile,coocon);
         brdirect(thefile,coodir);
         brnumber(thefile,coobeg);
         brnumber(thefile,cooend);
         brconfig(thefile,piecon);
         brdirect(thefile,piedir);
         brnumber(thefile,piebeg);
         brnumber(thefile,pieend);
      end
end;
(* end module book.brpiekey version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* begin module book.brdna *)
procedure brdna(var thefile: text; var dna: dnaptr);
(* read in dna from thefile *)
(* note: if the dna were circularized, by linking the last dnastring
to the first, then the cleardna routine could not clear properly,
and would loop forever... there is no reason to do that, since a simple
mod function will allow one to access the circle. *)
var
      ch: char;
      workdna: dnaptr;
begin
      getdna(dna);
      workdna:=dna;
      ch:=getto(thefile,['d']);
      read(thefile,ch); (* skipstar *)
      while (ch = '*') do
      begin
         read(thefile,ch); (* skip blank *)
         repeat
            read(thefile,ch);
            if ch in ['a','c','g','t'] then begin
               if workdna^.length=dnamax then begin
                  getdna(workdna^.next);
                  workdna:=workdna^.next
               end;
               workdna^.length:=succ(workdna^.length);
               workdna^.part[workdna^.length]:=chartobase(ch)
            end
         until eoln(thefile);
         readln(thefile); (* go to next line *)
         read(thefile,ch); (* ch is either '*' or 'd' *)
      end;
      readln(thefile)
end;
(* end module book.brdna version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* begin module book.brpiece *)
procedure brpiece(var thefile: text; var pie: pieceptr);
(* read in a piece *)
begin
      brpiekey(thefile,pie^.key);
      if numbered or (not skipunnum)
         then brdna(thefile,pie^.dna)
end;
(* end module book.brpiece version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* begin module book.brinit *)
procedure brinit(var book: text);
(* check that the book is ok to read, and
   set up the global variables for br routines *)
begin (* brinit *)
      (* halt if the book is bad (first word is 'halt') or the first
         character is not * *)
      reset(book);
      if not eof(book) then begin
         (* check for the date line *)
         if book^ <> '*' then begin
            if book^ <> 'h'
            then writeln(output, ' this is not the first line of a book:')
            else writeln(output, ' bad book:');
            write(output, ' ');

            while not (eoln(book) or eof(book)) do begin
               write(output, book^);
               get(book)
            end;
            writeln(output);
            halt
         end
      end
      else begin
         writeln(output, ' book is empty');
         halt
      end;

      (* initialize free storage *)
      freeline:=nil;
      freedna:=nil;

      readnumber:=true; (* usually we read in numbers for items *)
      number:=0; (* arbitrary value *)
      numbered:=false; (* the piece has no number (none yet read in) *)
      skipunnum:=false;
end; (* brinit *)
(* end module book.brinit version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* ************************************************************************ *)
(* end module package.brpiece version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* begin module book.brorgkey *)
procedure brorgkey(var thefile: text; var org: orgkey);
(* read organism key *)
begin
      with org do begin
         brheader(thefile,hea);
         getline(mapunit);
         brline(thefile,mapunit);
      end
end;
(* end module book.brorgkey version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* begin module book.brchrkey *)
procedure brchrkey(var thefile: text; var chr: chrkey);
(* read chromosome key *)
begin
      with chr do begin
         brheader(thefile,hea);
         brreanum(thefile,mapbeg);
         brreanum(thefile,mapend);
      end
end;
(* end module book.brchrkey version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* begin module book.getocp *)
procedure getocp(var thefile: text; var org: orgkey;   var orgchange: boolean;
                                                       var orgopen:   boolean;
                                    var chr: chrkey;   var chrchange: boolean;
                                                       var chropen:   boolean;
                                    var pie: pieceptr; var pieopen:   boolean);
(*
      get the next piece and its organism and chromosome keys.
      orgchange and chrchange indicate whether or not a new organism
      or chromosome was found.
      if a piece is not found, then eoln(book) will be true.
      orgopen, chropen and pieopen are used by getocp to tell when it has
      entered an organism, chromosome or piece.
      they should be set to false initially.
      there should be one triplet for each book read.
*)

var
      ch: char;
begin
      ch:='a';
      orgchange:=false;
      chrchange:=false;
      while not (ch in [' ','p']) do begin
         ch:=getto(thefile,['o','c','p']);
         if ch <> ' '
         then case ch of
            'o': if orgopen
               then orgopen:=false (* close organism *)
               else begin
                  brorgkey(thefile,org);
                  orgopen:=true;
                  orgchange:=true
               end;
            'c': if chropen
               then chropen:=false (* close chromosome *)
               else begin
                  brchrkey(thefile,chr);
                  chropen:=true;
                  chrchange:=true
               end;
            'p': if pieopen
               then begin
                  pieopen:=false; (* close last piece *)
                  ch:='a' (* prevent falling out of the loop *)
               end
               else begin
                  brpiece(thefile,pie);
                  pieopen:=true
               end
         end
      end
end;
(* end module book.getocp version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* ************************************************************************ *)
(* end module package.getocp version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* begin module book.getbase *)
function getbase(position: integer; pie: pieceptr):base;
(* get a base from the nth position (internal coordinates) of the
   piece.  no protection is made against positions outside the piece *)
var
      workdna: dnaptr;
      p: integer; (* the last base of the dna part *)
begin
      workdna:=pie^.dna;
      p:=dnamax;
      while position>p do begin
         p:=p+dnamax;
         workdna:=workdna^.next
      end;
      getbase:=workdna^.part[position-(p-dnamax)]
end;
(* end module book.getbase version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* begin module package.iwrite *)
(* ************************************************************************ *)
(* begin module book.iwcombk *)
procedure iwcombk(var book, afile: text);
(* make a comment in the file that says the name of the book *)
begin
      write(afile,'(* ');
      reset(book);
      if copylines(book, afile, 1) = 0 then begin
         writeln(output,
            ' book is empty, can not write comment for instructions');
         halt
      end;
      writeln(afile,'*)');
end;
(* end module book.iwcombk version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* begin module book.iwname *)
procedure iwname(var thefile: text; thename: name);
(* write the name to the file *)
var   c: integer;
begin
      for c:=1 to thename.length
         do write(thefile, thename.letters[c])
end;
(* end module book.iwname version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* begin module book.iworg *)
procedure iworg(var afile: text; org: orgkey);
(* write an organism specification.  no writeln is done to allow
   write orgchr to do this. *)
begin
      write(afile,'organism ');
      iwname(afile,org.hea.keynam);
      write(afile,';');
end;
(* end module book.iworg version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* begin module book.iwchr *)
procedure iwchr(var afile: text; chr: chrkey);
(* write an chromosome specification.  no writeln is done to allow
   write orgchr to do this. *)
begin
      write(afile,'chromosome ');
      iwname(afile,chr.hea.keynam);
      write(afile,';');
end;
(* end module book.iwchr version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* begin module book.iwpie *)
procedure iwpie(var afile: text; pie: piekey);
(* write a piece specification *)
begin
      write(afile,'piece ');
      iwname(afile,pie.hea.keynam);
      writeln(afile,';');
end;
(* end module book.iwpie version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* begin module book.iworgchr *)
procedure iworgchr(var afile: text;
                       org: orgkey; orgchange: boolean;
                       chr: chrkey; chrchange: boolean);
(* write both organism and chromosome specifications, based
   on orgchange and chrchange.  see getocp in the br routines. *)
begin
      if orgchange then iworg(afile,org);
      if orgchange and chrchange then write(afile,' ');
      if chrchange then iwchr(afile,chr);
      if orgchange or chrchange then writeln(afile)
end;
(* end module book.iworgchr version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* begin module book.iwget *)
procedure iwget(var afile: text; pie: pieceptr;
                    fromplace, pieceplace, toplace: integer;
                    flip: boolean);
(* print a get delila instruction in the orientation of pie,
   from fromplace to toplace pieceplace. +/- direction;
If flip is false, the piece direction is as on the piece,
if it is true, the it is the opposite direction. *)
procedure iwplace(relative: integer);
procedure iwrelative(relative: integer);
begin
           if relative>=0 then write(afile,' +', relative:1)
      else if relative< 0 then write(afile,' ',relative:1)
end;
begin (* iwplace *)
      write(afile,' ',pieceplace:1);
      case pie^.key.piedir of
         plus:  iwrelative(+relative);
         minus: iwrelative(-relative);
      end
end;
begin (* iwget *)
      write(afile,'get from');
      iwplace(fromplace);
      write(afile,' to');
      iwplace(toplace);

      write(afile,' direction');
      case pie^.key.piedir of
         plus:  case flip of
                  false: write(afile,' +');
                   true: write(afile,' -');
                end;
         minus: case flip of
                  false: write(afile,' -');
                   true: write(afile,' +');
                end
      end;         
 
      writeln(afile,';')
end;
(* end module book.iwget version = 'delmod 6.61 91 Mar 19 tds/gds' *)

(* ************************************************************************ *)
(* end module package.iwrite version = 'delmod 6.61 91 Mar 19 tds/gds' *)


procedure valuheader;  (* header for 'values' file *)

begin (* valuhead *)
      writeln(values,'* patser ',version:4:2,
                     '; evaluations of sequences in ');
      copyaline(book,values);
      write(values,'* by the matrix from ');
      reset(pattern);
      readln(pattern);  (* skip first line of pattern values *)
      write(values,'*');
      copyaline(pattern,values);
      writeln(values,'*');
end;

procedure instheader;  (* header for inst file *)

begin
      writeln(inst,'(* delila instructions from patser ',
         version:4:2,' evaluations of sequences in ');
      reset(book);
      copyaline(book,inst);
      reset(pattern);
      write(inst,'by the matrix from ');
      readln(pattern);  (* skip first line of pattern input *)
      copyaline(pattern,inst);
      writeln(inst,' *)');
      writeln(inst,'  ');
      writeln(inst,'default out-of-range reduce-range;');
      writeln(inst,'default numbering piece;');
      writeln(inst,'default key note off;');
      writeln(inst,' ');
end;

(* begin module matrix.readmatrix *)
procedure findlearnend(var pattern: text);
(* locate the end of the cyclic learning part of the pattern file *)
const
      learnmax = 16; (* the size of learnend *)
var
      (* this array contains the pattern that defines the end *)
      learnend: packed array[1..learnmax] of char;
      state: integer; (* how close we are to finding learnend *)
begin (* findlearnend *)
      reset(pattern);
      learnend := 'end of learning.';
      state := 1;
      while not eof(pattern) and (state < learnmax) do begin
         if learnend[state] = pattern^
         then state := succ(state)
         else state :=1;
         if eoln(pattern) then readln(pattern)
                          else get(pattern)
      end;
      if eof(pattern) then begin
         write  (output,'pattern matrix does not contain "');
         writeln(output,learnend,'" signal');
         halt
      end
end; (* findlearnend *)
procedure getcolon(var f: text);
(* move to the next colon in file f *)
begin (* getcolon *)
      while f^<>':' do begin
         get(f);
         if eof(f) then begin
            writeln(output,'pattern is missing colons');
            halt
         end
      end;
      get(f); (* move past the colon *)
end; (* getcolon *)

procedure readmatrix(var thefile: text; var wmatrix: matrix;
                     var beginning, width: integer);
(* read a pattern matrix wmatrix from the file thefile, returning
the aligned position for the first position (beginning) and the width *)
var   i : integer;  (* an index *)
      ba : base;  (* an index *)
begin (* readmatrix *)
      findlearnend(thefile);

      (* skip min and max values *)
      getcolon(thefile); getcolon(thefile);

      getcolon(thefile); (* skip to range *)
      read(thefile,beginning); (* read first position of the matrix *)

      getcolon(thefile); (* skip to width *)
      read(thefile,width); (* read width of the matrix *)

      if (width > maxmatrix) then begin
         writeln(output,'input matrix too large');
         halt;
      end;

      for i := 1 to width do   (* read in matrix *)
      begin
         getcolon(thefile); (* skip position label *)
         for ba := a to t do read(thefile,wmatrix[ba,i]);
         readln(thefile);
      end;
            (* set all the unused elements of wmatrix to 0 *)
      for i := (width + 1) to maxmatrix do
         for ba := a to t do wmatrix[ba,i] := 0;
end; (* readmatrix *)
(* end module matrix.readmatrix version = 'auxmod 1.37 85 apr 4 gds/tds'; *)

procedure printingminimum; (* determines the minimum 'valu' for which
                 a sequence will be identified.  it is either read from file
                 patserp or set by default to the funcmin of the pattern *)
begin (* printingminimum *)
      if not eof(patserp)
      then read(patserp,printmin)
      else begin
         findlearnend(pattern);
         getcolon(pattern); (* skip to minimum *)
         readln(pattern,printmin);
      end;
      writeln(values,'* sequences identified have a minimum value of ',
         printmin:4);
      writeln(values,'*');
(*
      writeln(values,'* value, sequence #, position');
*)
      writeln(values,'* sequence #, position, value');
end; (* printingminimum *)

procedure evaluate; (* each position in the sequence is placed in the
   'aligned position' of the wmatrix, and the 'valu' of the sequence at that
   position, according to the wmatrix, is calculated.  those positions which
   have 'valu' > than printmin are listed in the file 'values', and the
   instructions to get the region so evaluated are put into the file 'inst' *)

var
      valu,  (* value obtained by evaluating seq by matrix *)
      firstpos, (* the first position of the sequence to be evaluated for
                  this particular aligned base.  it is 'maxaligned' - 1 bases
                  before the aligned base. *)
      lastpos, (* the last position of the sequence to be evaluated for
                  this particular aligned base.  it is 'maxafter' bases after
                  the aligned base. *)
      alignedbase, (* the base in the piece aligned with the 0 position
                   of the pattern matrix *)
      length,  (* length of the piece dna *)
      i : integer;
      offend: boolean; (* this is true if the pattern matrix extends beyond
                        one end of the linear sequence being searched. *)

procedure getvalue(thispie: pieceptr; seqpos, matpos: integer);
(* this gets the base at position 'seqpos' in the sequence pointed to by
   'thispie', and evaluates it by the wmatrix at position 'matpos' *)
begin (* getvalue *)
      valu := valu + (wmatrix[getbase(seqpos,thispie),matpos]);
end; (* getvalue *)

begin (* evaluate *)
      getocp(book,org,orgchange,orgopen,chr,chrchange,chropen,thispie,pieopen);
      if not eof(book) then begin
         iworgchr(inst,org,orgchange,chr,chrchange);
         iwpie(inst,thispie^.key);
         length := piecelength(thispie);
         for alignedbase := 1 to length do
         begin
            valu := 0;
            offend := false;
            firstpos := alignedbase - maxaligned + 1;
            lastpos := alignedbase + maxafter;
            if ( firstpos > 0) then     (* not off beginning of piece *)
               if (lastpos <= length) then   (* not off end of piece *)
                  for i := firstpos to lastpos do
                     getvalue(thispie,i,i - firstpos + 1)
               else begin   (* off end of piece *)
                  for i := firstpos to length do
                     getvalue(thispie,i,i - firstpos + 1);
                  if (thispie^.key.piecon = circular) then
                           (* get bases from beginning *)
                     for i := 1 to (lastpos - length) do
                        getvalue(thispie,i,width + i - lastpos + length)
                  else offend := true
               end
            else begin   (* off beginning of piece *)
               if (lastpos <= length) then  (* not off end of piece *)
                  for i := 1 to lastpos do
                     getvalue(thispie,i,1 - firstpos + i)
               else begin  (* off end of piece too *)
                  for i := 1 to length do
                     getvalue(thispie,i,1 - firstpos + i);
                  if (thispie^.key.piecon = circular) then
                           (* get bases from beginning *)
                     for i := 1 to (lastpos - length) do
                        getvalue(thispie,i,width + i - lastpos + length)
               end;
               if (thispie^.key.piecon = circular) then
                           (* get bases from end *)
                  for i := (length + firstpos) to length do
                     getvalue(thispie,i,1 - firstpos - length + i)
               else offend := true;
            end;
            if (valu >= printmin) and (not offend) then begin
               write(values,
                            number:8,
                            inttopie(alignedbase,thispie):12,
                            (valu/scaleup):20:10);
(*
               if offend then
                  writeln(values,'   * pattern extends beyond sequence')
               else writeln(values);
*)
               writeln(values);

               iwget(inst,thispie,1-maxaligned,
                  inttopie(alignedbase,thispie),maxafter,false);
            end;
         end;
         clearpiece(thispie);
      end;
end; (* evaluate *)

procedure initialize;
begin (* initialize *)
      brinit(book);
      reset(pattern);
      reset(patserp);
      rewrite(values);
      rewrite(inst);

      (* find the value by which to divide all the values for output *)
      reset(scale);
      scaleup := 0;
      if not eof(scale) then read(scale,scaleup);
      if scaleup <= 0 then scaleup := 1;

      skipunnum := true;  (* we evaluate only numbered sequences *)
      new(thispie);
         (* initialize these for getocp procedure *)
      orgopen := false;
      chropen := false;
      pieopen := false;
end; (* initialize *)

begin (* patser *)
      writeln(output,'patser ',version:4:2);
      initialize;
      readmatrix(pattern,wmatrix,beginning,width);
(* debug writeln(output,wmatrix[a,4]:1,' [a,4]'); *)
(* debug writeln(output,'beginning=',beginning:1,' width=',width:1);*)
      valuheader;
      instheader;
      printingminimum;
      maxaligned := 1 - beginning;
      maxafter := width - maxaligned;
(* debug writeln(output,wmatrix[a,4]:1,' [a,4]'); *)
      while not eof(book) do  evaluate;
1: end. (* patser *)
