program pemowe(book,list,output);
(* molecular weight of first peptide in each piece of the book
   by thomas schneider

from modules: delman, delmods *)

label 1; (* end of program *)

const
(* begin module version *)
version = 2.20; (* of pemowe.p 1993 Jan 19
origin 1981 july 16 *)
(* end module version *)


(* begin module describe.pemowe *)
(*
name
      pemowe: peptide molecular weights

synopsis
      pemowe(book: in, list: out, output: out)

files
      book: any book from the delila system.  only one weight is given per
         piece in the book.  the first triplet of the piece is the first
         codon translated.
      list: a list of the piece numbers and names in the book, along with
         the molecular weight of the peptide and
         the number of atoms of the peptide for each piece.
      output: messages to the user.

description
      pemowe is designed to find the molecular weights of polypeptides
      that might be coded by a particular sequence.  it is to be used in
      cases where one knows where a particular peptide is, not when one wants
      the weights of all possible peptides.  one should use delila to
      construct the book.  the calculation of the weights takes into account
      loss of water for each peptide bond formed.  calculation ends at stop
      codons or at the end of the piece.

examples
      sanger et. al., nature 265: 687 (1977) on page 692 gives a list of
      calculated molecular weights from phix174.  these can be used to test
      pemowe, by using the delila instructions expepin.  the largest deviation
      from sanger's numbers is for gene j at 3 percent.

documentation
      data is from the crc handbook of chemistry and physics 60th ed, 1980.

author
      thomas d. schneider

bugs
      only one peptide per piece is calculated.  one could write another
      program that predicted peptides (like lister), and generated
      instructions for pulling out those peptides using delila.

*)
(* end module describe.pemowe *)

(* begin module book.const *)
(* constants needed for book manipulations *)

      dnamax = 3000;  (* length of dna arrays *)
      namelength = 20; (* maximum key name length *)
      linelength = 80; (* maximum line readable in book *)

(* end module book.const version = 'delmod 6.51 85 apr 17 tds/gds' *)

type
(* begin module book.type *)
(* types needed for book manipulations *)

      chset = set of 'a'..'z';

      (* types defined in book definition *)

      alpha = packed array[1..namelength] of char; (* this is not alfa *)

      (* name is a left justified string with blanks following the
         characters *)
      name = record
         letters: alpha;
         length: 0..namelength (* zero means an unspecified structure *)
      end;

      lineptr = ^line;
      line = record (* a line of characters *)
         letters: packed array [1..linelength] of char;
         length: 0..linelength;
         next: lineptr
      end;

      direction = (plus, minus, dircomplement, dirhomologous);
      configuration = (linear, circular);
      state = (on, off);


      header = record (* header of key *)
         keynam: name; (* key name of structure *)
         fulnam: lineptr; (* full name of structure *)
         note:   lineptr (* note key *)
      end;


      (* base types *)
      base = (a,c,g,t);
      dnaptr = ^dnastring;
      dnarange = 0..dnamax;
      seq = packed array[1..dnamax] of base;
      dnastring = record
         part: seq;
         length: dnarange;
         next: dnaptr
      end;


      orgkey = record (* organism key *)
         hea: header;
         mapunit: lineptr (* genetic map units *)
      end;

      chrkey = record (* chromosome key *)
         hea: header;
         mapbeg: real; (* number of genetic map beginning *)
         mapend: real (* number of genetic map ending *)
      end;

      pieceptr = ^piece;
      piekey = record (* piece key *)
         hea: header;
         mapbeg: real; (* genetic map beginning *)
         coocon: configuration; (* configruation (circular/linear) *)
         coodir: direction; (* direction (+/-) relative to genetic map *)
         coobeg: integer; (* beginning nucleotide *)
         cooend: integer; (* ending nucleotide *)
         piecon: configuration; (* configruation (circular/linear) *)
         piedir: direction; (* direction (+/-) relative to coordinates *)
         piebeg: integer; (* beginning nucleotide *)
         pieend: integer; (* ending nucleotide *)
      end;


      piece = record
         key: piekey;
         dna: dnaptr
      end;

      reference = record
         pienam : name;  (* name of piece referred to *)
         mapbeg : real;  (* genetic map beginning *)
         refdir : direction;  (* direction relative to coordinates *)
         refbeg : integer;  (* beginning nucleotide *)
         refend : integer;  (* ending nucleotide *)
      end;

      genkey = record  (* gene key *)
         hea : header;
         ref : reference;
      end;

      trakey = record  (* transcript key *)
         hea : header;
         ref : reference;
      end;

      markerptr = ^marker;
      markey = record (* marker key *)
         hea : header;
         ref : reference;
         sta : state;
         phenotype : lineptr;
         next : markerptr;
      end;

      marker = record
         key : markey;
         dna : dnaptr;
      end;

(* end module book.type version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module amino.type *)
      aminoacid = packed array[1..3] of char;
(* end module amino.type version = 'delmod 6.51 85 apr 17 tds/gds' *)

var
      book, list: text; (* the book and resulting listing *)

(* begin module book.var *)
(* ************************************************************************ *)
(* global variables needed for book manipulations *)

      (* free storage: *)
      freeline: lineptr; (* unused lines *)
      freedna: dnaptr; (* unused dnas *)

      readnumber: boolean; (* whether to read a number from the notes, or
                              to read in the notes *)
      number: integer; (* the number of the item just read *)
      numbered: boolean; (* true when the item just read is numbered *)
      skipunnum: boolean; (* a control variable to allow skipping of
                             un-numbered items in the book *)

(* ************************************************************************ *)
(* end module book.var version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module package.primitive *)
(* ************************************************************************ *)
(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module copylines *)
function copylines(var fin, fout: text; n: integer): integer;
(* copy n lines of file fin to file fout.
the actual number of lines copied is returned. *)
var
      index: integer; (* the current line number *)
begin (* copylines *)
      index := 0;
      while (not eof(fin)) and (index < n) do begin
         copyaline(fin, fout);
         index := succ(index)
      end;

      copylines := index
end; (* copylines *)
(* end module copylines version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* ************************************************************************ *)
(* end module package.primitive version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module package.getpiece *)
(* ************************************************************************ *)
(* begin module package.brpiece *)
(* ************************************************************************ *)
(* begin module book.basis *)
(* procedures needed for book manipulations *)

(* get procedures should be used for all linked lists of records *)

procedure getline(var l: lineptr);
(* obtain a line from the free line list or by making a new one *)
begin
      if freeline<>nil
         then begin
            l:=freeline;
            freeline:=freeline^.next
         end
         else new(l);
      l^.length:=0;
      l^.next:=nil
end;

procedure getdna(var l: dnaptr);
begin
      if freedna<>nil
         then begin
            l:=freedna;
            freedna:=freedna^.next
         end
         else new(l);
      l^.length:=0;
      l^.next:=nil
end;

(* clear procedures should be called each time the records are no longer needed
         failure to do this may result in a stack overflow. *)

procedure clearline(var l: lineptr);
(* return a line to the free line list *)
var   lptr: lineptr;
begin
      if l<>nil then begin
         lptr:=l;
         l:=l^.next;
         lptr^.next:=freeline;
         freeline:=lptr
      end
end;

procedure cleardna(var l: dnaptr);
var   lptr: dnaptr;
begin
      if l<>nil then begin
         lptr:=l;
         l:=l^.next;
         lptr^.next:=freedna;
         freedna:=lptr
      end
end;

procedure clearheader(var h: header);
(* clear the header h (remove lines to free storage) *)
begin
      with h do begin
         clearline(fulnam);
         while note<>nil do clearline(note)
      end
end;

procedure clearpiece(var p: pieceptr);
(* clear the dna of the piece *)
begin
      while p^.dna<>nil do cleardna(p^.dna);
      clearheader(p^.key.hea)
end;

function chartobase(ch:char):base;
(* convert a character into a base *)
begin
      case ch of
         'a': chartobase:=a;
         'c': chartobase:=c;
         'g': chartobase:=g;
         't': chartobase:=t
      end
end;

function basetochar(ba:base):char;
(* convert a base into a character *)
begin
      case ba of
         a: basetochar:='a';
         c: basetochar:='c';
         g: basetochar:='g';
         t: basetochar:='t';
      end
end;

function complement(ba:base):base;
(* take the complement of ba *)
begin
      case ba of
         a: complement:=t;
         c: complement:=g;
         g: complement:=c;
         t: complement:=a;
      end
end;

function pietoint(p: integer; pie: pieceptr): integer;
(* p is a coordinate on the piece.
   we want to transform p into a number
   from 1 to n: an internal coordinate system for
   easy manipulation of piece coordinates *)
var   i: integer; (* an intermediate value *)
begin
      with pie^.key do begin
         case piedir of
            plus:  if p>=piebeg
                   then i:=p-piebeg+1
                   else i:=(p-coobeg)+(cooend-piebeg)+2;
            minus: if p<=piebeg
                   then i:=piebeg-p+1
                   else i:=(cooend-p)+(piebeg-coobeg)+2
         end;
         pietoint:=i
      end
end;

function inttopie(i: integer; pie: pieceptr):integer;
(* i is in the range 1 to some maximum.  it is an internal coordinate
   system for the program.  we want to do a
   coordinate transformation to obtain
   a value in the range of the piece called pie:
   i=1 corresponds to piebeg and
   i=its maximum corresponds to pieend *)
var   p: integer; (* an intermediate value *)
begin
      with pie^.key do begin
         case piedir of
            plus: begin
               p:=piebeg+(i-1);
               if p>cooend
                  then if coocon=circular
                     then p:=p-(cooend-coobeg+1)
            end;
            minus: begin
               p:=piebeg-(i-1);
               if p<coobeg
                  then if coocon=circular
                     then p:=p+(cooend-coobeg+1)
            end
         end;
         inttopie:=p
      end
end;

function piecelength(pie: pieceptr): integer;
(* return the length of the dna in pie *)
begin
      piecelength:=pietoint(pie^.key.pieend,pie)
end;

(* end module book.basis version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module book.getto *)
function getto(var thefile: text; ch: chset): char;
(* search the file for a character in the first line which is a
   member of the set ch. *)
var achar: char;
begin
      achar:=' ';
      while (not(achar in ch)) and (not eof(thefile))
         do readln(thefile,achar);
      if (achar in ch) then getto:=achar
                       else getto:=' '
end;
(* end module book.getto version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module book.skipstar *)
procedure skipstar(var thefile:text);
(* skip start of line (or star = '*'). *)
begin (* skipstar *)
      if thefile^ <> '*' then begin
         writeln(output,' procedure skipstar: bad book');
         writeln(output,' "*" expected as first character on the line, but "',
                        thefile^,'" was found');
         halt
      end;
      get(thefile); (* skip the star *)

      if thefile^ <> ' ' then begin
         writeln(output,' procedure skipstar: bad book');
         writeln(output,' "* " expected on a line but "*',
                        thefile^,'" was found');
         halt
      end;
      get(thefile) (* skip the blank *)
end; (* skipstar *)
(* end module book.skipstar version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module book.brreanum *)
procedure brreanum(var thefile: text; var reanum: real);
(* read a real number from the file *)
begin
      skipstar(thefile);
      readln(thefile,reanum);
end;
(* end module book.brreanum version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module book.brnumber *)
procedure brnumber(var thefile: text; var num: integer);
(* read a number from the file *)
begin
      skipstar(thefile);
      readln(thefile,num)
end;
(* end module book.brnumber version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module book.brname *)
procedure brname(var thefile: text; var nam: name);
(* read a name from the file *)
var   i: integer; (* an index to the name *)
      c: char; (* a character read *)
begin (* brname *)
      skipstar(thefile);
      with nam do begin
         length:=0;
         repeat
            length:=succ(length);
            read(thefile,c);
            letters[length] := c
         until (eoln(thefile)) or
               (length>=namelength) or
               (letters[length]=' ');
         if letters[length]=' ' then length:=length-1;
         if length<namelength
            then for i:=length+1 to namelength do letters[i]:=' '
      end;
      readln(thefile)
end; (* brname *)
(* end module book.brname version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module book.brline *)
procedure brline(var thefile: text; var l: lineptr);
(* read a line from the file *)
var
      i,j: integer;
      acharacter: char;
begin
      skipstar(thefile);
      i:=0;
      while (not eoln(thefile)) do begin
         i:=succ(i);
         read(thefile,acharacter);
         l^.letters[i]:=acharacter
      end;
      if i<l^.length then for j:=i+1 to l^.length do l^.letters[j]:=' ';
      l^.length:=i;
      l^.next:=nil;
      readln(thefile)
end;
(* end module book.brline version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module book.brdirect *)
procedure brdirect(var thefile: text; var direct: direction);
(* read a direction *)
var   ch: char;
begin
      skipstar(thefile);
      readln(thefile,ch);
      if ch='+' then direct:=plus
                else direct:=minus
end;
(* end module book.brdirect version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module book.brconfig *)
procedure brconfig(var thefile: text; var config: configuration);
(* read a configuration *)
var   ch: char;
begin
      skipstar(thefile);
      readln(thefile,ch);
      if ch='l' then config:=linear
                else config:=circular
end;
(* end module book.brconfig version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module book.brnotenumber *)
procedure brnotenumber(var thefile: text; var note: lineptr);
(* book note reading to obtain the number of the object.
the procedure returns the value of the number as a global.
(this is not such a good practice, but we are stuck with it for now.) *)
begin (* brnotenumber *)
      note:=nil;
      numbered := false;
      number := 0; (* force number to zero if there
         is no number at all *)
      (* the next character is n or * depending on whether there are notes *)
      if thefile^ = 'n' then begin
         readln(thefile);
         if thefile^ <> 'n' then begin
            skipstar(thefile);
            if not eoln(thefile) then begin
               if thefile^ = '#' then begin
                  numbered := true;
                  get(thefile); (* move past the number symbol *)
                  read(thefile,number);
               end
            end;
            repeat
               readln(thefile)
            until thefile^ = 'n';
            readln(thefile)
         end
         else readln(thefile)
      end
end; (* brnotenumber *)
(* end module book.brnotenumber version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module book.brnote *)
procedure brnote(var thefile: text; var note: lineptr);
(* read note key *)
var
      newnote: lineptr; (* the new note *)
      previousnote: lineptr; (* the last line of the notes *)
begin (* brnote *)
      note:=nil;
      if thefile^ = 'n' then begin (* enter note *)
         readln(thefile);
         if thefile^ <> 'n' then begin (* abort null note (n/n) *)
            getline(note);
            newnote:=note;
            while thefile^ <> 'n' do begin (* wait until end of note *)
               brline(thefile,newnote);
               previousnote:=newnote;
               (* get next note *)
               getline(newnote^.next);
               newnote:=newnote^.next;
            end;
            (* last note was not used, so: *)
            clearline(newnote);
            previousnote^.next:=nil;
            readln(thefile)
         end
         else readln(thefile)
      end
end; (* brnote *)
(* end module book.brnote version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module book.brheader *)
procedure brheader(var thefile: text; var hea: header);
(* read the header of a key. *)
begin
      with hea do begin
         (* read key name *)
         brname(thefile,keynam);

         (* read full name *)
         getline(fulnam);
         brline(thefile,fulnam);

         (* read note key *)
         if readnumber then brnotenumber(thefile,note)
                       else brnote(thefile,note)
      end
end;
(* end module book.brheader version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module book.brpiekey *)
procedure brpiekey(var thefile: text; var pie: piekey);
(* read piece key *)
begin
      with pie do begin
         brheader(thefile,hea);
         brreanum(thefile,mapbeg);
         brconfig(thefile,coocon);
         brdirect(thefile,coodir);
         brnumber(thefile,coobeg);
         brnumber(thefile,cooend);
         brconfig(thefile,piecon);
         brdirect(thefile,piedir);
         brnumber(thefile,piebeg);
         brnumber(thefile,pieend);
      end
end;
(* end module book.brpiekey version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module book.brdna *)
procedure brdna(var thefile: text; var dna: dnaptr);
(* read in dna from thefile *)
(* note: if the dna were circularized, by linking the last dnastring
to the first, then the cleardna routine could not clear properly,
and would loop forever... there is no reason to do that, since a simple
mod function will allow one to access the circle. *)
var
      ch: char;
      workdna: dnaptr;
begin
      getdna(dna);
      workdna:=dna;
      ch:=getto(thefile,['d']);
      read(thefile,ch); (* skipstar *)
      while (ch = '*') do
      begin
         read(thefile,ch); (* skip blank *)
         repeat
            read(thefile,ch);
            if ch in ['a','c','g','t'] then begin
               if workdna^.length=dnamax then begin
                  getdna(workdna^.next);
                  workdna:=workdna^.next
               end;
               workdna^.length:=succ(workdna^.length);
               workdna^.part[workdna^.length]:=chartobase(ch)
            end
         until eoln(thefile);
         readln(thefile); (* go to next line *)
         read(thefile,ch); (* ch is either '*' or 'd' *)
      end;
      readln(thefile)
end;
(* end module book.brdna version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module book.brpiece *)
procedure brpiece(var thefile: text; var pie: pieceptr);
(* read in a piece *)
begin
      brpiekey(thefile,pie^.key);
      if numbered or (not skipunnum)
         then brdna(thefile,pie^.dna)
end;
(* end module book.brpiece version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module book.brinit *)
procedure brinit(var book: text);
(* check that the book is ok to read, and
   set up the global variables for br routines *)
begin (* brinit *)
      (* halt if the book is bad (first word is 'halt') or the first
         character is not * *)
      reset(book);
      if not eof(book) then begin
         (* check for the date line *)
         if book^ <> '*' then begin
            if book^ <> 'h'
            then writeln(output, ' this is not the first line of a book:')
            else writeln(output, ' bad book:');
            write(output, ' ');

            while not (eoln(book) or eof(book)) do begin
               write(output, book^);
               get(book)
            end;
            writeln(output);
            halt
         end
      end
      else begin
         writeln(output, ' book is empty');
         halt
      end;

      (* initialize free storage *)
      freeline:=nil;
      freedna:=nil;

      readnumber:=true; (* usually we read in numbers for items *)
      number:=0; (* arbitrary value *)
      numbered:=false; (* the piece has no number (none yet read in) *)
      skipunnum:=false;
end; (* brinit *)
(* end module book.brinit version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* ************************************************************************ *)
(* end module package.brpiece version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module book.getpiece *)
procedure getpiece(var thefile: text; var pie: pieceptr);
(* move to and read in the next piece in the book *)
var   ch: char;
begin
      ch:=getto(thefile,['p']); (* get to the next p(iece) in the book *)
      if ch<>' ' then begin
         brpiece(thefile,pie);
         ch:=getto(thefile,['p']); (* read past closing p *)
      end
end;
(* end module book.getpiece version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* ************************************************************************ *)
(* end module package.getpiece version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module book.getbase *)
function getbase(position: integer; pie: pieceptr):base;
(* get a base from the nth position (internal coordinates) of the
   piece.  no protection is made against positions outside the piece *)
var
      workdna: dnaptr;
      p: integer; (* the last base of the dna part *)
begin
      workdna:=pie^.dna;
      p:=dnamax;
      while position>p do begin
         p:=p+dnamax;
         workdna:=workdna^.next
      end;
      getbase:=workdna^.part[position-(p-dnamax)]
end;
(* end module book.getbase version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module amino.getaminoacid *)
procedure getaminoacid(position: integer; pie: pieceptr;
                       var aa: aminoacid);
(* get the amino acid corresponding to a codon starting in pie at position.
   the amino acid is blank ('   ') if it is outside a linear piece.
   amb = amber nonsense (uag)
   och = ochre nonsense (uaa)
   opa = opal  nonsense (uga)
*)
var
      b0, b1, b2: base; (* bases that correspond to the amino acid *)
      length: integer; (* length of the piece *)
      done: boolean; (* flag for out of bounds test *)
begin
      length:=pietoint(pie^.key.pieend,pie);
      done:=false;

      (* deal with position out of the piece *)
      if (position < 1) or (position > length - 2)
      then case pie^.key.piecon of
         linear: begin (* ignore out of bounds requests *)
            aa:='   ';
            done:=true
         end;
         circular: begin (* modify the local copy of position
                            to be within the circle. *)
            while position < 1 do position:=position + length;
            while position > length do position:=position - length;

            b0:=getbase(position,pie);

            (* step through the codon, making sure to stay on the circle. *)
            position:=succ(position);
            if position > length then position:=1;
            b1:=getbase(position,pie);

            position:=succ(position);
            if position > length then position:=1;
            b2:=getbase(position,pie);
         end
      end
      else begin (* pick up the bases at the position *)
         b0:=getbase(position+0,pie);
         b1:=getbase(position+1,pie);
         b2:=getbase(position+2,pie);
      end;

      if not done then case b0 of
         t: case b1 of
            t: case b2 of
               t: aa:='phe';
               c: aa:='phe';
               a: aa:='leu';
               g: aa:='leu';
            end;
            c: case b2 of
               t: aa:='ser';
               c: aa:='ser';
               a: aa:='ser';
               g: aa:='ser';
            end;
            a: case b2 of
               t: aa:='tyr';
               c: aa:='tyr';
               a: aa:='och';
               g: aa:='amb';
            end;
            g: case b2 of
               t: aa:='cys';
               c: aa:='cys';
               a: aa:='opa';
               g: aa:='trp';
            end;
         end;
         c: case b1 of
            t: case b2 of
               t: aa:='leu';
               c: aa:='leu';
               a: aa:='leu';
               g: aa:='leu';
            end;
            c: case b2 of
               t: aa:='pro';
               c: aa:='pro';
               a: aa:='pro';
               g: aa:='pro';
            end;
            a: case b2 of
               t: aa:='his';
               c: aa:='his';
               a: aa:='gln';
               g: aa:='gln';
            end;
            g: case b2 of
               t: aa:='arg';
               c: aa:='arg';
               a: aa:='arg';
               g: aa:='arg';
            end;
         end;
         a: case b1 of
            t: case b2 of
               t: aa:='ile';
               c: aa:='ile';
               a: aa:='ile';
               g: aa:='met';
            end;
            c: case b2 of
               t: aa:='thr';
               c: aa:='thr';
               a: aa:='thr';
               g: aa:='thr';
            end;
            a: case b2 of
               t: aa:='asn';
               c: aa:='asn';
               a: aa:='lys';
               g: aa:='lys';
            end;
            g: case b2 of
               t: aa:='ser';
               c: aa:='ser';
               a: aa:='arg';
               g: aa:='arg';
            end;
         end;
         g: case b1 of
            t: case b2 of
               t: aa:='val';
               c: aa:='val';
               a: aa:='val';
               g: aa:='val';
            end;
            c: case b2 of
               t: aa:='ala';
               c: aa:='ala';
               a: aa:='ala';
               g: aa:='ala';
            end;
            a: case b2 of
               t: aa:='asp';
               c: aa:='asp';
               a: aa:='glu';
               g: aa:='glu';
            end;
            g: case b2 of
               t: aa:='gly';
               c: aa:='gly';
               a: aa:='gly';
               g: aa:='gly';
            end;
         end;
      end;
end;
(* end module amino.getaminoacid version = 'delmod 6.51 85 apr 17 tds/gds' *)


function aamw(aa: aminoacid): real;
(* return the molecular weight of the amino acid.
   data from crc handbook of chemistry and physics 60th ed, 1980
note: 1 was subtracted from values for asp and glu since they lose their
hydrogen near ph7. *)
begin (* aamw *)
     if aa = 'gly' then aamw := 75.07
else if aa = 'ala' then aamw := 89.10
else if aa = 'val' then aamw := 117.15
else if aa = 'leu' then aamw := 131.18
else if aa = 'ile' then aamw := 131.18 (* not in crc, assumed to be like leu *)
else if aa = 'ser' then aamw := 105.10
else if aa = 'thr' then aamw := 119.12
else if aa = 'cys' then aamw := 121.16
else if aa = 'met' then aamw := 149.21
else if aa = 'asp' then aamw := 132.11 (* see note above *)
else if aa = 'asn' then aamw := 132.13
else if aa = 'glu' then aamw := 146.13
else if aa = 'gln' then aamw := 146.15
else if aa = 'lys' then aamw := 146.19
else if aa = 'arg' then aamw := 174.20
else if aa = 'his' then aamw := 155.16
else if aa = 'phe' then aamw := 165.19
else if aa = 'tyr' then aamw := 181.19
else if aa = 'trp' then aamw := 204.23
else if aa = 'pro' then aamw := 115.13
else if aa = 'amb' then aamw := 0
else if aa = 'och' then aamw := 0
else if aa = 'opa' then aamw := 0
else if aa = '   ' then aamw := 0
end; (* aamw *)

function aaatoms(aa: aminoacid): integer;
(* return the number of atoms in amino acid.  Data from page 54 of:
   @book{Darnell1986,
   author = "J. Darnell
    and H. Lodish
    and D. Baltimore",
   title = "Molecular Cell Biology",
   year = "1986",
   publisher = "Scientific American Books, Inc.",
   address = "N. Y."}
note: 1 was subtracted from values for asp and glu since they lose their
hydrogen near ph7. *)
const
   backbone = 9; (* number of atoms in the amino acid without the side group *)
begin (* aaatoms *)
     if aa = 'gly' then aaatoms :=  1 + backbone
else if aa = 'ala' then aaatoms :=  4 + backbone
else if aa = 'val' then aaatoms := 10 + backbone
else if aa = 'leu' then aaatoms := 13 + backbone
else if aa = 'ile' then aaatoms := 13 + backbone
else if aa = 'ser' then aaatoms :=  5 + backbone
else if aa = 'thr' then aaatoms :=  8 + backbone
else if aa = 'cys' then aaatoms :=  5 + backbone
else if aa = 'met' then aaatoms := 11 + backbone
else if aa = 'asp' then aaatoms :=  6 + backbone
else if aa = 'asn' then aaatoms :=  8 + backbone
else if aa = 'glu' then aaatoms :=  9 + backbone
else if aa = 'gln' then aaatoms := 11 + backbone
else if aa = 'lys' then aaatoms := 16 + backbone
else if aa = 'arg' then aaatoms := 18 + backbone
else if aa = 'his' then aaatoms := 11 + backbone (* without the charged H *)
else if aa = 'phe' then aaatoms := 14 + backbone
else if aa = 'tyr' then aaatoms := 15 + backbone
else if aa = 'trp' then aaatoms := 18 + backbone
else if aa = 'pro' then aaatoms :=  8 + backbone
else if aa = 'amb' then aaatoms := 0
else if aa = 'och' then aaatoms := 0
else if aa = 'opa' then aaatoms := 0
else if aa = '   ' then aaatoms := 0
end; (* aamw *)

function peptideatoms(pie: pieceptr; position: integer): integer;
(* return the number of atoms in the peptide that starts
at internal coordinate position.  note that each peptide bond costs
one molecule of water... *)
const
      wateratoms = 3; (* number of atoms in a water molecule *)
var
      index: integer; (* location in pie *)
      totalatoms: integer; (* total number of atoms *)
      aa: aminoacid; (* the current aa *)
      atoms: integer; (* the number of atoms in an aa *)
      done: boolean; (* are we? *)
begin (* peptideatoms *)
      index:=position; (* start where we are told *)
      totalatoms:=0;
      done:=false;
      while not done do begin
         getaminoacid(index,pie,aa);
         atoms:=aaatoms(aa);
         if atoms > 0.0
         then begin
            (* remove an H2O if a bond is to form: *)
            if totalatoms > 0.0 then totalatoms:=totalatoms - wateratoms;
            (* form the bond: *)
            totalatoms := totalatoms + atoms;

            (* is it a circle? *)
            if pie^.key.piecon = circular then begin
               (* keep inside the circle *)
               if index > piecelength(pie)
                  then index := index - piecelength(pie);
               (* stop at once around: *)
               if index = position then done := true
            end
         end
         else done := true;
         index := index + 3; (* move a codon foreward *)
      end;
      peptideatoms := totalatoms
end; (* peptideatoms *)

procedure peptidemw(pie: pieceptr; position: integer;
                    var totalmw: real; (* total molecular weight *)
                    var aminoacids: integer);
(* Return the molecular weight of the peptide that starts
at internal coordinate position.  Note that each peptide bond costs
one molecule of water...  Also return the number of amino acids *)
const
      watermw = 18.01534; (* molecular weight of H2O from crc 60th ed. 1980 *)
var
      index: integer; (* location in pie *)
      aa: aminoacid; (* the current aa *)
      mw: real; (* the weight of aa *)
      done: boolean; (* are we? *)
begin (* peptidemw *)
      index:=position; (* start where we are told *)
      totalmw:=0;
      done:=false;
      aminoacids := 0;
      while not done do begin
         getaminoacid(index,pie,aa);
         mw:=aamw(aa);
         if mw > 0.0
         then begin
            (* remove an H2O if a bond is to form: *)
            if totalmw > 0.0 then totalmw:=totalmw - watermw;
            (* form the bond: *)
            totalmw := totalmw + mw;
            aminoacids := aminoacids + 1;

            (* is it a circle? *)
            if pie^.key.piecon = circular then begin
               (* keep inside the circle *)
               if index > piecelength(pie)
                  then index := index - piecelength(pie);
               (* stop at once around: *)
               if index = position then done := true
            end
         end
         else done := true;
         index := index + 3; (* move a codon foreward *)
      end;
end; (* peptidemw *)

procedure themain;
(* the main procedure of the program *)
const
   watermw = 18.01534; (* molecular weight of H2O from crc 60th ed. 1980 *)
   wateratoms = 3; (* number of atoms in a water molecule *)
var
   aminoacids: integer; (* number of amino acids in the chain *)
   mw: real; (* the molecular weight of a peptide *)
   atoms: integer; (* the number of atoms in a peptide *)
   totalmw: real; (* the total molecular weight of all peptides in the book *)
   totalatoms: integer; (* total number of atoms in all peptides in the book *)
   pie: pieceptr; (* the piece being examined *)
begin
   writeln(output,'pemowe ',version:4:2);
   brinit(book);
   new(pie);

   rewrite(list);
   writeln(list,' pemowe ',version:4:2,': peptide molecular weights');
   write(list,' ');
   copyaline(book,list);

   writeln(list);
   writeln(list,' only mw from the peptide starting',
                ' at piece beginning is given.');
   writeln(list,' fmet is not used for initial aug or gug codons.');
   writeln(list,' source: crc 60th ed. 1980.');
   writeln(list);

   totalmw := watermw; (* pre-add the water atoms *)
   totalatoms := wateratoms; (* pre-add the water atoms *)
   while not eof(book) do begin
      getpiece(book,pie);
      if not eof(book) then begin
         peptidemw(pie,1,mw,aminoacids);
         totalmw := totalmw + mw - watermw;

         atoms := peptideatoms(pie,1);
         totalatoms := totalatoms + atoms - wateratoms;

         write(list,' number ',number:4,
                    ' (',pie^.key.hea.keynam.letters,') ',
                    ' ',mw:8:2,' dalton');
         write(list,', ',atoms:8,' atoms ',aminoacids:1,' amino acids');
         writeln(list);
         clearpiece(pie)
      end
   end;
   writeln(list);

   writeln(list);
   write(list, ' ',totalmw:8:2,' dalton');
   write(list, ', ',totalatoms:8,' atoms');
   writeln(list);
end;
begin
   themain;
1: end.
