program rara(data, xyin, output);
(* rara: rank-rank reformulation of a data set

  Tom Schneider
  network address: toms@alum.mit.edu


  1993
 *)

label 1; (* end of program *)

const
(* begin module version *)
version = 1.02; (* of rara.p 1993 March 16
origin 1992 Mar 16 *)
(* end module version *)

(* begin module describe.rara *)
(*
name
   rara: rank-rank reformulation of a data set

synopsis
   rara(data: in, xyin: in, output: out)

files
   data: a data set with two columns.  '*' on the start of lines
      are comments copied to the output xyin file.
   xyin: Doubly sorted data.  The first two columns are the original two data
      columns.  The first data column is sorted.  The third column is the rank
      of the first column (1 to n, in order).  The forth column is the rank of
      the second column (1 to n but no longer in order).

   output: messages to the user

description
   To test data correlations but to make them insensitive to outliers, the data
   can be ranked and then graphed or the correlation coefficient found by
   xyplo.

   First, the data pairs are sorted on the second data column.  The second data
   column is then assigned ranks (1 to n).  The data are then sorted again on
   the first data column and the first data column is assigned ranks.  This
   leaves the first data column sorted.

examples

documentation

see also
   xyplo.p

author
   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.rara *)

(* begin module rara.const *)
   maxdata = 10000; (* the dataitems *)

   infofield = 12; (* size of field for printing information in bits *)
   infodecim = 6; (* number of decimal places for printing information *)
       (* these are used for conlist only *)
   nfield = 6; (* size of field for printing n, the number of sites *)

(* end module rara.const *)

type
   datatype = record
      sortdata: real; (* the data to sort the record on *)
      x, y: real; (* the raw data *)
      xrank, yrank: integer; (* the rank of the data *)
   end;

   (* position in the data: *)
   position = 0..maxdata;

   (* data to be sorted *)
   maptype =  array[1..maxdata] of datatype;

var
   data, xyin: text; (* a file used by this program *)
   map: maptype; (* the data to be sorted *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 4.11; (@ of prgmod.p 1991 Apr 22 *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 4.11; (@ of prgmod.p 1991 Apr 22 *)

(* begin module rara.sortstuff *)
function lessthan(a, b: position): boolean;
(* see quicksort *)
begin
   lessthan:=(map[a].sortdata < map[b].sortdata)
end;

procedure swap(a,b: position);
(* see quicksort *)
var
   hold: datatype;
begin
   hold:=map[a];
   map[a]:=map[b];
   map[b]:=hold
(*;write(output,'a=',a:1,', b=',b:1); print (@ for testing *)
end;
(* end module rara.sortstuff *)

(* begin module quicksort *)

procedure quicksort(left, right: position);
(* quick sort a list between positions left and right, into ascending order.
   a position is simply a scalar of the form 0..max.
   the array to be sorted is dimensioned 1..max.
(the difference in the ranges is important to the correct operation
of the sort...)
   two external routines are used:
      function lessthan(a, b: position): boolean is a generalized test for
          value-at-a < value-at-b.
      procedure swap(a, b: position) switches the items at positions a and b.
   since these routines are external, the procedure is general.

   this procedure taken from the book
   'algorithms + data structures = programs' by niklaus wirth,
   prentice-hall, inc., englewood cliffs, n.j.(1976), pp. 76-82 *)
var
      lower, upper: position; (* the positions looked at currently *)
      center: position; (* the rough center of the region being sorted *)
begin
      lower  := left;
      center := (left + right) div 2;
      upper  := right;

      repeat
         while lessthan(lower, center) do lower := succ(lower);
         while lessthan(center, upper) do upper := pred(upper);

         if lower <= upper then begin
            (* keep track of the center through the map: *)
                 if lower = center then center:=upper
            else if upper = center then center:=lower;
            swap(lower, upper);
            lower := succ(lower);
            upper := pred(upper)
         end
      until lower > upper;

      if  left < upper then quicksort(left,  upper);
      if lower < right then quicksort(lower, right)
end;

(* end module quicksort version = 4.11; (@ of prgmod.p 1991 Apr 22 *)


(* begin module rara.themain *)
procedure themain(var data, xyin: text);
(* the main procedure of the program *)
var
   number: position; (* number of data pairs read in *)
   p: position; (* current position in the data *)
begin
   writeln(output,'rara ',version:4:2);
   rewrite(xyin);
   writeln(xyin,'* rara ',version:4:2);

   number := 0;
   reset(data);
   while not eof(data) do begin
      if data^ = '*'
      then copyaline(data,xyin)
      else begin
         number := succ(number);
         readln(data, map[number].x,map[number].y);
      end
   end;

   (* sort on the second data item *)
   for p := 1 to number
   do map[p].sortdata := map[p].y;

   quicksort(1,number);

   for p := 1 to number do begin
       map[p].xrank := 0;
       map[p].yrank := p;
   end;

   (* prepare for the next sort on the first data item *)
   for p := 1 to number
   do map[p].sortdata := map[p].x;

   quicksort(1,number);

   for p := 1 to number do begin
       map[p].xrank := p;
   end;

   for p := 1 to number do begin
      write(xyin,' ',map[p].x:infofield:infodecim);
      write(xyin,' ',map[p].y:infofield:infodecim);
      write(xyin,' ',map[p].xrank:nfield);
      write(xyin,' ',map[p].yrank:nfield);
      writeln(xyin);
   end;

end;
(* end module rara.themain *)

begin
   themain(data, xyin);
1: end.
