program reform(fin, fout, input, output);
(* reformat raw sequences
   Thomas Schneider

module libraries: delman, delmods, prgmods *)

label 1; (* end of program *)

const
(* begin module version *)
version = 1.21; (* of reform 1994 sep 5
origin before 1982 *)
(* end module version *)

(* begin module describe.reform *)
(*
name
      reform: raw sequences reformatted

synopsis
      reform(fin: in, fout: out, input: intty, output: out)

files
      fin: the raw sequences to be reformatted.  the file must contain
         only the letters: 'a', 'c', 'g', 't', and/or 'u'.
      fout: the reformatted sequence.
      input: the user defines how to reformat the sequence:
         reformat - the sequence is only reformatted.
         invert - the order of the sequence stays the same, but the bases
            are complemented.
         complement - the order of the sequence is reversed and the bases
            are complemented.
         the user also specifies the number of bases to be printed on
         each line of fout.
      output: messages to the user.

description
      Reform allows one to type a file containing raw sequence data typed in
      whatever form is convenient, and to convert it into a form that
      the merge program can use to compare to a second typed copy.
      For example, when a sequence is to be entered from two strands,
      the sequence and its complement, one can enter both strands and then
      invert the second strand for comparison by merge.  Alternatively,
      the second strand could be entered backwards and the complement taken
      using this program.

see also
      merge.p

author
      Thomas Schneider

bugs
      The first typed line is ignored because of a problem with
      the standard input procedures on Unix.  Simply type a carriage
      return when the program starts up.

*)
(* end module describe.reform *)

      (* more constants *)
      maxseq = 1000; (* the maximum number of bases that can be held in
                      an array at once (the arrays are in a linked list, so
                      this does not limit the total number that can be
                      complemented at once, it only affects the efficiency. *)

(* begin module interact.const *)
      maxstring = 150; (* the maximum string *)
(* end module interact.const version = 'prgmod 3.96  85 mar 18 tds'; *)

type
(* begin module interact.type *)
      string = record (* a string of characters *)
         letters: array[1..maxstring] of char; (* the letters in the string *)
         length: integer; (* the number of characters in the string *)
         current: integer; (* the letter we are working on *)
      end;
(* end module interact.type version = 'prgmod 3.96  85 mar 18 tds'; *)

var
      fin, fout: text; (* sequence input and processed sequence *)
      operation: char; (* what operation to do (r,i,c) *)
      linelength: integer; (* length of lines in fout *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module package.interact *)
(* ************************************************************************ *)
(* begin module interact.prompt *)
procedure prompt(var afile: text);
(* prompt a file.
   the prompt is sent to the output file, and
   a line is read into the pascal line buffer.
   (for the cyber system this means to readln afile.)
   guarantee no bomb *)
begin (* prompt *)
      if eof(afile) then reset(afile);
      readln(afile);
end; (* prompt *)
(* end module interact.prompt version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module interact.readchar *)
procedure readchar(var afile:text; var ch: char);
(* read a character from afile, guarantee no bomb *)
begin (* readchar *)
      if eof(afile) then prompt(afile);
      read(afile,ch);
(*writeln(output,'"',ch,'"') *)
end; (* readchar *)
(* end module interact.readchar version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module interact.clearstring *)
procedure clearstring(var ribbon: string);
(* empty the string *)
var   index: integer; (* to the ribbon *)
begin (* clearstring *)
      with ribbon do begin
         for index := 1 to maxstring do letters[index] := ' ';
         length := 0;
         current := 0;
      end
end; (* clearstring *)
(* end module interact.clearstring version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module interact.readstring *)
procedure readstring(var afile: text; var line: string);
(* read in a string from afile, protect against bombing *)
var
      index: integer; (* for line *)
      cha: char; (* a character read in *)
      done: boolean; (* used for removing trailing blanks from the line *)
      acceptable: boolean; (* was the line typed short enough? *)
begin (* readstring *)
      with line do begin
         repeat
            clearstring(line);
            prompt(afile);
            index := 0; (* we now count characters *)
            while (not eoln(afile)) and (index < maxstring) do begin
               index:=succ(index);
               readchar(afile,cha);
               line.letters[index]:=cha
            end;
            if not eoln(afile)
            then begin
               writeln(output, 'type lines shorter than ', (maxstring+1): 1,
                               ' characters.  please retype the line...');
               acceptable := false
            end
            else acceptable := true
         until acceptable;

         length := index;
         if length > 0 then begin
            done := false;
            repeat (* remove blanks from the line.  note that a while loop
                  can not be used because one must avoid letters[0], since that
                  position does not exist... *)
               if letters[length] = ' '
               then length := pred(length)
               else done := true;

               if length = 0 then done := true
            until done
         end;

         if length > 0 then current := 1
                       else current := 0
      end
end; (* readstring *)
(* end module interact.readstring version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module interact.figurestring *)
procedure figurestring(
      var line: string; (* a string of characters to figure out *)
      var first: integer; (* first found non-blank character in the line *)
      var last:  integer; (* last character before a blank after first *)
      var whzat: char; (* what the token is *)
      var c:     char;  (* the first character of the token *)
      var i:     integer; (* integer value of token if it is integer; or 0 *)
      var r:     real); (* the real value if it is real; or 0.0 *)
(* figurestring figures out the tokens in a string.
 it recognizes words, integers, reals and poorly formed numbers.
 you can easily use it to parse lines.

our goal is to figure out what thing is on a string.
start looking at the current place on the line.
first and last are the first 'token' in line after start.
the current place is updated to the letter after last.
the thing found is described by the value of whzat:
      'c': character (when the token does not begin with a digit, '+', or '-')
      'i': integer
      'r': real
      ' ': blank line
      'g': garbage, cannot figure it out
and the value of the thing found is the appropriate variable *)
var
      numbers: set of '0'..'9';
      sign: integer; (* sign of a number *)
      numberstart: integer; (* the point a number starts,
                               beyond its sign, if any *)
      point: integer; (* location of decimal point *)
      power: integer; (* of 10 that a number is *)

      l: integer; (* an index for dissecting numbers *)

function figureinteger(first,last:integer):integer;
(* figure the integer in the token *)
var
      i: integer; (* index *)
      sum, increment: integer;
begin (* figureinteger *)
      power:=1; (* start at ones place *)
      sum:=0; (* start sum at zero *)
      for i:=last downto first do begin
         case line.letters[i] of
            '0': increment:=0;
            '1': increment:=1;
            '2': increment:=2;
            '3': increment:=3;
            '4': increment:=4;
            '5': increment:=5;
            '6': increment:=6;
            '7': increment:=7;
            '8': increment:=8;
            '9': increment:=9
         end;
         sum:=sum+power*increment;
         power:=power*10
      end;
      figureinteger:=sum
end; (* figureinteger *)

begin (* figurestring *)
      numbers:=['0','1','2','3','4','5','6','7','8','9'];
(* c:=' '; i:=0; r:=0.0; do not affect these variables unless necessary *)
      point:=0;
      whzat := '.'; (* assume that we have someting to work on *)
      (* now to see if that is true: *)

      with line do if (length = 0) or (current < 1) or (current > length)
      then whzat := ' '
      else begin
         (* figure out where the first token is in the line *)
         first:=line.current;
         while (line.letters[first]=' ') and (first < line.length)
            do first:=succ(first);
         if (first = line.length) and (line.letters[first] = ' ')
            then whzat := ' ';
      end;

      if whzat <> ' ' then begin
         last:=first;
         while (line.letters[last]<>' ') and (last < line.length)
            do last:=succ(last);
         if line.letters[last] = ' ' then last := pred(last);

         (* the token is between inclusive first and last *)

         c:=line.letters[first];
         if (c in numbers) or (c in ['+','-']) then begin
            if c in ['+','-'] then begin
               case c of
                  '+': sign:=+1;
                  '-': sign:=-1;
               end;
               numberstart:=succ(first)
            end
            else begin
               sign:=+1;
               numberstart:=first
            end;

            whzat:='i';
            for l:=numberstart to last do begin
               if not(line.letters[l] in numbers)
               then if line.letters[l]='.' (* we found a period *)
                  then if whzat='i' (* if so far it is numbers *)
                     then begin
                        whzat:='r'; (* it is actually real *)
                        point:=l
                     end
                     else whzat:='g' (* it is a second '.', ie garbage *)
                  else whzat:='g' (* it is garbage *)
            end;
            (* if it is only numbers, it is integer *)

            (* build number *)
            (* if it ends in a period, it is integer *)
            if (whzat = 'r') and (point = last) then whzat:='i';

            if whzat = 'i' then begin
               if point = last (* had an ending decimal point *)
               then i:=sign * figureinteger(numberstart,pred(last))
               else i:=sign * figureinteger(numberstart,last);
               r:=i
            end
            else if whzat = 'r' then begin
               i:=figureinteger(numberstart,point-1);
               r:=sign * (i+figureinteger(point+1,last)/power);
               i:=sign * i
            end
         end
         else begin
            whzat:='c';
         end;

         (* move the start to just beyond the last character of the token *)
         line.current:=succ(last)
      end
end; (* figurestring *)

(* end module interact.figurestring version = 'prgmod 3.96  85 mar 18 tds'; *)

(* ************************************************************************ *)
(* end module package.interact version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module interact.writestring *)
procedure writestring(var tofile: text; var s: string);
(* write the string s to file tofile, no writeln *)
var   i: integer; (* index to s *)
begin (* writestring *)
      with s do for i := 1 to length do write(tofile, letters[i])
end; (* writestring *)
(* end module interact.writestring version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module interact.flagstring *)
procedure flagstring(var afile: text; var buffer: string);
(* flag an error in the buffer at the current place, and clear the buffer *)
begin (* flagstring *)
      with buffer do length := current; (* chop off the rest of the buffer *)
      writestring(afile, buffer); (* show the buffer *)
      write(afile,'? ');
      clearstring(buffer)
end; (* flagstring *)
(* end module interact.flagstring version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module interact.nostring *)
function nostring(var buffer: string): boolean;
(* true if there are no characters in the rest of the buffer;
   false if there are characters.
   also, if there is no buffer, then buffer.length is set to 0 *)
var   answer: boolean; (* the answer returned *)
procedure kill;
(* destroy the line *)
begin (* kill *)
      answer := true; (* blood and gore *)
      clearstring(buffer) (* total death *)
end; (* kill *)
begin (* nostring *)
      with buffer do begin
         if length > 0 then begin
            if length < maxstring
            then while (letters[current] = ' ') and (current < length)
               do current := succ(current);

            if current <= maxstring
            then if letters[current] = ' '
                 then kill
                 else answer := false
            else kill
         end
         else kill
      end;
      nostring := answer
end; (* nostring *)
(* end module interact.nostring version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module interact.getchar *)
procedure getchar(var afile: text; var buffer: string;
                  var cha: char; var gotten: boolean);
(* get a character from the buffer, or refill the buffer and
   let the calling program figure out whether the buffer has non
   blank characters in it. *)
var
      (* variables for calling figurestring: *)
      first: integer;
      last: integer;
      what: char;
      int: integer;
      rea: real;
begin (* getchar *)
      if buffer.length = 0 then begin
         gotten:=false;
         readstring(afile,buffer)
      end
      else begin
         figurestring(buffer,first,last,what,cha,int,rea);
         gotten := (what <> ' ')
      end
end; (* getchar *)
(* end module interact.getchar version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module interact.getinteger *)
procedure getinteger(var afile: text; var buffer: string;
                     var int: integer; var gotten: boolean);
(* get the integer int from the buffer or interactive file afile *)
var
      (* variables for calling figurestring: *)
      first: integer;
      last: integer;
      what: char;
      cha: char;
      rea: real;
begin (* getinteger *)
      if buffer.length = 0
      then begin
         gotten:=false;
         readstring(afile,buffer)
      end
      else begin
         figurestring(buffer,first,last,what,cha,int,rea);

         if what <> 'i'
         then begin
            flagstring(output,buffer);
            writeln(output,' please type an integer');
            gotten:=false
         end
         else gotten:=true
      end
end; (* getinteger *)
(* end module interact.getinteger version = 'prgmod 3.96  85 mar 18 tds'; *)

procedure getcontrols(var operation: char; var linelength: integer);
(* read from the input file the control variables *)
var
      buffer: string; (* interactive buffered line from input *)
      gotten: boolean; (* has the desired object been obtained? *)
begin (* getcontrols *)
      clearstring(buffer);

      repeat
      if nostring(buffer)
         then writeln(output,'reformat, invert, complement (r/i/c)');
         getchar(input,buffer,operation,gotten);
         if gotten then if not (operation in ['r','i','c'])
            then begin
               gotten := false;
               clearstring(buffer);
               writeln(output,'type one of these letters: r i c')
            end
      until gotten;

      repeat
         if nostring(buffer)
         then writeln(output,'number of bases per line:');
         getinteger(input,buffer,linelength,gotten);

         if gotten
         then if linelength < 1 then begin
            writeln(output,'length must be positive');
            gotten := false
         end
      until gotten
end; (* getcontrols *)

procedure dooperation (var fin, fout: text;
                       operation: char; linelength: integer);
(* do the operation
      r reformat
      i invert
      c complement
on sequence in fin copied to fout with defined line length *)

type
      base = (a, c, g, t);

var
      count: integer; (* count of bases written to fout *)

(* the following three functions were stolen from module book.basis *)
function chartobase(ch:char):base;
(* convert a character into a base *)
begin
      case ch of
         'a': chartobase:=a;
         'c': chartobase:=c;
         'g': chartobase:=g;
         't': chartobase:=t
      end
end;

function basetochar(ba:base):char;
(* convert a base into a character *)
begin
      case ba of
         a: basetochar:='a';
         c: basetochar:='c';
         g: basetochar:='g';
         t: basetochar:='t';
      end
end;

function complement(ba:base):base;
(* take the complement of ba *)
begin
      case ba of
         a: complement:=t;
         c: complement:=g;
         g: complement:=c;
         t: complement:=a;
      end
end;

(* begin module findnonblank *)
procedure findnonblank(var afile: text; var ch: char);
(* find the next non blank character in a file, return it in ch. *)
begin
      ch:=' ';
      while (not eof(afile)) and
            (ch = ' ')
      do begin
         read(afile,ch);
         if eoln(afile) then readln(afile)
      end
end;
(* end module findnonblank version = 'delmod 6.51 85 apr 17 tds/gds' *)

function getbase(var b: base): boolean;
(* get the next base in fin.  if one was found, return true,
otherwise false *)
var   ch: char; (* a character found in fin *)
begin (* getbase *)
      findnonblank (fin, ch);
(*writeln(output,'"',ch,'"');*)

      if ch <> ' ' then begin (* found *)
         if ch = 'u' then ch := 't';
         if not(ch in ['a','c','g','t']) then begin
            writeln(output,'the character "',ch,'" is not allowed',
                           ' in a raw sequence file...');
            halt
         end;
         b:=chartobase(ch);
         getbase := true;
      end
      else getbase := false (* not found *)
end; (* getbase *)

procedure putbase(b: base);
(* put base b to fout as a character, with lines of linelength *)
begin (* putbase *)
      write(fout, basetochar(b));
      count := succ(count);
      if (count mod linelength) = 0 then writeln(fout)
end; (* putbase *)

procedure rawreformat;
(* reformat the raw sequence *)
var   b: base; (* the base currently delt with *)
begin (* rawreformat *)
      while not eof(fin)
         do if getbase(b)
            then putbase(b)
end; (* rawreformat *)

procedure rawinvert;
(* invert the raw sequence *)
var   b: base; (* the base currently delt with *)
begin (* rawinvert *)
      while not eof(fin)
         do if getbase(b)
            then putbase(complement(b))
end; (* rawinvert *)

procedure rawcomplement;
(* complement the raw sequence *)
type
      seqptr = ^seq;
      seq = record (* part of the complement built backwards *)
         bases: packed array[1..maxseq] of base;
         place: integer; (* where to start writing bases out *)
         next: seqptr
      end;

var
      root, newroot: seqptr; (* the root of the sequence in a linked list,
         and another pointer to build the list *)
      b: base; (* the base currently delt with *)
      index: integer; (* into root bases *)
begin (* rawcomplement *)
      new(root);
      with root^ do begin
         next := nil;
         place := maxseq;
      end;

      (* read the sequence into the root linked list -
         backwards with complement *)
      while not eof(fin) do begin
         if getbase(b) then begin
            b:=complement(b); (* do the complement *)

            (* check to see if this link is full *)
            if root^.place = 0 then begin
               root^.place := 1; (* the first base *)

               (* add on another link *)
               new(newroot);
               with newroot^ do begin
                  next := root;
                  place := maxseq
               end;
               root := newroot
            end;

            (* now put the base into the link *)
            with root^ do begin
               bases[place] := b; (* stuff base in *)
               place := pred(place) (* move backwards to unused spot *)
            end
         end
      end;

      (* be sure to point to the first base *)
      with root^ do place := succ(place);

      (* now write the root list out.
         (note that we lose the linked list into outer space,
          but who cares???) *)

      while root <> nil
         do with root^ do begin
            for index := place to maxseq
               do putbase(bases[index]);

               (* cast the fool link adrift: *)
               root := root^.next (* see ya later ya sucker... *)
         end
end; (* rawcomplement *)

begin (* dooperation *)
      count := 0;

      case operation of
         'r': rawreformat;
         'i': rawinvert;
         'c': rawcomplement
      end;

      (* close up the last line *)
      if (count mod linelength) <> 0 then writeln(fout);

      write(output,' ',count:1,' bases ');
      case operation of
         'r': writeln(output,'reformatted');
         'i': writeln(output,'inverted');
         'c': writeln(output,'complemented')
      end;
end; (* dooperation *)

begin (* reform *)
      writeln(output,'reform ',version:4:2);
      reset(fin);
      rewrite(fout);

      getcontrols(operation, linelength);
      dooperation(fin, fout, operation, linelength);
1: end. (* reform *)
