program stirling(output);
(* stirling: test of stirling's formula

  Tom Schneider
  network address: toms@alum.mit.edu


 *)

const
(* begin module version *)
version = 1.07; (* of stirling.p 1995 Feb 3
origin: 1988 June 24 *)
(* end module version *)

(* begin module describe.stirling *)
(*
name
   stirling: test of Stirling's formula

synopsis
   stirling(output: out)

files
   output: a table of Stirling's approximation

description

examples

documentation
   Stirling's approximation for factorial is compared to the exact factorial
   function.  The results can be plotted with xyplo.

see also
   xyplo.p

author
   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.stirling *)

(* begin module stirling.lnnf *)
function lnnf(n: integer):real;
(* calculate ln(n!) *)
var
  i: integer; (* counter *)
  s: real; (* running sum *)
begin
 s := 0;
 for i := 2 to n do  s := s +ln(i);
 lnnf := s
end;
(* end module stirling.lnnf *)

(* begin module stirling.stirling *)
function stirling(n: integer): real;
(* calculate n ln(n) - n *)
begin
 stirling := (n*ln(n)) - n
end;
(* end module stirling.stirling *)

function beststirling(n: integer): real;
(* better approx of ln(n!) *)
var
  pi: real; (* 3.14159... *)
begin
  pi := 4.0 * arctan(1.0);
  beststirling := (n + 1/2)*ln(n) - n + (1/2)* ln(2*pi);
end;

function reduce(a: real; n: integer): real;
(* reduce a by n *)
begin
  reduce := ( ln(n) - (a/n))/ln(2);
end;

(* begin module stirling.themain *)
procedure themain(var fout: text);
(* the main procedure of the program *)
var
  a: real; (* actual value of ln(n!) *)
  b: real; (* best approximation of ln(n!) *)
  s: real; (* stirling's approximation to ln(n!) *)
  n: integer; (* counter of the tests *)
begin
  writeln(output,'stirling ',version:4:2);

  writeln(fout,'* Stirling''s approximation ',version:4:2);
  writeln(fout,'* 3 columns: n, flag, value');
  writeln(fout,'* flag = a means ln(n!)');
  writeln(fout,'* flag = b means (n + 1/2)*ln(n) - n + (1/2)* ln(2*pi)');
  writeln(fout,'* flag = s means n*ln(n)-n (the approximation)');
  writeln(fout,'* flag = d means ln(n!) - [n*ln(n)-n]');

  for n:=1 to 53 do begin

     a := lnnf(n);
     b := beststirling(n);
     s := stirling(n);

     write(fout,n:5);
     write(fout,' a', a:15:5);
     writeln(fout);

     write(fout,n:5);
     write(fout,' b', b:15:5);
     writeln(fout);

     write(fout,n:5);
     write(fout,' s', s:15:5);
     writeln(fout);

     write(fout,n:5);
     write(fout,' d', (a-s):15:5);
     writeln(fout);

  end;

(*
  repeat
     writeln(output);
     write(output,'n: ');
     read(input,n);
     if n > 0 then begin
       a := lnnf(n);
       c := beststirling(n);
       s := stirling(n);

       writeln(output,'                             ln(n!) = ',a:10:5);
       writeln(output,'Complex Stirling''s approx to ln(n!) = ',c:10:5);
       writeln(output,' Simple Stirling''s approx to ln(n!) = ',s:10:5);

       writeln(output);
       writeln(output,
        '                             log2(n) - log2(n!)/n = ',reduce(a,n):10:5);
       writeln(output,
       'Complex Stirling''s approx to log2(n) - log2(n!)/n = ',reduce(c,n):10:5);
       writeln(output,
       ' Simple Stirling''s approx to log2(n) - log2(n!)/n = ',reduce(s,n):10:5);

    end
  until n = 0;
*)


end;
(* end module stirling.themain *)

begin
   themain(output);
end.
