program thethepc(input, output);
(* thethepc: find duplicate words in a file
   Tom Schneider, 1989 *)

label 1; (* end of program *)

const
(* begin module version *)
version = 2.01; (* of thethepc.p 1996 July 25
name changed to thethepc: 1996 July 25
last significant change: 1989 April 14
origin 1989 April 10 *)
(* end module version *)

(* begin module describe.thethepc *)
(*
name
   thethepc: find duplicate words in a file

synopsis
   thethepc(input: in, output: out)

files
   input:  text of a paper in which one might have put a
      a word in duplicate, as these two lines demonstrate!
   output: messages to the user, the program name and version number,
      followed by a list of the duplicate words (if any).

description
   The program searches through the input stream for duplicate tokens and shows
   them on the output.  A token is separtated from other tokens by spaces or
   carriage returns.  The purpose of the program is to find duplicate words
   in a paper.  For this reason, the program's normal mode is to ignore
   words with special characters (anything other than a-zA-Z).  This can be
   altered by changing constants (see technical notes below).

examples
   This line contains a duplicate word word that should be picked up.

author
   Thomas D. Schneider
   Frederick, Maryland 21701
   toms@alum.mit.edu

bugs
   Does not handle arbitrarily long lines.

technical notes
   There are two boolean constants that may be set to adjust how the program
   works:  onlycharacters and caseindependent.  See the comments on their
   definitions The maximum buffer length is given by constant maxstring.

   Note: the name of this program is thethepc because there is a (better) c
   version, and I don't want to get them mixed up.

*)
(* end module describe.thethepc *)

(* more constants *)
   onlycharacters = true;  (* if true then only characters a-z allowed
      in the tokens picked up *)
   caseindependent = true; (* if true, then case is ignored *)

(* begin module interact.const *)
      maxstring = 1024; (* the maximum string *)
(* end module interact.const version = 'prgmod 3.98  86 nov 11 tds'; *)

type
(* begin module interact.type *)
      string = record (* a string of characters *)
         letters: array[1..maxstring] of char; (* the letters in the string *)
         length: integer; (* the number of characters in the string *)
         current: integer; (* the letter we are working on *)
      end;
(* end module interact.type version = 'prgmod 3.98  86 nov 11 tds'; *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'prgmod 3.98  86 nov 11 tds'; *)

(* begin module package.interact *)
(* ************************************************************************ *)
(* begin module interact.prompt *)
procedure prompt(var afile: text);
(* prompt a file.
   the prompt is sent to the output file, and
   a line is read into the pascal line buffer.
   (for the cyber system this means to readln afile.)
   guarantee no bomb *)
begin (* prompt *)
      if eof(afile) then reset(afile);
      readln(afile);
end; (* prompt *)
(* end module interact.prompt version = 'prgmod 3.98  86 nov 11 tds'; *)

(* begin module interact.readchar *)
procedure readchar(var afile:text; var ch: char);
(* read a character from afile, guarantee no bomb *)
begin (* readchar *)
      if eof(afile) then prompt(afile);
      read(afile,ch);
(*writeln(output,'"',ch,'"') *)
end; (* readchar *)
(* end module interact.readchar version = 'prgmod 3.98  86 nov 11 tds'; *)

(* begin module interact.clearstring *)
procedure clearstring(var ribbon: string);
(* empty the string *)
var   index: integer; (* to the ribbon *)
begin (* clearstring *)
      with ribbon do begin
         for index := 1 to maxstring do letters[index] := ' ';
         length := 0;
         current := 0;
      end
end; (* clearstring *)
(* end module interact.clearstring version = 'prgmod 3.98  86 nov 11 tds'; *)

(* begin module interact.readstring *)
procedure readstring(var afile: text; var line: string);
(* read in a string from afile, protect against bombing *)
var
      index: integer; (* for line *)
      cha: char; (* a character read in *)
      done: boolean; (* used for removing trailing blanks from the line *)
      acceptable: boolean; (* was the line typed short enough? *)
begin (* readstring *)
      with line do begin
         repeat
            clearstring(line);
            prompt(afile);
            index := 0; (* we now count characters *)
            while (not eoln(afile)) and (index < maxstring) do begin
               index:=succ(index);
               readchar(afile,cha);
               line.letters[index]:=cha
            end;
            if not eoln(afile)
            then begin
               writeln(output, 'type lines shorter than ', (maxstring+1): 1,
                               ' characters.  please retype the line...');
               acceptable := false
            end
            else acceptable := true
         until acceptable;

         length := index;
         if length > 0 then begin
            done := false;
            repeat (* remove blanks from the line.  note that a while loop
                  can not be used because one must avoid letters[0], since that
                  position does not exist... *)
               if letters[length] = ' '
               then length := pred(length)
               else done := true;

               if length = 0 then done := true
            until done
         end;

         if length > 0 then current := 1
                       else current := 0
      end
end; (* readstring *)
(* end module interact.readstring version = 'prgmod 3.98  86 nov 11 tds'; *)

(* begin module interact.figurestring *)
procedure figurestring(
      var line: string; (* a string of characters to figure out *)
      var first: integer; (* first found non-blank character in the line *)
      var last:  integer; (* last character before a blank after first *)
      var whzat: char; (* what the token is *)
      var c:     char;  (* the first character of the token *)
      var i:     integer; (* integer value of token if it is integer; or 0 *)
      var r:     real); (* the real value if it is real; or 0.0 *)
(* figurestring figures out the tokens in a string.
 it recognizes words, integers, reals and poorly formed numbers.
 you can easily use it to parse lines.

our goal is to figure out what thing is on a string.
start looking at the current place on the line.
first and last are the first 'token' in line after start.
the current place is updated to the letter after last.
the thing found is described by the value of whzat:
      'c': character (when the token does not begin with a digit, '+', or '-')
      'i': integer
      'r': real
      ' ': blank line
      'g': garbage, cannot figure it out
and the value of the thing found is the appropriate variable *)
var
      numbers: set of '0'..'9';
      sign: integer; (* sign of a number *)
      numberstart: integer; (* the point a number starts,
                               beyond its sign, if any *)
      point: integer; (* location of decimal point *)
      power: integer; (* of 10 that a number is *)

      l: integer; (* an index for dissecting numbers *)

function figureinteger(first,last:integer):integer;
(* figure the integer in the token *)
var
      i: integer; (* index *)
      sum, increment: integer;
begin (* figureinteger *)
      power:=1; (* start at ones place *)
      sum:=0; (* start sum at zero *)
      for i:=last downto first do begin
         case line.letters[i] of
            '0': increment:=0;
            '1': increment:=1;
            '2': increment:=2;
            '3': increment:=3;
            '4': increment:=4;
            '5': increment:=5;
            '6': increment:=6;
            '7': increment:=7;
            '8': increment:=8;
            '9': increment:=9
         end;
         sum:=sum+power*increment;
         power:=power*10
      end;
      figureinteger:=sum
end; (* figureinteger *)

begin (* figurestring *)
      numbers:=['0','1','2','3','4','5','6','7','8','9'];
(* c:=' '; i:=0; r:=0.0; do not affect these variables unless necessary *)
      point:=0;
      whzat := '.'; (* assume that we have someting to work on *)
      (* now to see if that is true: *)

      with line do if (length = 0) or (current < 1) or (current > length)
      then whzat := ' '
      else begin
         (* figure out where the first token is in the line *)
         first:=line.current;
         while (line.letters[first]=' ') and (first < line.length)
            do first:=succ(first);
         if (first = line.length) and (line.letters[first] = ' ')
            then whzat := ' ';
      end;

      if whzat <> ' ' then begin
         last:=first;
         while (line.letters[last]<>' ') and (last < line.length)
            do last:=succ(last);
         if line.letters[last] = ' ' then last := pred(last);

         (* the token is between inclusive first and last *)

         c:=line.letters[first];
         if (c in numbers) or (c in ['+','-']) then begin
            if c in ['+','-'] then begin
               case c of
                  '+': sign:=+1;
                  '-': sign:=-1;
               end;
               numberstart:=succ(first)
            end
            else begin
               sign:=+1;
               numberstart:=first
            end;

            whzat:='i';
            for l:=numberstart to last do begin
               if not(line.letters[l] in numbers)
               then if line.letters[l]='.' (* we found a period *)
                  then if whzat='i' (* if so far it is numbers *)
                     then begin
                        whzat:='r'; (* it is actually real *)
                        point:=l
                     end
                     else whzat:='g' (* it is a second '.', ie garbage *)
                  else whzat:='g' (* it is garbage *)
            end;
            (* if it is only numbers, it is integer *)

            (* build number *)
            (* if it ends in a period, it is integer *)
            if (whzat = 'r') and (point = last) then whzat:='i';

            if whzat = 'i' then begin
               if point = last (* had an ending decimal point *)
               then i:=sign * figureinteger(numberstart,pred(last))
               else i:=sign * figureinteger(numberstart,last);
               r:=i
            end
            else if whzat = 'r' then begin
               i:=figureinteger(numberstart,point-1);
               r:=sign * (i+figureinteger(point+1,last)/power);
               i:=sign * i
            end
         end
         else begin
            whzat:='c';
         end;

         (* move the start to just beyond the last character of the token *)
         line.current:=succ(last)
      end
end; (* figurestring *)

(* end module interact.figurestring version = 'prgmod 3.98  86 nov 11 tds'; *)

(* ************************************************************************ *)
(* end module package.interact version = 'prgmod 3.98  86 nov 11 tds'; *)

(* begin module interact.token *)
procedure token(var buffer, atoken: string; var gotten: boolean);
(* get a token from the buffer *)
var
      (* variables for calling figurestring: *)
      first: integer;
      last: integer;
      what: char;
      cha: char;
      int: integer;
      rea: real;
      index: integer; (* to the buffer *)
begin
      figurestring(buffer,first,last,what,cha,int,rea);

      if what = ' '
      then gotten := false
      else begin
         clearstring(atoken);
         for index := first to last
            do atoken.letters[index-first+1] := buffer.letters[index];
         atoken.length := last - first + 1;
         atoken.current := 1;

         gotten:=true
      end
end;
(* end module interact.token version = 'prgmod 3.98  86 nov 11 tds'; *)
 
(* begin module interact.gettoken *)
procedure gettoken(var afile: text; var buffer: string;
                   var atoken: string; var gotten: boolean);
(* get a token from the buffer or interactive file afile *)
begin (* gettoken *)
      if buffer.length = 0
      then begin
         gotten:=false;
         readstring(afile,buffer)
      end
      else token(buffer,atoken,gotten)
end; (* gettoken *)
(* end module interact.gettoken version = 'prgmod 3.98  86 nov 11 tds'; *)

(* begin module interact.writestring *)
procedure writestring(var tofile: text; var s: string);
(* write the string s to file tofile, no writeln *)
var   i: integer; (* index to s *)
begin (* writestring *)
      with s do for i := 1 to length do write(tofile, letters[i])
end; (* writestring *)
(* end module interact.writestring version = 'prgmod 3.98  86 nov 11 tds'; *)

(* begin module interact.getstring *)
procedure getstring(var afile: text; var buffer: string;
                    var gotten: boolean);
(* get a string from a file not using string calls.  this lets one
obtain lines from a file without interactive prompts *)
var   index: integer; (* of buffer *)
begin (* getstring *)
      clearstring(buffer);
      if eof(afile)
      then gotten := false
      else begin
         index := 0;
         while (not eoln(afile)) and (index < maxstring) do begin
            index := succ(index);
            read(afile, buffer.letters[index])
         end;

         if not eoln(afile) then begin
            writeln(output, ' getstring: a line exceeds maximum string size (',
                            maxstring:1,')');
            halt
         end;

         buffer.length := index;
         buffer.current := 1;
         readln(afile);
         gotten := true
      end
end; (* getstring *)
(* end module interact.getstring *)

(* begin module equalstring *)
function equalstring(a, b: string): boolean;
(* test for equality between two strings at current positions *)
var   index: integer; (* index to both strings *)
      equal: boolean; (* are letters in a and b the same? *)
begin (* equalstring *)
      if a.length = b.length
      then begin
         index := 1;
         repeat
            equal := (a.letters[index] = b.letters[index]);
            index := succ(index)
         until (not equal) or (index > a.length);
         equalstring := equal
      end
      else equalstring := false
end; (* equalstring *)
(* end module equalstring *)

(* begin module decapitilize *)
procedure decapitilize(var c: char); 
(* convert the character c to lower case *) 
var 
   n: integer; (* c is the n'th letter of the alphabet *)
begin 
   n := ord(c); 

   if (n >= ord('A')) and 
      (n <= ord('Z'))   
   then c := chr( n - ord('A') + ord('a')) 
end; 
(* end module decapitilize *)

(* begin module decapstring *)
procedure decapstring(var t: string);
(* convert capital letters in t to lower case *)
var
   i: integer; (* index to position in t *)
begin
   with t do for i := 1 to length do decapitilize(letters[i]);
end;
(* end module decapstring *)

(* begin module alphabetic *)
function alphabetic(var c: char): boolean;
(* test that character c is alphabetic (either case) *)
var
   n: integer; (* c is the n'th letter of the alphabet *)
begin
   n := ord(c);

   if ( (n >= ord('A')) and (n <= ord('Z')) )
      or
      ( (n >= ord('a')) and (n <= ord('z')) )
   then alphabetic := true
   else alphabetic := false
end;
(* end module alphabetic *)

(* begin module whatisthere *)
procedure whatisthere(t: string; var charsok: boolean);
(* what characters are in the string t?  If it contains non-alphabetic
characters, charsok is false. *)
var
   i: integer; (* index to position in t *)
begin
   charsok := true;
   with t do for i := 1 to length
   do charsok := charsok and alphabetic(letters[i]);

   (* This works because any false case (ie, non alphabetic) will cause the
   entire thing to be false *)
end;
(* end module whatisthere *)

(* begin module thethepc.themain *)
procedure themain(var infile, outfile: text);
(* the main procedure of the program *)
var
   gotten: boolean; (* true if the buffer line was found *)
   no: integer; (* current line number in the file *)
   lasttokenline: integer; (* the line number of the previous token *)

   (* true means string bufferone contains the most recent line,
      otherwise buffertwo has it. *)
   bonefirst: boolean;
   bufferone: string; (* a buffer read from the input file *)
   buffertwo: string; (* the other buffer read from the input file *)

   (* true means string tokenone contains the most recent token,
      otherwise tokentwo has it. *)
   tonefirst: boolean;
   tokenone: string; (* a buffers containing a token *)
   tokentwo: string; (* a buffers containing a token *)

   charsok: boolean; (* true means that the characters in the string
      are acceptable for showing the string to the person.  This allows
      the program to filter cases like 'end;' out because it 
      contains a non-alphabetic character *)

procedure comparetokens(var buffer: string);
(* tokens from the line buffer *)
begin
   getstring(infile, buffer, gotten);

   if gotten then begin
      buffer.current := 1;

      while (buffer.current <= buffer.length) and (gotten = true) do begin
         if tonefirst
         then begin
            gettoken(infile, buffer, tokenone, gotten);
            if caseindependent then decapstring(tokenone);
            if onlycharacters then whatisthere(tokenone,charsok);
         end
         else begin
            gettoken(infile, buffer, tokentwo, gotten);
            if caseindependent then decapstring(tokentwo);
            if onlycharacters then whatisthere(tokentwo,charsok);
         end;

         if gotten then begin
(*
            write(outfile,'lead: (');
            if tonefirst then writestring(outfile, tokenone)
                         else writestring(outfile, tokentwo);

            write(outfile,') follow: (');
            if tonefirst then writestring(outfile, tokentwo)
                         else writestring(outfile, tokenone);

            writeln(outfile,')');
*)

            if equalstring(tokenone,tokentwo) and charsok then begin
               writeln(outfile,'------------');
               write(outfile,'duplicated token: "');
               writestring(outfile,tokenone);
               writeln(outfile,'"');

               if no <> lasttokenline then begin
                  writeln(outfile,'Found across lines ',
                                  lasttokenline:1,'-',no:1,':');
                  if bonefirst then writestring(outfile, buffertwo)
                               else writestring(outfile, bufferone);
                  writeln(outfile);
                  if bonefirst then writestring(outfile, bufferone)
                               else writestring(outfile, buffertwo);
                  writeln(outfile);
                  if lasttokenline < no - 1
                  then writeln(outfile,'(only the last two lines are shown)');
               end
               else begin
                  writeln(outfile,'Found in line ',no:1,':');
                  writestring(outfile, buffer);
                  writeln(outfile);
               end;
            end;

            tonefirst := not tonefirst;
            lasttokenline := no;
         end;
      end;
   end;
end;

begin
   writeln(outfile,'thethepc ',version:4:2);
   bonefirst := true;
   tonefirst := true;
   no := 0;
   lasttokenline := no;
   charsok := true;

   while not eof(infile) do begin
      no := no + 1;

      if bonefirst
      then comparetokens(bufferone)
      else comparetokens(buffertwo);

      bonefirst := not bonefirst
   end;
end;
(* end module thethepc.themain *)

begin
   themain(input,output);
1: end.
