program tkod(input, xyin, xyplop, output);
(* tkod: read od values from tk data
   tom schneider *)

label 1;  (* end of program *)

const
(* begin module version *)
version = 1.13;  (* of tkod.p 1995 July 26
origin 1987 august 3 *)
(* end module version *)

(* begin module describe.tkod *)
(*
name
   tkod: read od values from tk data

files
   tkod(input: in, xyin: out, xyplop: out, output: out)

synopsis
   input: output of the tk program; containing a header line,
      and a series of plates that describe the names of wells,
      their optical densities at various times and od620 data.
   xyin: the data from the input rearranged for xyplo
   xyplop: control file for xyplo
   output: messages to the user.

description
   Tkod takes OD620 data from 96 well plates as provided by the
   tk program and converts the data into a form that the xyplo
   program can use to plot with.

see also
   titer.p
   beta-galactosidase assay protocol
   tk program (basic)
   xyplo.p

author
   Tom Schneider

(* end module describe.tkod *)

var
   xyin, xyplop: text; (* files to control the graph *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text); 
(* copy a line from file fin to file fout *)
begin (* copyaline *) 
      while not eoln(fin) do begin  
         fout^ := fin^;   
         put(fout);   
         get(fin) 
      end;  
      readln(fin);  
      writeln(fout);  
end; (* copyaline *)
(* end module copyaline version = "delmod 6.21 84 may 19 tds/gds"; *) 

(* begin module tkod.header *)
procedure header(var infile, xyin: text);
(* handle the header of the xyin, copy from input *)
var x: integer; (* index to the lines copied *)
begin
   rewrite(xyin);
   writeln(xyin,'* tkod ',version:4:2);
   for x:=1 to 3 do begin
     write(xyin,'* ');
     copyaline(infile, xyin);
   end;
   writeln(xyin,'* data are:');
   writeln(xyin,'* name, time (sec), od, exp(time/3600)');

end;
(* end module tkod.header *)

(* begin module tkod.rowcol *)
procedure rowcol(var outfile: text; row, column: integer);
(* print out to file outfile the row and column as a alphanumeric.
Thus the first row and column are a1, and the last is h12 *)
begin
   case column of
      1: write(outfile,'a');
      2: write(outfile,'b');
      3: write(outfile,'c');
      4: write(outfile,'d');
      5: write(outfile,'e');
      6: write(outfile,'f');
      7: write(outfile,'g');
      8: write(outfile,'h');
   end;
   write(outfile,row:1)
end;
(* end module tkod.rowcol *)


(* begin module tkod.movedata *)
procedure movedata(var infile, outfile: text);
(* move data from infile to xyin *)
var
   c: char; (* a character for moving around infile *)
   column: integer; (* current column *)
   datum: real; (* a bit of data from the input file *)
   etime: real; (* the exponential of time expressed in hours,
                   this allows correct regression of the data *)
   ro: integer; (* the current row according to the data *)
   row: integer; (* current row *)
   time: integer; (* time of the data from start of experiment (seconds) *)
begin
   (* skip lines *)
   readln(infile);
   readln(infile);

   (* skip '*' character and grab time in seconds from start *)
   readln(infile,c,time);

   (* convert time to hours by dividing by 3600, then take exponent
      to allow regression *)
   etime := exp(time/3600);

   (* skip a to h descriptive line *)
   readln(infile);
   
   (* read data and dump out to xyin *)
   for row := 1 to 12 do begin
      read(infile, ro);
      if ro <> row then begin
         writeln(output,' bad row number: ',ro:1,
                        ' seen where ',row:1,' expected');
         halt
      end;

      for column := 1 to 8 do begin
         read(infile, datum);

         (* if row < 10 then write(outfile,' '); (* make names line up *)
         rowcol(outfile,row,column);

         writeln(outfile,' ', time:10,' ',datum:6:3,' ',etime:10:8)
      end;
      readln(infile)
   end;

   (* skip lines *)
   readln(infile);
end;
(* end module tkod.movedata *)

(* begin module tkod.makexyplop *)
procedure makexyplop(var xyplop: text);
(* make the parameter file for xyplo *)
var
   column: integer; (* current column *)
   row: integer; (* current row *)
begin
   rewrite(xyplop);
   writeln(xyplop,' x 0 5000  zx min max         ');
   writeln(xyplop,' y -0.1 0.1 zy min max        ');
   writeln(xyplop,'10 10     xinterval yinterval ');
   writeln(xyplop,'4 4       xwidth    ywidth    ');
   writeln(xyplop,'0 3       xdecimal  ydecimal  ');
   writeln(xyplop,'6 8       xsize     ysize     ');
   writeln(xyplop,'exp(time) (hours)');
   writeln(xyplop,'OD 620');
   writeln(xyplop,'n         zc                  ');
   writeln(xyplop,'n 2       zxl base            ');
   writeln(xyplop,'n 2       zyl base            ');
   writeln(xyplop,'          --------------------');
   writeln(xyplop,'4 3       xcolumn   ycolumn   ');
   writeln(xyplop,'1         symbol column       ');
   writeln(xyplop,'0  0      xscolumn  yscolumn  ');
   writeln(xyplop,'          --------------------');

   for row := 1 to 12 do begin
      for column := 1 to 8 do begin
         writeln(xyplop,'f         symbol to plot      ');
         rowcol(xyplop,row,column);
         writeln(xyplop,'        symbol flag         ');
         writeln(xyplop,'0.05      symbol sizex        ');
         writeln(xyplop,'0.05      symbol sizey        ');
         writeln(xyplop,'nl 0.05   no connection (examp');
         writeln(xyplop,'l 0.05    linetype  size      ');
         writeln(xyplop,'          --------------------');
      end;
   end;
   writeln(xyplop,'.');
   writeln(xyplop,'          --------------------');
end;
(* end module tkod.makexyplop *)

(* begin module themain *)
procedure themain(var xyin, xyplop: text);
(* the main procedure of tkod *)
begin
   writeln(output,'tkod ',version:4:2);

   header(input, xyin);
   while not eof(input) do movedata(input, xyin);

   makexyplop(xyplop);
end;
(* end module themain *)

begin
   themain(xyin, xyplop);
1: end.
