program transpose(fin, transposep, fout, output);
(* transpose: transpose a matrix of integers

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  toms@alum.mit.edu (permanent)
  https://alum.mit.edu/www/toms (permanent)

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.05; (* of transpose.p 2012 Mar 14
2012 Mar 14, 1.05: handle 0 decimals
2010 Nov 02, 1.04: transposep takes width, decimals
2010 Nov 02, 1.03: fix documentation
2007 Oct 13, 1.02: match freb better
2007 Oct 12, 1.01: functional but not quite
2007 Oct 12, 1.00: origin
*)
updateversion = 1.05; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.transpose *)
(*
name
   transpose: transpose a matrix of integers

synopsis
   transpose(fin: in, transposep: in, fout: out, output: out)

files

   fin: A file containing five lines.
      The first line is the name of the site.  The remaining four
      lines are the numbers of A, C, G and T
      at each position in an aligned set of sequences.  Words
      in the file that do not begin with a number are dropped

   fout:  The transpose of the file with numbering in the first colum.
      This output is ready for conversion to a Delila book by the freb
      program.

   transposep:  parameters to control the program.  The file must contain the
      following parameters, one per line:

      parameterversion: The version number of the program.  This allows the
         user to be warned if an old parameter file is used.

      width (integer): width of output numbers

      decimal (integer): decimal places of output numbers

      zerobase (integer): base that should have the zero coordinate

   output: messages to the user

description

   Convert a table of numbers given as four rows into four columns.

examples

documentation

see also

   freb.p

author

   Thomas Dana Schneider

bugs

technical notes

   The largest number of items in a row is given by constant maxstore.

*)
(* end module describe.transpose *)

   maxstore = 1000; (* maximum storage *)

var
   fin, (* file used by this program *)
   transposep, (* file used by this program *)
   fout: text; (* file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 5.38; (@ of prgmod.p 2007 May 15 *)

(* begin module skipblanks *)

(* 2003 July 31: tab is considered a blank character *)

function isblank(c: char): boolean;
(* is the character c blank or tab? *)
const
   tab = 9; (* tab character *)
begin
   isblank := (c = ' ') or (ord(c) = tab)
end;

procedure skipblanks(var thefile: text);
(* skip over blanks until a non-blank, or end of line, is found *)
begin
   while isblank(thefile^) and not eoln(thefile) do get(thefile);
end;

procedure skipnonblanks(var thefile: text);
(* skip over nonblanks until a blank, or end of line, is found *)
begin
   while (not isblank(thefile^)) and not eoln(thefile) do get(thefile);
end;

procedure skipcolumn(var thefile: text);
(* skip over a data column *)
begin
   skipblanks(thefile); skipnonblanks(thefile)
end;
(* end module skipblanks version = 5.38; (@ of prgmod.p 2007 May 15 *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 5.38; (@ of prgmod.p 2007 May 15 *)

(* begin module transpose.themain *)
procedure themain(var fin, transposep, fout: text);
(* the main procedure of the program *)
var
   parameterversion: real; (* parameter version number *)
   store: array[1..4, 1..maxstore] of real;
   base: integer; (* count of the 4 bases *)
   count: integer; (* the number of initial columns read *)
   c: integer; (* counter for output *)
   wid: integer; (* width of output numbers *)
   dec: integer; (* decimal places of output numbers *)
   zerobase: integer; (* base that should have the zero coordinate *)

procedure writeout(var f: text);
(* write out the data *)
var
   base: integer; (* count of the 4 bases *)
   c: integer; (* counter for output *)
begin
   (* write out the data *)
   for c := 1 to count do begin
      write(f, (c-zerobase):4);
      for base := 1 to 4 do begin
         if dec > 0
         then write(f, ' ', store[base,c]:wid:dec)
         else write(f, ' ', round(store[base,c]):wid);
      end;
      writeln(f);
   end;
end; (* writeout *)

begin
   writeln(output,'transpose ',version:4:2);

   reset(transposep);
   readln(transposep, parameterversion);
   if round(100*parameterversion) < round(100*updateversion) then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;

   readln(transposep, wid);
   readln(transposep, dec);
   readln(transposep, zerobase);

   reset(fin);
   rewrite(fout);

   (* copy the name line *)
   copyaline(fin, fout);

   (* read in the data *)
   writeln(output,'Reading data. "+" = accepted number, "-" = not a number');
   for base := 1 to 4 do begin
      count := 0;
      while not eoln(fin) do begin
         skipblanks(fin);
         if fin^ in ['0','1','2','3','4','5','6','7','8','9'] then begin
            count := succ(count);
            read(fin, store[base,count]);
            write(output,'+');
         end
         else begin
            write(output,'-');
            skipcolumn(fin);
         end
      end;
      writeln(output, ' base ', base:1);
      readln(fin);
   end;

   writeln(output);
   writeln(output, 'number of columns read: ',count:1);

   for base := 1 to 4 do begin
      write(output, 'base ', base:1, ' |');
      for c := 1 to count do begin
         write(output, ' ', store[base,c]:4);
      end;
      writeln(output);
   end;

   writeout(output);
   writeout(fout);

end;
(* end module transpose.themain *)

begin
   themain(fin, transposep, fout);
1: end.
