program waveplot(searchfeatures, waveplotp, xyin, output);
(* waveplot: plot sites around the wave of B-form DNA

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  toms@alum.mit.edu (permanent)
  https://alum.mit.edu/www/toms (permanent)

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.09; (* of waveplot.p 2012 Nov 21
2012 Nov 21, 1.09: give better single line definition
2012 Nov 21, 1.08: improve documentation
2012 Nov 20, 1.07: improve documentation
2012 Nov 08, 1.06: debug
2012 Nov 08, 1.05: clean up
2012 Nov 08, 1.04: backup
2012 Nov 08, 1.03: read through quote columns, add color plot
2012 Nov 08, 1.02: circle fractions should not be quantized
2012 Nov 08, 1.01: backup position display
2012 Nov 08, 1.00: origin
*)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.waveplot *)
(*
name
   waveplot: plot sites around the wave of B-form DNA

synopsis
   waveplot(searchfeatures: in, waveplotp: in, xyin: out, output: out)

files

   searchfeatures:  output of the search program containing the
   locations of a pattern along a sequence.

   xyin:  input to the xyplo program for plotting

   waveplotp:  parameters to control the program.  The file must contain the
      following parameters, one per line:

      parameterversion: The version number of the program.  This allows the
         user to be warned if an old parameter file is used.

      The parameters are like those in live:

      zerobase: integer; to set the phase of the wave by.
      wavelength: integer; The wave length in bases.  Usually 10.6.

      riconsensus: integer; Maximum value to divide Ri by to normalize
      color saturation.  If it is 0 or less, the saturation is set to
      1.0.

      These variables are from live but maybe not needed:
      wavevertical: integer; amount to shift colored wave vertically
      waveheight: integer; height of colored wave

   output: messages to the user

description

   When one identifies a set of sites, how are they oriented in
   physical space around the DNA relative to each other?  A zero
   coordinate is chosen by the user.  All sites that are exact
   multiples of 10.6 bp have the same orientation as a site at this
   coordinate.  The orientation of other sites is determined by
   repeatedly subtracting 10.6 bases per DNA twist from a coordinate
   until the value is above the zero coordinate but below 10.6.  The
   actual distance can be changed from 10.6 to whatever value the user
   desires.

examples

documentation

see also

   {The program that generates searchfeatures} search.p

   {A program that generates marks for lister maps
    that are colors rotating through the spectrum.
    The wavelength can be adjusted as here.} live.p

   {Program to plot the data} xyplo.p
   {Example xyplop file for xyplo:} waveplot.xyplop

author

   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.waveplot *)

(* begin module const.waveplot *)
   wid = 9; (* width of output numbers *)
   dec = 5; (* decimal places of output numbers *)
(* end module const.waveplot *)

var
   searchfeatures, (* file used by this program *)
   waveplotp, (* file used by this program *)
   xyin: text; (* file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 7.73; {of delmod.p 2012 Nov 08} *)

(* begin module skipblanks *)

(* 2003 July 31: tab is considered a blank character *)

function isblank(c: char): boolean;
(* is the character c blank or tab? *)
begin
   isblank := (c = ' ') or (c = chr(9))
end;

procedure skipblanks(var thefile: text);
(* skip over blanks until a non-blank, or end of line, is found *)
begin
   while isblank(thefile^) and not eoln(thefile) do get(thefile);
end;

procedure skipnonblanks(var thefile: text);
(* skip over nonblanks until a blank, or end of line, is found *)
begin
   while (not isblank(thefile^)) and not eoln(thefile) do get(thefile);
end;

procedure skipcolumn(var thefile: text);
(* skip over a data column *)
begin
   skipblanks(thefile); skipnonblanks(thefile)
end;
(* end module skipblanks version = 7.73; {of delmod.p 2012 Nov 08} *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 7.73; {of delmod.p 2012 Nov 08} *)

(* begin module skipquote *)
procedure skipquote(var thefile: text);
(* skip over a data column - if it starts with a quote,
skip to the next quote of the same kind. *)
var
   c: char; (* the first character of the column *)
   done: boolean; (* done analyzing a quoted string *)
begin
{writeln(output, 'BUBBA skipquote start');}
   skipblanks(thefile);
   c := thefile^;
{writeln(output, 'BUBBA skipquote start c=<',c,'>');}
   if (c = '"') or (c = '''') then begin
      (* Note: '''' is the symbol for a single quote *)
      get(thefile);
      done := false;
      while not done do begin
{write(output,'(A)',thefile^);}
         if eoln(thefile) then begin
write(output,'(eoln)');
            done := true;
         end
         else begin
            if thefile^ = c then begin
               done := true;
{write(output,'(B)',thefile^);}
               get(thefile); (* move past the final quote mark *)
            end
            else begin
{write(output,'(C)',thefile^);}
               get(thefile);
            end
         end;
{
writeln(output, 'BUBBA skipquote halt');
halt;
}
      end
   end
   else begin
      skipnonblanks(thefile);
{writeln(output, 'BUBBA skipquote skipnonblanks');}
   end;
{writeln(output, 'BUBBA skipquote end');}
end;
(* end module skipquote *)

(* begin module backspace *)
procedure backspace(var f: text; n: integer);
(* produce n backspaces *)
var
  i: integer; (* index for n *)
begin
  for i := 1 to n
  do write(output,''); (* the backspace character *)
end;
(* end module backspace version = 7.73; {of delmod.p 2012 Nov 08} *)

(* begin module realmodulo *)
function realmodulo(position: real; wavelength: real): real;
(* reduce the initial position to a number between 0 and 1 by
repeatedly subtracting the wavelength *)
var
   p: real; (* internal variable for result *)
begin
   p := position;
{writeln(output,'wavelength = ',wavelength:5:3);}
   while p > wavelength do begin
{writeln(output,'p = ',p:5:3);}
      p := p - wavelength;
   end;
   realmodulo := p
end;
(* end module realmodulo *)

(* begin module waveplot.themain *)
procedure themain(var searchfeatures, waveplotp, xyin: text);
(* the main procedure of the program *)
var
   parameterversion: real; (* parameter version number *)

   position: real; (* position on a DNA or RNA sequence *)
   orientation: real; (* orientation on a DNA or RNA sequence *)
   firstbase: real; (* the first position *)
   reducedposition: real; (* the position reduced to a number between
                             0 and the wavelength *)
   normalizedposition: real; (* the reducedposition 
                                divided by the wavelength *)

   {These variables are from the live program: }
   zerobase: integer; (* to set the phase of the wave by. *)
   wavelength: real; (* The wave length in bases.  Usually 10.6. *)
   {These variables are from live but maybe not needed: }
   wavevertical: integer; (* amount to shift colored wave vertically *)
   waveheight: integer; (* height of colored wave *)

   ri: real; (* individual information, bits *)
   riconsensus: real; (* consensus for the site *)
   hue: real; (* color parameter *)
   saturation: real; (* color parameter *)
   brightness: real; (* color parameter *)
begin
   writeln(output,'waveplot ',version:4:2);
   reset(waveplotp);
   readln(waveplotp, parameterversion);
   if round(100*parameterversion) < round(100*updateversion) then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;
   readln(waveplotp, zerobase);
   readln(waveplotp, wavelength);
if not eof(waveplotp) then begin
   readln(waveplotp, riconsensus);
end
else begin
   riconsensus := 20.0;
end;
   writeln(output, 'riconsensus : ',riconsensus:wid:dec );
   readln(waveplotp, wavevertical);
   readln(waveplotp, waveheight);

   reset(searchfeatures);
   rewrite(xyin);

   firstbase := maxint;
   write(output, 'coordinate: ');

{
@ hg19-chr1 198668711.0 +1 "CpG" "" R   0.000000   1.000000   0.000000   1.000000
}

   while not eof (searchfeatures) do begin
       if eoln(searchfeatures)
       then readln(searchfeatures)
       else if searchfeatures^ = '@' then begin
          (* read in things *)
          skipcolumn(searchfeatures); (* skip the @ *)
{writeln(output,'BUBBA 1');}
          skipcolumn(searchfeatures); (* accession *)
{writeln(output,'BUBBA 2');}
          read(searchfeatures, position); (* position *)
{writeln(output,'BUBBA 3');}
          read(searchfeatures, orientation); (* orientation *)
{writeln(output,'BUBBA 4');}
          skipquote(searchfeatures); (* name *)
{writeln(output,'BUBBA 5');}
{write(output,'<othertag>');copyaline(searchfeatures, output); halt;}
          skipquote(searchfeatures); (* othertag *)
{writeln(output,'BUBBA 6');}
{write(output,'<ri>');copyaline(searchfeatures, output); halt;}
          if searchfeatures^ in ['1','2','3','4','5','6','7','8','9','0']
          then read(searchfeatures,ri) (* ri *)
          else skipcolumn(searchfeatures); (* color info *)
{writeln(output,'BUBBA 7 ri = ', ri:wid:dec);}
          readln(searchfeatures);
{writeln(output,'BUBBA 8');}

          if firstbase = maxint then begin
             firstbase := position; (* capture the first base *)
          end;

          (* compute things *)
          reducedposition := realmodulo(position, wavelength);
          normalizedposition := reducedposition / wavelength;
          hue := 1.0;
          if riconsensus <= 0
          then saturation := 1.0
          else saturation := ri/riconsensus;
          brightness := 1.0;
          (* output things *)
          writeln(xyin, round(position):wid,
                        ' ',round((position - firstbase)):wid,
                        ' ',reducedposition:wid:dec,
                        ' ',normalizedposition:wid:dec,
                        ' ',ri:wid:dec,
                        ' ',hue:wid:dec,
                        ' ',saturation:wid:dec,
                        ' ',brightness:wid:dec);
          write(output, round(position):wid);
          backspace(output, wid);
{ example @ line:
@ hg19-chr1 198668711.0 +1 "CpG" ""     0.000000   0.000000   0.000000   0.000000
}
       end
       else begin
          readln(searchfeatures);
       end
   end;
   writeln(output);

end;
(* end module waveplot.themain *)

begin
   themain(searchfeatures, waveplotp, xyin);
1: end.

{
searchfeatures file:
********************************************************************************
* search 6.95

* THE CURRENT FEATURE DEFINITION IS:
* define "CpG" "-" "[]" "[]" 0.0 1.0

* cg.
define "CpG" "-" "[]" "[]" 0.0 1.0

@ hg19-chr1 198668711.0 +1 "CpG" ""     0.000000   0.000000   0.000000   0.000000
@ hg19-chr1 198668726.0 +1 "CpG" ""     0.000000   0.000000   0.000000   0.000000

********************************************************************************
}
