program winrs(book, cmp, winrsp, xyin, output);
(* winrs: Compute Rs/window vs window in a single program

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  https://alum.mit.edu/www/toms 

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.17; (* of winrs.p 2016 Jan 25
2016 Jan 25, 1.17: debug: add local variables:
  sed -i.bak 's|procedure BuildPrecomputedTable;|procedure BuildPrecomputedTable;va
+r n: integer;|g' winrs.p
  sed -i.bak 's|procedure DoTheCrunchAndMakeTheXyin;|procedure DoTheCrunchAndMakeTh
+eXyin;var b: base;WindowSize,PositionInWindow, Window: integer;|g' winrs.p
1996 Apr 10, 1.16: functional
1996 Apr  8, 1.00: origin
origin 1996 April 8 *)
(* end module version *)

(* begin module describe.winrs *)
(*
name
   winrs: Compute Rs/window vs window in a single program

synopsis
   winrs(book: in, cmp: in, winrsp: in, xyin: out, output: out)

files
   book: A book from Delila.  Only the first DNA piece in the book is used.

   cmp: A composition from the comp program.
       If cmp is empty, then equal frequencies are assumed.

   winrsp:  :  Parameters to control the program: one per line

      StartCoordinate (integer):  The base of the sequence to be used for the
         zero coordinate of the first window.  For simplicity, internal
         coordinates are used: the sequence begins at base 1.

      MinimumWindowSize (integer): The shortest window to use for the
         computations.

      MaximumWindowSize (integer): The longest window to use for the
         computations.

      SpaceBetweenWindows (integer):  Windows are spaced so as to not
         collide.  You can add extra distance with this parameter.
         If it is negative, the windows will overlap.

      NumberofWindows (integer): The number of windows to use for
         each computation.

      * The program will object if any of these parameters cause a window to
      go outside the sequence.

   xyin: output of this program, input to the xyplo graphics program.
         columns:
         1: window size
         2: Rs, Rsequence, bits
         3: SD, Standard Deviation of Rs, bits
         4: Rs/(window size), bits/base
         5: SD/(window size), bits/base

   output: messages to the user

description

   A window size is chosen.  The first sequence in the book is broken into
   fragments of this size.  These are aligned and the information content,
   Rsequence (Rs) is computed.  This process is repeated for a range of
   window sizes defined in the parameters.  The output is a data file (xyin)
   containing columns for window size (w), Rs and the standard deviation of
   Rs (SD).  There are also two normalized columns for Rs/w and SD/w, which
   are reports in bits per base.  Repeating structures will appear as peaks
   in a graph of Rs/w versus w.

examples

documentation
   [1996 April 8]

see also
   lidel.p, winscan.p, delila.p, encode.p, rseq.p, xyplo.p

author
   Thomas Dana Schneider

bugs
   The program should produce a xyplop so that it can put
   the genomic information on the curve automatically.

technical notes
   Paremeter dnamax must be set as large as possible so that the entire
   sequence can fit into a single array and therefore be accessed as rapidly
   as possible.  Since Delila stores sequences with 2 bits/base (assuming a
   good compiler) so one can compute the size of the array in bytes.

*)
(* end module describe.winrs *)

(* begin module winrs.const *)
(* Constants needed for book manipulations.  This replaces module book.const
so that dnamax can be controlled *)

      dnamax = 400000; (* length of dna arrays *)
      namelength = 20; (* maximum key name length *)
      linelength = 80; (* maximum line readable in book *)

      LongestWindowAllowed = 5000; (* longest window allowed *)
      infofield = 8; (* field size for bits *)
      infodecim = 5; (* decimal places for bits *)

(* end module winrs.const *)

type
(* begin module book.type *)
(* types needed for book manipulations *)

      chset = set of 'a'..'z';

      (* types defined in book definition *)

      alpha = packed array[1..namelength] of char; (* this is not alfa *)

      (* name is a left justified string with blanks following the
         characters *)
      name = record
         letters: alpha;
         length: 0..namelength (* zero means an unspecified structure *)
      end;

      lineptr = ^line;
      line = record (* a line of characters *)
         letters: packed array [1..linelength] of char;
         length: 0..linelength;
         next: lineptr
      end;

      direction = (plus, minus, dircomplement, dirhomologous);
      configuration = (linear, circular);
      state = (on, off);


      header = record (* header of key *)
         keynam: name; (* key name of structure *)
         fulnam: lineptr; (* full name of structure *)
         note:   lineptr (* note key *)
      end;


      (* base types *)
      base = (a,c,g,t);
      dnaptr = ^dnastring;
      dnarange = 0..dnamax;
      seq = packed array[1..dnamax] of base;
      dnastring = record
         part: seq;
         length: dnarange;
         next: dnaptr
      end;


      orgkey = record (* organism key *)
         hea: header;
         mapunit: lineptr (* genetic map units *)
      end;

      chrkey = record (* chromosome key *)
         hea: header;
         mapbeg: real; (* number of genetic map beginning *)
         mapend: real (* number of genetic map ending *)
      end;

      pieceptr = ^piece;
      piekey = record (* piece key *)
         hea: header;
         mapbeg: real; (* genetic map beginning *)
         coocon: configuration; (* configruation (circular/linear) *)
         coodir: direction; (* direction (+/-) relative to genetic map *)
         coobeg: integer; (* beginning nucleotide *)
         cooend: integer; (* ending nucleotide *)
         piecon: configuration; (* configruation (circular/linear) *)
         piedir: direction; (* direction (+/-) relative to coordinates *)
         piebeg: integer; (* beginning nucleotide *)
         pieend: integer; (* ending nucleotide *)
      end;


      piece = record
         key: piekey;
         dna: dnaptr
      end;

      reference = record
         pienam : name;  (* name of piece referred to *)
         mapbeg : real;  (* genetic map beginning *)
         refdir : direction;  (* direction relative to coordinates *)
         refbeg : integer;  (* beginning nucleotide *)
         refend : integer;  (* ending nucleotide *)
      end;

      genkey = record  (* gene key *)
         hea : header;
         ref : reference;
      end;

      trakey = record  (* transcript key *)
         hea : header;
         ref : reference;
      end;

      markerptr = ^marker;
      markey = record (* marker key *)
         hea : header;
         ref : reference;
         sta : state;
         phenotype : lineptr;
         next : markerptr;
      end;

      marker = record
         key : markey;
         dna : dnaptr;
      end;

(* end module book.type version = 'delmod 6.66 95 Oct 24 tds/gds' *)

(* begin module comp.type *)
            (* the composition is stored in a tree of these nodes *)
      compnodeptr = ^compnode; (* points to a node of the tree *)
      compnode = record    (* a node of the composition tree *)
         count : integer; (* the number of oligos for this node *)
         son : array[base] of compnodeptr; (* the pointers to 'descendants'
                                           of this node of the tree *)
      end;

            (* spiders are used to make the composition tree *)
      spiderptr = ^spider; (* points to a 'spider' *)
      spider = record  (* a spider climbs the composition tree, its path
                        determined by the sequences, and increments 'count' at
                        all the nodes it passes, thereby determining the
                        composition *)
         depth : integer; (* the level of the node now at *)
         place : compnodeptr; (* a pointer to the current node *)
         next : spiderptr; (* the next spider in the collection *)
      end;

            (* the total number of composition entries at a given level
               is stored in the linked list of type comptotal *)
      compptr = ^comptotal;
      comptotal = record
         count: integer;  (* the number at a given level *)
         next: compptr;  (* pointer to the next level totals *)
      end;

           (* 'path' is used in printing the tree *)
      pathptr = ^path; (* pointer into a path on the tree *)
      path = record  (* the path of bases to get to a particular node *)
         bas : base;
         next : pathptr;
      end;
(* end module comp.type version = 5.35; (@ of rseq.p 1995 Feb 21 *)

var
   book, cmp, winrsp, (* files used by this program *)
   xyin: text; (* file used by this program *)

(* begin module book.var *)
(* ************************************************************************ *)
(* global variables needed for book manipulations *)

      (* free storage: *)
      freeline: lineptr; (* unused lines *)
      freedna: dnaptr; (* unused dnas *)

      readnumber: boolean; (* whether to read a number from the notes, or
                              to read in the notes *)
      number: integer; (* the number of the item just read *)
      numbered: boolean; (* true when the item just read is numbered *)
      skipunnum: boolean; (* a control variable to allow skipping of
                             un-numbered items in the book *)

(* ************************************************************************ *)
(* end module book.var version = 'delmod 6.66 95 Oct 24 tds/gds' *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 5.35; (@ of rseq.p 1995 Feb 21 *)

(* ************************************************************************ *)
(* ************************************************************************ *)
(* ************************************************************************ *)
(* begin module package.getpiece *)
(* ************************************************************************ *)
(* begin module package.brpiece *)
(* ************************************************************************ *)
(* begin module book.basis *)
(* procedures needed for book manipulations *)

(* get procedures should be used for all linked lists of records *)

procedure getline(var l: lineptr);
(* obtain a line from the free line list or by making a new one *)
begin
      if freeline<>nil
         then begin
            l:=freeline;
            freeline:=freeline^.next
         end
         else new(l);
      l^.length:=0;
      l^.next:=nil
end;

procedure getdna(var l: dnaptr);
begin
      if freedna<>nil
         then begin
            l:=freedna;
            freedna:=freedna^.next
         end
         else new(l);
      l^.length:=0;
      l^.next:=nil
end;

(* clear procedures should be called each time the records are no longer needed
         failure to do this may result in a stack overflow. *)

procedure clearline(var l: lineptr);
(* return a line to the free line list *)
var   lptr: lineptr;
begin
      if l<>nil then begin
         lptr:=l;
         l:=l^.next;
         lptr^.next:=freeline;
         freeline:=lptr
      end
end;

procedure cleardna(var l: dnaptr);
var   lptr: dnaptr;
begin
      if l<>nil then begin
         lptr:=l;
         l:=l^.next;
         lptr^.next:=freedna;
         freedna:=lptr
      end
end;

procedure clearheader(var h: header);
(* clear the header h (remove lines to free storage) *)
begin
      with h do begin
         clearline(fulnam);
         while note<>nil do clearline(note)
      end
end;

procedure clearpiece(var p: pieceptr);
(* clear the dna of the piece *)
begin
      while p^.dna<>nil do cleardna(p^.dna);
      clearheader(p^.key.hea)
end;

function chartobase(ch:char):base;
(* convert a character into a base *)
begin
      case ch of
         'a': chartobase:=a;
         'c': chartobase:=c;
         'g': chartobase:=g;
         't': chartobase:=t
      end
end;

function basetochar(ba:base):char;
(* convert a base into a character *)
begin
      case ba of
         a: basetochar:='a';
         c: basetochar:='c';
         g: basetochar:='g';
         t: basetochar:='t';
      end
end;

function complement(ba:base):base;
(* take the complement of ba *)
begin
      case ba of
         a: complement:=t;
         c: complement:=g;
         g: complement:=c;
         t: complement:=a;
      end
end;

function pietoint(p: integer; pie: pieceptr): integer;
(* p is a coordinate on the piece.
   we want to transform p into a number
   from 1 to n: an internal coordinate system for
   easy manipulation of piece coordinates *)
var   i: integer; (* an intermediate value *)
begin
      with pie^.key do begin
         case piedir of
            plus:  if p>=piebeg
                   then i:=p-piebeg+1
                   else i:=(p-coobeg)+(cooend-piebeg)+2;
            minus: if p<=piebeg
                   then i:=piebeg-p+1
                   else i:=(cooend-p)+(piebeg-coobeg)+2
         end;
         pietoint:=i
      end
end;

function inttopie(i: integer; pie: pieceptr):integer;
(* i is in the range 1 to some maximum.  it is an internal coordinate
   system for the program.  we want to do a
   coordinate transformation to obtain
   a value in the range of the piece called pie:
   i=1 corresponds to piebeg and
   i=its maximum corresponds to pieend *)
var   p: integer; (* an intermediate value *)
begin
      with pie^.key do begin
         case piedir of
            plus: begin
               p:=piebeg+(i-1);
               if p>cooend
                  then if coocon=circular
                     then p:=p-(cooend-coobeg+1)
            end;
            minus: begin
               p:=piebeg-(i-1);
               if p<coobeg
                  then if coocon=circular
                     then p:=p+(cooend-coobeg+1)
            end
         end;
         inttopie:=p
      end
end;

function piecelength(pie: pieceptr): integer;
(* return the length of the dna in pie *)
begin
      piecelength:=pietoint(pie^.key.pieend,pie)
end;

(* end module book.basis version = 'delmod 6.66 95 Oct 24 tds/gds' *)

(* begin module book.getto *)
function getto(var thefile: text; ch: chset): char;
(* search the file for a character in the first line which is a
   member of the set ch. *)
var achar: char;
begin
      achar:=' ';
      while (not(achar in ch)) and (not eof(thefile))
         do readln(thefile,achar);
      if (achar in ch) then getto:=achar
                       else getto:=' '
end;
(* end module book.getto version = 'delmod 6.66 95 Oct 24 tds/gds' *)

(* begin module book.skipstar *)
procedure skipstar(var thefile:text);
(* skip start of line (or star = '*'). *)
begin (* skipstar *)
      if thefile^ <> '*' then begin
         writeln(output,' procedure skipstar: bad book');
         writeln(output,' "*" expected as first character on the line, but "',
                        thefile^,'" was found');
         halt
      end;
      get(thefile); (* skip the star *)

      if thefile^ <> ' ' then begin
         writeln(output,' procedure skipstar: bad book');
         writeln(output,' "* " expected on a line but "*',
                        thefile^,'" was found');
         halt
      end;
      get(thefile) (* skip the blank *)
end; (* skipstar *)
(* end module book.skipstar version = 'delmod 6.66 95 Oct 24 tds/gds' *)

(* begin module book.brreanum *)
procedure brreanum(var thefile: text; var reanum: real);
(* read a real number from the file *)
begin
      skipstar(thefile);
      readln(thefile,reanum);
end;
(* end module book.brreanum version = 'delmod 6.66 95 Oct 24 tds/gds' *)

(* begin module book.brnumber *)
procedure brnumber(var thefile: text; var num: integer);
(* read a number from the file *)
begin
      skipstar(thefile);
      readln(thefile,num)
end;
(* end module book.brnumber version = 'delmod 6.66 95 Oct 24 tds/gds' *)

(* begin module book.brname *)
procedure brname(var thefile: text; var nam: name);
(* read a name from the file *)
var   i: integer; (* an index to the name *)
      c: char; (* a character read *)
begin (* brname *)
      skipstar(thefile);
      with nam do begin
         length:=0;
         repeat
            length:=succ(length);
            read(thefile,c);
            letters[length] := c
         until (eoln(thefile)) or
               (length>=namelength) or
               (letters[length]=' ');
         if letters[length]=' ' then length:=length-1;
         if length<namelength
            then for i:=length+1 to namelength do letters[i]:=' '
      end;
      readln(thefile)
end; (* brname *)
(* end module book.brname version = 'delmod 6.66 95 Oct 24 tds/gds' *)

(* begin module book.brline *)
procedure brline(var thefile: text; var l: lineptr);
(* read a line from the file *)
var
      i,j: integer;
      acharacter: char;
begin
      skipstar(thefile);
      i:=0;
      while (not eoln(thefile)) do begin
         i:=succ(i);
         read(thefile,acharacter);
         l^.letters[i]:=acharacter
      end;
      if i<l^.length then for j:=i+1 to l^.length do l^.letters[j]:=' ';
      l^.length:=i;
      l^.next:=nil;
      readln(thefile)
end;
(* end module book.brline version = 'delmod 6.66 95 Oct 24 tds/gds' *)

(* begin module book.brdirect *)
procedure brdirect(var thefile: text; var direct: direction);
(* read a direction *)
var   ch: char;
begin
      skipstar(thefile);
      readln(thefile,ch);
      if ch='+' then direct:=plus
                else direct:=minus
end;
(* end module book.brdirect version = 'delmod 6.66 95 Oct 24 tds/gds' *)

(* begin module book.brconfig *)
procedure brconfig(var thefile: text; var config: configuration);
(* read a configuration *)
var   ch: char;
begin
      skipstar(thefile);
      readln(thefile,ch);
      if ch='l' then config:=linear
                else config:=circular
end;
(* end module book.brconfig version = 'delmod 6.66 95 Oct 24 tds/gds' *)

(* begin module book.brnotenumber *)
procedure brnotenumber(var thefile: text; var note: lineptr);
(* book note reading to obtain the number of the object.
the procedure returns the value of the number as a global.
(this is not such a good practice, but we are stuck with it for now.) *)
begin (* brnotenumber *)
      note:=nil;
      numbered := false;
      number := 0; (* force number to zero if there
         is no number at all *)
      (* the next character is n or * depending on whether there are notes *)
      if thefile^ = 'n' then begin
         readln(thefile);
         if thefile^ <> 'n' then begin
            skipstar(thefile);
            if not eoln(thefile) then begin
               if thefile^ = '#' then begin
                  numbered := true;
                  get(thefile); (* move past the number symbol *)
                  read(thefile,number);
               end
            end;
            repeat
               readln(thefile)
            until thefile^ = 'n';
            readln(thefile)
         end
         else readln(thefile)
      end
end; (* brnotenumber *)
(* end module book.brnotenumber version = 'delmod 6.66 95 Oct 24 tds/gds' *)

(* begin module book.brnote *)
procedure brnote(var thefile: text; var note: lineptr);
(* read note key *)
var
      newnote: lineptr; (* the new note *)
      previousnote: lineptr; (* the last line of the notes *)
begin (* brnote *)
      note:=nil;
      if thefile^ = 'n' then begin (* enter note *)
         readln(thefile);
         if thefile^ <> 'n' then begin (* abort null note (n/n) *)
            getline(note);
            newnote:=note;
            while thefile^ <> 'n' do begin (* wait until end of note *)
               brline(thefile,newnote);
               previousnote:=newnote;
               (* get next note *)
               getline(newnote^.next);
               newnote:=newnote^.next;
            end;
            (* last note was not used, so: *)
            clearline(newnote);
            previousnote^.next:=nil;
            readln(thefile)
         end
         else readln(thefile)
      end
end; (* brnote *)
(* end module book.brnote version = 'delmod 6.66 95 Oct 24 tds/gds' *)

(* begin module book.brheader *)
procedure brheader(var thefile: text; var hea: header);
(* read the header of a key. *)
begin
      with hea do begin
         (* read key name *)
         brname(thefile,keynam);

         (* read full name *)
         getline(fulnam);
         brline(thefile,fulnam);

         (* read note key *)
         if readnumber then brnotenumber(thefile,note)
                       else brnote(thefile,note)
      end
end;
(* end module book.brheader version = 'delmod 6.66 95 Oct 24 tds/gds' *)

(* begin module book.copyheader *)
procedure copyheader(fromhea: header; var tohea: header);
(* copy the header fromhea into tohea.  Note that the
linked objects are NOT copied, but merely pointed to. *)
begin
   tohea.keynam.letters := fromhea.keynam.letters;
   tohea.keynam.length  := fromhea.keynam.length;
   tohea.note           := fromhea.note;
   tohea.fulnam         := fromhea.fulnam;
end;
(* end module book.copyheader version = 'delmod 6.66 95 Oct 24 tds/gds' *)

(* begin module book.brpiekey *)
procedure brpiekey(var thefile: text; var pie: piekey);
(* read piece key *)
begin
      with pie do begin
         brheader(thefile,hea);
         brreanum(thefile,mapbeg);
         brconfig(thefile,coocon);
         brdirect(thefile,coodir);
         brnumber(thefile,coobeg);
         brnumber(thefile,cooend);
         brconfig(thefile,piecon);
         brdirect(thefile,piedir);
         brnumber(thefile,piebeg);
         brnumber(thefile,pieend);
      end
end;
(* end module book.brpiekey version = 'delmod 6.66 95 Oct 24 tds/gds' *)

(* begin module book.brdna *)
procedure brdna(var thefile: text; var dna: dnaptr);
(* read in dna from thefile *)
(* note: if the dna were circularized, by linking the last dnastring
to the first, then the cleardna routine could not clear properly,
and would loop forever... there is no reason to do that, since a simple
mod function will allow one to access the circle. *)
var
      ch: char;
      workdna: dnaptr;
begin
      getdna(dna);
      workdna:=dna;
      ch:=getto(thefile,['d']);
      read(thefile,ch); (* skipstar *)
      while (ch = '*') do
      begin
         read(thefile,ch); (* skip blank *)
         repeat
            read(thefile,ch);
            if ch in ['a','c','g','t'] then begin
               if workdna^.length=dnamax then begin
                  getdna(workdna^.next);
                  workdna:=workdna^.next
               end;
               workdna^.length:=succ(workdna^.length);
               workdna^.part[workdna^.length]:=chartobase(ch)
            end
         until eoln(thefile);
         readln(thefile); (* go to next line *)
         read(thefile,ch); (* ch is either '*' or 'd' *)
      end;
      readln(thefile)
end;
(* end module book.brdna version = 'delmod 6.66 95 Oct 24 tds/gds' *)

(* begin module book.brpiece *)
procedure brpiece(var thefile: text; var pie: pieceptr);
(* read in a piece *)
begin
      brpiekey(thefile,pie^.key);
      if numbered or (not skipunnum)
         then brdna(thefile,pie^.dna)
end;
(* end module book.brpiece version = 'delmod 6.66 95 Oct 24 tds/gds' *)

(* begin module book.brinit *)
procedure brinit(var book: text);
(* check that the book is ok to read, and
   set up the global variables for br routines *)
begin (* brinit *)
      (* halt if the book is bad (first word is 'halt') or the first
         character is not * *)
      reset(book);
      if not eof(book) then begin
         (* check for the date line *)
         if book^ <> '*' then begin
            if book^ <> 'h'
            then writeln(output, ' this is not the first line of a book:')
            else writeln(output, ' bad book:');
            write(output, ' ');

            while not (eoln(book) or eof(book)) do begin
               write(output, book^);
               get(book)
            end;
            writeln(output);
            halt
         end
      end
      else begin
         writeln(output, ' book is empty');
         halt
      end;

      (* initialize free storage *)
      freeline:=nil;
      freedna:=nil;

      readnumber:=true; (* usually we read in numbers for items *)
      number:=0; (* arbitrary value *)
      numbered:=false; (* the piece has no number (none yet read in) *)
      skipunnum:=false;
end; (* brinit *)
(* end module book.brinit version = 'delmod 6.66 95 Oct 24 tds/gds' *)

(* ************************************************************************ *)
(* end module package.brpiece version = 'delmod 6.66 95 Oct 24 tds/gds' *)

(* begin module book.getpiece *)
procedure getpiece(var thefile: text; var pie: pieceptr);
(* move to and read in the next piece in the book *)
var   ch: char;
begin
      ch:=getto(thefile,['p']); (* get to the next p(iece) in the book *)
      if ch<>' ' then begin
         brpiece(thefile,pie);
         ch:=getto(thefile,['p']); (* read past closing p *)
      end
end;
(* end module book.getpiece version = 'delmod 6.66 95 Oct 24 tds/gds' *)

(* ************************************************************************ *)
(* end module package.getpiece version = 'delmod 6.66 95 Oct 24 tds/gds' *)
(* ************************************************************************ *)
(* ************************************************************************ *)
(* ************************************************************************ *)

(* begin module book.getbase *)
function getbase(position: integer; pie: pieceptr):base;
(* get a base from the nth position (internal coordinates) of the
   piece.  no protection is made against positions outside the piece *)
var
      workdna: dnaptr;
      p: integer; (* the last base of the dna part *)
begin
      workdna:=pie^.dna;
      p:=dnamax;
      while position>p do begin
         p:=p+dnamax;
         workdna:=workdna^.next
      end;
      getbase:=workdna^.part[position-(p-dnamax)]
end;
(* end module book.getbase version = 'delmod 6.66 95 Oct 24 tds/gds' *)

(* begin module copyaline *)
procedure copyaline(var fin, data: text);
(* copy a line from file fin to file data *)
begin (* copyaline *)
      while not eoln(fin) do begin
         data^ := fin^;
         put(data);
         get(fin)
      end;
      readln(fin);
      writeln(data);
end; (* copyaline *)
(* end module copyaline version = 5.35; (@ of rseq.p 1995 Feb 21 *)

(* begin module comp.readcomp *)

(* begin module skipoligo *)
procedure skipoligo(var thefile: text);
(* this procedure is used to skip over an oligonucleotide string.
   it reads until it comes to a base, and then continues to read until
   it comes to a blank. no checking is done for non-base characters
   in between. *)
var ch : char;
begin
      repeat read(thefile,ch)
      until (ch in ['a','c','g','t']);
      repeat read(thefile,ch)
      until (ch = ' ');
end;
(* end module skipoligo version = 'auxmod 1.37 85 apr 4 gds/tds'; *)

procedure readcomp(var comp: text; var compmax: integer; readmax: integer;
                   var root: compnodeptr; var monocomptotal: compptr);
(* this procedure requires modules: comp.type, skipoligo, halt;
   this procedure reads from file 'comp' a composition and puts it into
   the tree pointed to by the 'root' pointer.  'compmax' is the depth
   of the composition tree which is stored.  it is the minimum of
   'readmax', the requested depth, and 'detcomp', the depth for which
   the input file composition was determined.  'monocomptotal' points
   to the beginning (for the monos) of a linked list which gives the
   totals for each level of the composition. *)
type  (* for efficient reading  of the information into the tree, each
         node that has a successor is stored in a queue as well as put
         into the tree.  the variables of the type list are the queue
         elements. *)
    listptr = ^list;
    list = record
      item: compnodeptr;  (* points to the composition tree node *)
      next: listptr;  (* points to the next list item in the queue *)
    end;

var  freeitem,  (* the list of unused list pointers *)
     listitem, (* an item in the queue *)
     first,  (* the first item in the queue *)
     last: listptr;  (* the last item in the queue *)
     comptot,  (* the comp total for a given level *)
     newcomptot: compptr;  (* for adding to the string of comptot"s *)
     detcomp,  (* the level to which the input composition was determined *)
     level,  (* of the composition being read, i.e., monos, dis, ... *)
     number: integer;  (* read from the 'comp' file *)
     ch: char;  (* for reading from 'comp' *)
     ba: base;  (* an index *)

(* getitem and clearitem provide efficient use of linked list storage by
   keeping a list of unused pointers that can be allocated instead of
   always creating 'new' ones  *)
procedure getitem(var l: listptr);
(* obtain a listitem from the free list or by making a new one *)
begin
      if freeitem<>nil
         then begin
            l:=freeitem;
            freeitem:=freeitem^.next
         end
         else new(l);
      l^.next:=nil
end;


procedure clearitem(var l: listptr);
(* return a listitem to the free list *)
var   lptr: listptr;
begin
      if l<>nil then begin
         lptr:=l;
         l:=l^.next;
         lptr^.next:=freeitem;
         freeitem:=lptr
      end
end;

begin  (* readcomp *)
      reset(comp);
      if eof(comp) then begin
         writeln(output,' error: no composition file provided');
         halt;
      end;
      readln(comp); (* skip the program identification *)
      readln(comp); (* skip the book identification *)
      readln(comp,detcomp); (* obtain the determined composition *)
      readln(comp); (* skip the blank line *)

            (* determine the level of composition to be stored *)
      if (readmax < 1) then begin
         writeln(output);
         writeln(output,' warning: 0 or negative oligo length requested');
         writeln(output,' composition used is depth ',detcomp:1);
         writeln(output);
         compmax := detcomp
      end
      else if (readmax > detcomp) then begin
         writeln(output);
         writeln(output,' warning: requested composition oligo',
                        ' length (',readmax:1,')');
         writeln(output,' is larger than the determined composition',
                        ' oligo length (',detcomp:1,').');
         writeln(output,' composition used is to depth ',detcomp:1);
         writeln(output);
         compmax := detcomp
      end
      else compmax := readmax;

      new(root);
      new(monocomptotal);
      new(comptot);
      new(first);
      new(last);
      first^.item := root;
      first^.next := nil;
      last := first;
      new(freeitem);
      freeitem^.next := nil;

         (* read in the total number of bases from the composition *)
      readln(comp); (* skip the * *)
      readln(comp); (* skip the information line *)
      readln(comp,number);
      root^.count := number;

      repeat
         read(comp,ch); (* skip the space *)
         read(comp,ch); (* the ch determines what to do next *)
         if (ch = '*') then begin  (* determine the level about to be read *)
            readln(comp);
            readln(comp,level);
            if (level = 1) then monocomptotal := comptot
            else begin
               new(newcomptot);
               comptot^.next := newcomptot;
               comptot := newcomptot;
            end;
            comptot^.count := 0;
            comptot^.next := nil
         end
         else begin  (* read the composition values *)
            for ba := a to t do begin
               skipoligo(comp);
               read(comp,number);
               if (number <> 0) then begin
                  new(first^.item^.son[ba]);
                  first^.item^.son[ba]^.count := number;
                  comptot^.count := comptot^.count + number;
                  getitem(listitem);
                  last^.next := listitem;
                  last := listitem;
                  last^.next := nil;
                  last^.item := first^.item^.son[ba]
               end
               else first^.item^.son[ba] := nil;
            end;
            clearitem(first);
            readln(comp);
         end;
      until level = compmax;
         (* for the last level of compositions to be read we don"t
            need to store the nodes in the queue because their successors
            will not be read in. *)
      repeat
         for ba := a to t do begin
            skipoligo(comp);
            read(comp,number);
            if (number <> 0) then begin
               new(first^.item^.son[ba]);
               first^.item^.son[ba]^.count := number;
               comptot^.count := comptot^.count + number;
            end
            else first^.item^.son[ba] := nil;
         end;
         clearitem(first);
         readln(comp);
      until (first = nil);
end;  (* readcomp *)

function getcount(root: compnodeptr; start: pathptr): integer;
(* this function follows the tree from the node 'root' (which
   may be the root of a subtree) along the path initiated with
   'start', and returns the count of the resulting node.  if the
   path through the tree ever hits a null node in the process
   the count is returned to be zero. *)
var place: compnodeptr;  (* a place in the composition tree *)
    point: pathptr;  (* a point in the path through the tree *)
begin
      point:= start;
      place := root;
      while ((place <> nil) and (point <> nil)) do begin
         place := place^.son[point^.bas];
         point := point^.next;
      end;
      if (place = nil) then getcount := 0
      else getcount := place^.count;
end;

(* end module comp.readcomp version = 5.35; (@ of rseq.p 1995 Feb 21 *)

(* begin module comp.getmonocomposition *)
procedure getmonocomposition(var cmp: text;
                             var gna,gnc,gng,gnt: integer;
                             var data: text;
                             var equalmono: boolean);
(* get the mononucleotide composition in gna to gnt from the file cmp,
messages about what is returned go to data.  if the file cmp is
empty, equal mononucleotides are assumed and equalmono is true *)
var
      (* variables for composition procedures readcomp and getcount *)
      compmax: integer;
      root: compnodeptr;
      monocomptotal: compptr;
      start: pathptr;
      b1: base; (* base for comp *)
      count: integer; (* one of the base counts *)
begin (* getmonocomposition *)
      reset(cmp);
      if eof(cmp)
      then begin
         writeln(data,'* the composition file is empty.');
         writeln(data,'* (equal mononucleotides assumed)');
         gna := 1; gnc := 1; gng := 1; gnt := 1;
         equalmono := true
      end
      else begin
         write(data,'*');
         copyaline(cmp,data);
         write(data,'*');
         copyaline(cmp,data);

         readcomp(cmp,compmax,1,root,monocomptotal);

         (* obtain the mono composition *)
         new(start);
         start^.next := nil;
         for b1 := a to t do begin
            start^.bas := b1;
            count := getcount(root,start);
            case b1 of
               a: gna := count;
               c: gnc := count;
               g: gng := count;
               t: gnt := count
            end
         end;
         dispose(start);
         equalmono := false
      end
end; (* getmonocomposition *)
(* end module comp.getmonocomposition version = 5.35; (@ of rseq.p 1995 Feb 21 *)

(* begin module calehnb *)
procedure calehnb(n: integer;
                  gna,gnc,gng,gnt: integer;
                  var hg, ehnb, varhnb: real
              (* ; debugging: boolean *));
(* calculate e(hnb) in bits/bp (ehnb) for a number (n) of example sequence
sites.  gna to gnt are the composition to use for the genome probabilities
of a to t.  the genomic entropy hg and the variance var(hnb)
(=varhnb) are also calculated.

if the variable debugging is passed to the procedure then the individual
combinations of hnb are displayed.

      note: this procedure should not be broken into smaller
procedures so that it remains efficient.
      version = 3.02; of procedure calehnb 1983 nov 23 *)
const
      maxsize = 200; (* the largest n allowed *)
      accuracy = 10000; (* less than (1/accuracy) bits error is demanded
      for the sum of pnb (see variable 'total') at the end of the procedure *)

var
      log2: real; (* natural log of 2, used to find log base 2 *)
      logn: real; (* log of n *)
      nlog2: real; (* n * log2 *)

      gn: integer; (* sum of gna..gnt *)
      logpa,logpc,logpg,logpt: real; (* logs of genome probabilities *)

      (* log of n factorial is the sum of i=1 to n of log(i).
      the array below represents these logs up to n *)
      logfact: array[0..maxsize] of real;

      (* precalculated values of -p*log2(p), where p=nb/n for
         nb = 0 .. n.  m stands for minus *)
      mplog2p: array[0..maxsize] of real;

      i: integer; (* index for logfact and mplog2p *)
      logi: real; (* natural log of i *)

      na, nc, ng, nt: integer; (* numbers of bases in a site *)
      done: boolean; (* true when the loop is completed *)

      pnb: real; (* multinomial probability of a combination
                    of na, nc, ng, nt *)
      hnb: real; (* entropy for a combination of na..nt *)
      pnbhnb: real; (* pnb*hnb, an intermediate result *)
      sshnb: real; (* sum of squares of hnb *)

      (* variables for testing program correctness: *)
      total: real; (* sum of pnb over all combinations of na..nt
         if this is not 1.00, the program is in error *)
      counter: integer; (* counts the number of times through
         the loop *)

begin (* calehnb *)
      (* prevent access to outside the arrays: *)
      if n > maxsize then begin
         writeln(output,' procedure calehnb: n > maxsize (',
                        n:1,'>',maxsize:1,')');
         halt
      end;
      counter := 0;
      total := 0.0;

      log2 := ln(2);
      logn := ln(n);
      nlog2 := n * log2;

      (* get logs of genome probabilities *)
      gn := gna + gnc + gng + gnt;
      logpa := ln(gna/gn);
      logpc := ln(gnc/gn);
      logpg := ln(gng/gn);
      logpt := ln(gnt/gn);

      (* find genomic entropy *)
      hg := -(gna*logpa + gnc*logpc + gng*logpg + gnt*logpt)/(gn*log2);

      ehnb := 0.0; (* start error entropy at zero *)
      sshnb := 0.0;

      (* make table of log of n factorial up to n
         and entropies for nb/n *)
      logfact[0] := 0.0; (* factorial(0) = 0 *)
      mplog2p[0] := 0.0;
      for i := 1 to n do begin
         logi := ln(i);
         logfact[i] := logfact[i - 1] + logi;
         mplog2p[i] := - i * (logi - logn) / nlog2
      end;

      (* begin by looking at the combination with all a: na = n *)
      na := n;
      nc := 0;
      ng := 0;
      nt := 0;

      (* the following loop simulates a number of nested loops
      of the form:
         for b1=a to t do
            for b2=b1 to t do
               for b3=b2 to t do
                  ...
                     for bn=b(n-1) to t do ...
      the resulting set of variables increase in alphabetic order
      since no inner loop variable can have a value less than any
      outer loop.  the number of times through the inner-most loop
      is given by:
         o = (n + 1)*(n + 2)*(n + 3)/6
      in the case where there are four symbols (a,c,g,t) and n is
      the number of nested loops.
         a recursive set of loops would be possible, but it
      would use up too much memory in practical cases (up to n=150
      or higher).  a second algorithm sequests the loop variables
      into an array and increments them there.  however, the goal
      is to get all possible combinations for na, nc, ng, nt, where
      the sum of these is n.  the nested loops provide all the
      combinations in alphabetic order, assuring that there can not
      be any duplicates.  to find nb (one of na..nt) one would look
      at which of the variables b1 to bn were of value b.  this is
      a wasteful operation.
         the loop below simulates the array of control variables
      by changing each nb directly.
*)
      done := false;

      repeat
         (* pnb is calculated by taking the log of the expression

                    fact(n)          na     nc     ng     nt
         pnb = ------------------- pa   * pc   * pg   * pt  .
               fact(na).. fact(nt)

         log(pnb) generates a series of sums, allowing
         the calculation to proceed by addition and
         multiplication rather than multiplication and
         exponentiation.  the factorials become tractable
         in this way *)

         pnb := exp(logfact[n] (* n factorial *)
               - logfact[na]
               - logfact[nc]
               - logfact[ng]
               - logfact[nt]
               + na * logpa
               + nc * logpc
               + ng * logpg
               + nt * logpt);

         hnb :=  mplog2p[na]
               + mplog2p[nc]
               + mplog2p[ng]
               + mplog2p[nt];

         pnbhnb := pnb * hnb;

         ehnb := ehnb + pnbhnb;

         sshnb := sshnb + pnbhnb * hnb; (* sum of squares of hnb *)

         (* the following section keeps track of the calculation
         and writes out the current set of nb. *)
         counter := succ(counter);
(*         if debugging then begin
            write(output,' ',counter:2,' ');
            for i := 1 to na do write(output,'a');
            for i := 1 to nc do write(output,'c');
            for i := 1 to ng do write(output,'g');
            for i := 1 to nt do write(output,'t');
            write(output,' ',na:3,nc:3,ng:3,nt:3);
            writeln(output,' pnb = ',pnb:10:5);
         end;  *)
         total := total + pnb;

         (* the remaining portion of this repeat loop generates
         the values of na, nc, ng and nt.  notice that
         there are 7 possibilities at each loop increment.
         other than the stop, in each case the sum of
         na+nc+ng+nt remains constant (=n). *)
         if nt > 0
         then begin (* ending on a t - do outer loops *)
            if ng > 0
            then begin (* turn g into t *)
               ng := ng - 1;
               nt := nt + 1
            end
            else if nc > 0
            then begin (* turn one c into g,
               and all t to g (note ng = 0 initially) *)
               nc := nc - 1;
               ng := nt + 1;
               nt := 0
            end
            else if na > 0
            then begin (* turn one a into c and
               all g and t to c. (note ng=nc=0 initially) *)
               na := na - 1;
               nc := nt + 1;
               nt := 0
            end
            else done := true (* since nt = n *)
         end
         else begin (* no t - increment innermost loop *)
            if ng > 0
            then begin (* turn g into t *)
               ng := ng - 1;
               nt := nt + 1
            end
            else if nc > 0
            then begin (* turn c into g *)
               nc := nc - 1;
               ng := ng + 1
            end
            else begin (* na > 0; turn a into c *)
               na := na - 1;
               nc := nc + 1
            end
         end
      until done;

      (* final adjustment: we only have the sum of squares so far *)
      varhnb := sshnb - ehnb*ehnb;

      (* if this message appears, there is either a bug in the code or
         the computer cannot be as accurate as requested *)
      if accuracy <> round(accuracy * total) then begin
         writeln(output,' procedure calehnb: the sum of probabilities is');
         writeln(output,' not accurate to one part in ',accuracy:1);
         writeln(output,' the sum of the probabilities is ',total:10:8);
      end;

      (* if this message appear, then there is an error in the
         repeat-until loop: it did not repeat as many times as
         is expected from the algorithm *)
      if counter <> round((n+1)*(n+2)*(n+3)/6) then begin
         writeln(output,' procedure calehnb: program error, the number of');
         writeln(output,' calculations is in error');
         halt
      end;
(*      writeln(output, '    total: ',total:10:5);
      writeln(output,'    count = ',counter:1);
      writeln(output,'    (n+1)*(n+2)*(n+3)/6 = ',
                          round((n+1)*(n+2)*(n+3)/6):1); *)
end; (* calehnb *)
(* end module calehnb version = 5.35; (@ of rseq.p 1995 Feb 21 *)

(* begin module calaehnb *)
procedure calaehnb(n: integer;
                   gna,gnc,gng,gnt: integer;
                   var hg, aehnb, avarhnb: real);
(* calculate ae(hnb) in bits/bp (=aehnb) for a number (n) of example sequence
sites.  gna to gnt are the composition to use for the genome probabilities
of a to t.  the genomic entropy (=hg) and the variance avar(hnb) (=avarhnb) are
also calculated.
      this procedure gives approximations for e(hnb) and var(hnb).
      it is based on a formula derived by jeff haemer.
      see also:
      g. p. basharin
      theory probability appl. 4(3): 333-336 (1959)
      'on a statistical estimate for the entropy of a sequence
      of independent random variables'

      version = 3.01; of procedure calaehnb 1983 nov 23 *)
var
      log2: real; (* natural log of 2 *)

      gn: integer; (* sum of gna..gnt *)
      pa, pc, pg, pt: real; (* genomic probabilities *)
      e: real; (* the approximate sampling error *)

begin (* calaehnb *)
      log2 := ln(2);

      gn := gna + gnc + gng + gnt;
      pa := gna/gn;
      pc := gnc/gn;
      pg := gng/gn;
      pt := gnt/gn;

      hg := -( (pa*ln(pa))
              +(pc*ln(pc))
              +(pg*ln(pg))
              +(pt*ln(pt)))/log2;

      e := 3/(2*log2*n);

      aehnb := hg - e;

      avarhnb := e * e
end; (* calaehnb *)
(* end module calaehnb version = 5.35; (@ of rseq.p 1995 Feb 21 *)

(* begin module winrs.themain *)
procedure themain(var book, cmp, winrsp, xyin: text);
(* the main procedure of the program *)
const
   debugging = false; (* are we debugging the code? *)
   kickover = 50; (* e(hnb) for n values above this are estimated from
                     ae(hnb) = hg - 3/(2 ln(2) n) while those below
                     are calculated exactly *)
var
   apiece: pieceptr; (* a piece of DNA *)
   b: base; (* a base in the sequence *)
   bases: array[a..t] of integer; (* numbers of bases at a position *)
   Ehnb: real; (* small sample correction: 2 - e(n) *)
   equalmono: boolean; (* true if equal mononucleotides
      were assumed by getmonocomposition *)
   f: real; (* frequency of a base *)
   flogf: array[0..LongestWindowAllowed] of real;
      (* - f log2 f for f=1/n precomputed *)
   gna,gnc,gng,gnt: integer; (* genomic composition *)
   hg: real; (* genomic uncertainty *)
   length: integer; (* the length of apiece *)
   ln2: real; (* precompute ln(2) *)
   MinimumWindowSize: integer; (* minimum window size *)
   MaximumWindowSize: integer; (* maximum window size *)
   n: integer; (* index to flogf *)
   NumberOfWindows: integer; (* number of windows to use for computation.
      This is the same as the number of aligned sequences. *)
   PositionInWindow: integer; (* position in all window *)
   PositionOnSequence: integer; (* position in one specific window *)
   Rs: real; (* Rsequence *)
   SD: real; (* Standard Deviation *)
   SpaceBetweenWindows: integer; (* increment to next window zero *)
   StartCoordinate: integer; (* start coordinate for windows *)
   varhnb: real; (* var(hnb) *)
   Window: integer; (* index to the windows of a given size *)
   WindowSize: integer; (* index to the size of windows *)

procedure die;
(* to prevent accidental use of an old xyin file,
destroy it before halting *)
begin
   rewrite(xyin);
   halt
end;

procedure ReadParametersIn;
(* read parameters in *)
begin
   reset(winrsp);
   readln(winrsp,StartCoordinate);
   readln(winrsp,MinimumWindowSize);
   readln(winrsp,MaximumWindowSize);
   readln(winrsp,SpaceBetweenWindows);
   readln(winrsp,NumberOfWindows);

   (* check parameters *)

   if StartCoordinate < 1
   then begin
      writeln(output,'StartCoordinate cannot be less than 1');
      die
   end;
   (* The check for exceeding end of sequence is later after sequence has
      been read in *)

   if MinimumWindowSize < 1
   then begin
      writeln(output,'MinimumWindowSize cannot be less than 1');
      die
   end;

   if MaximumWindowSize < 1
   then begin
      writeln(output,'MaximumWindowSize cannot be less than 1');
      die
   end;

   if MaximumWindowSize > LongestWindowAllowed
   then begin
      writeln(output,'MaximumWindowSize cannot exceed',
                     ' LongestWindowAllowed = ',LongestWindowAllowed:1);
      die
   end;

   if MinimumWindowSize > MaximumWindowSize
   then begin
      writeln(output,'No computation:',
             ' MinimumWindowSize exceeds MaximumWindowSize');
      die
   end;

   if NumberOfWindows < 1
   then begin
      writeln(output,'No computation:',
             ' NumberOfWindows must be > 0');
      die
   end;

end;

procedure WriteParametersOut;
(* write parameters out *)
begin
   rewrite(xyin);
   writeln(xyin,'* winscan ',version:4:2);
   writeln(xyin,'*');
   writeln(xyin,'* paremeters:');
   writeln(xyin,'* ',StartCoordinate:10,' Start Coordinate');
   writeln(xyin,'* ',MinimumWindowSize:10,' Minimum Window Size');
   writeln(xyin,'* ',MaximumWindowSize:10,' Maximum Window Size');
   writeln(xyin,'* ',SpaceBetweenWindows:10,' Space Between Windows');
   writeln(xyin,'* ',NumberOfWindows:10,' Number Of Windows');

   writeln(xyin,'*');
   if SpaceBetweenWindows < 0
   then begin
   writeln(xyin,'*');
      writeln(output,'WINDOWS OVERLAP');
      writeln(xyin,'* WINDOWS OVERLAP');
   end
   else begin
      writeln(output,'windows do not overlap');
      writeln(xyin,'* windows do not overlap');
   end;

   writeln(xyin,'*');
   writeln(xyin,'* book:');
   reset(book);
   if eof(book) then begin
      writeln(output,'book is empty');
      die;
   end;
   copyaline(book,xyin);
end;

procedure ReadInThePiece;
(* read in the piece *)
begin
   brinit(book);
   new(apiece);
   getpiece(book,apiece);
   length:=piecelength(apiece);
   writeln(output,'Sequence length is ',length:1,' bases.');
   if length > dnamax then begin
      writeln(output,'The first piece in the book has a length of ',
                     length:1,' bases.');
      writeln(output,'This is larger than can be stored, dnamax is ',
                     dnamax:1,' bases.');
      die;
   end;
   writeln(xyin,'* ',apiece^.key.hea.keynam.letters,
                  ' ',length:6,' bases');
end;

procedure DoMoreChecksNowThatWeHaveTheLength;
(* do more checks now that we have the length *)
begin
   if StartCoordinate > length
   then begin
      writeln(output,'StartCoordinate cannot exceed sequence length');
      die
   end;

   if StartCoordinate < 0 then begin
      writeln(output,'StartCoordinate must be positive.');
      die;
   end;

   if (StartCoordinate - 1)
      + (MaximumWindowSize + SpaceBetweenWindows) * (NumberOfWindows)
      > length then begin
      writeln(output,'Windows cannot extend outside to the right.');
      writeln(output, '(StartCoordinate[=',StartCoordinate:1,'] - 1)');
      writeln(output,
     '+ (MaximumWindowSize[=',MaximumWindowSize:1,']',
     ' + SpaceBetweenWindows[=',SpaceBetweenWindows:1,'])',
     ' * (NumberOfWindows[=',NumberOfWindows:1,'])');
      writeln(output,' must not exceed the sequence length[=',length:1,'].');
      die;
   end;
end;

procedure BuildPrecomputedTable;
(* build precomputed table *)
var
   n: integer; (* index to number of windows *)

begin
   (* get composition *)
   getmonocomposition(cmp,gna,gnc,gng,gnt,
         output, equalmono);

   (* compute sample size correction for NumberOfWindows *)
   if NumberOfWindows <= kickover
   then begin (* not using approximation *)
      calehnb (NumberOfWindows, gna,gnc,gng,gnt, hg, Ehnb, varhnb);
   end
   else begin (* using approximation *)
      calaehnb(NumberOfWindows, gna,gnc,gng,gnt, hg, Ehnb, varhnb);
   end;

   writeln(xyin,'* Ehnb = ',Ehnb:infofield:infodecim,' bits');
   writeln(xyin,'* Hg   = ',hg:infofield:infodecim,' bits');
   writeln(xyin,'* 2-Hg = ',
                   (2-hg):infofield:infodecim,' bits per base');

   ln2 := ln(2);
   flogf[0] := 0.0; (* can't take log of zero so do this one separately *)
   for n := 1 to NumberOfWindows - 1 do begin
      f := n/NumberOfWindows;
      flogf[n] := - f * ln(f) / ln2;
   end;
   flogf[NumberOfWindows] := 0.0; (* set zero to avoid small error *)

if debugging then
for n := 0 to NumberOfWindows do begin
writeln(xyin,'* flogf[',n:1,']=',flogf[n]:infofield:infodecim);
end;

end;

procedure ReportDataColumnsOfXyin;
(* report data columns of xyin *)
begin
   (* report data columns of xyin *)
   writeln(xyin,'*');
   writeln(xyin,'* data columns:');
   writeln(xyin,'* 1: Window Size');
   writeln(xyin,'* 2: Rs, Rsequence, bits');
   writeln(xyin,'* 3: SD, Standard Deviation of Rs, bits');
   writeln(xyin,'* 4: Rs/(Window Size), bits/base');
   writeln(xyin,'* 5: SD/(Window Size), bits/base');
end;

procedure DoTheCrunchAndMakeTheXyin;
(* Do the crunch and make the xyin *)
var
   b: base; (* index to the bases *)
   WindowSize, PositionInWindow, Window: integer; (* window parameters *)
begin
   for WindowSize := MinimumWindowSize to MaximumWindowSize do begin
if debugging then
writeln(xyin,'* WindowSize: ',WindowSize:3);

      SD := sqrt(WindowSize*varhnb);
      Rs := Ehnb * WindowSize;
      (* scan across all windows and compute Rs for each *)
      for PositionInWindow := StartCoordinate
                           to StartCoordinate + WindowSize - 1 do begin

         for b := a to t do bases[b] := 0;
         PositionOnSequence  := PositionInWindow;
if debugging then
write(xyin,'* position=',PositionInWindow:3,' ');

         for Window := 1 to NumberOfWindows do begin
            if PositionOnSequence > length then begin
               writeln(output,'PROGRAM ERROR:');
               writeln(output,'PositionOnSequence = ',PositionOnSequence:1);
               writeln(output,'This is past the end of the sequence.');
               halt
            end;

            b := apiece^.dna^.part[PositionOnSequence];

if debugging then
write(xyin,basetochar(b));

            bases[b] := succ(bases[b]);
            PositionOnSequence := PositionOnSequence
                                   + WindowSize
                                   + SpaceBetweenWindows;

         end;

if debugging then begin
for b := a to t do write(xyin,' ',basetochar(b),'=',bases[b]:1);
writeln(xyin);
end;

         for b := a to t do Rs := Rs - flogf[bases[b]];
      end;

      writeln(xyin,WindowSize:infodecim,
                  ' ',Rs:infofield:infodecim,
               ' ',SD:infofield:infodecim,
                  ' ',Rs/(WindowSize):infofield:infodecim,
                  ' ',SD/(WindowSize):infofield:infodecim);
   end;
end;

begin
   writeln(output,'winrs ',version:4:2);
   ReadParametersIn;
   WriteParametersOut;
   ReadInThePiece;
   DoMoreChecksNowThatWeHaveTheLength;
   BuildPrecomputedTable;
   ReportDataColumnsOfXyin;
   DoTheCrunchAndMakeTheXyin;
end;
(* end module winrs.themain *)

begin
   themain(book, cmp, winrsp, xyin);
1: end.
