program zipf(zipfp, data, xyin, output);
(* zipf: Monte Carlo simulation for Peter Shenkin's problem

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  https://alum.mit.edu/www/toms 

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.33; (* of zipf.p 2016 January 26
2016 Jan 26, 1.33: remove random(0) - standardize.
1993 Jan 26, 1.32: functional
1992 Jan 13, 1.00: origin
origin 1992 Jan 13 *)
(* end module version *)

(* begin module describe.zipf *)
(*
name
   zipf: Monte Carlo simulation for Peter Shenkin's problem

synopsis
   zipf(zipfp: in, data: out, xyin: out, output: out)

files
   zipfp:  parameters to control the program
      first line: integer, number of correlation coefficients to create
      second line: integer, number of symbols for each correlation coefficient.
         eg, 20 means amino acids.
      third line: character.  't' means use Tom's method, 'p' means use Peter's.
      fourth line: character.  'g' means to graph the simplex.
   data:  a list of correlation coefficients.  This is to be input
      to the genhis program.
   xyin:  data for graphing the simplex.  The graph is generated with the
      xyplo program.
   output: messages to the user

description

   1992 Jan 13  Returned call to Stephen Altschul 496-2475.  He suggested that
   Peter Shenkin's results of rank versus log of probability are due to random
   effects.  This is easy to test with a Monte Carlo simulation:

   Tom's method
      chose s (eg 20) random numbers
      find their sum
      divide each number by the sum to produce s random numbers which
         sum to 1.
      sort the numbers
      take the log versus the rank
      determine the correlation coefficient
      repeat to get distribution of correlation coefficients.

   Peter's method
      chose s-1 random numbers between 0 and 1
      sort the numbers
      take the differences to produce 20 numbers that sum to 1
      resort the numbers
      take the log versus the rank
      determine the correlation coefficient
      repeat to get distribution of correlation coefficients.

   Graph of simplex.  The numbers all add to 1 for either method.  They are
   points in an s dimensional space.  The volume they fit into is a hyper plane
   of s-1 dimensions since they sum to 1, called a simplex.  The distribution
   of the points can be visualized by projecting onto a plane and graphing with
   the xyplo program.  The projection is done by using polar coordinates.
   There is a vector P from the center of the simplex to each point to graph.
   There is a vector, A, from the center of the simplex to the point where the
   first coordinate has value 1 and all others are zero.  The magnitude of P is
   determined, and the angle between P and A determines an angle.  These
   numbers are in polar coordinates.  They are converted to rectangular
   coordinates in the xyin file.  If s = 3, then the simplex is a simple plane
   reaching between the three points A=(1,0,0), B=(0,1,0) and C=(0,0,1).  The
   projection takes this equilateral triangle onto the xy plane.  In higher
   dimensions, the points are collapsed to the xy plane, so high dimensional
   effects are expected.  This means that the center should tend to become
   empty, and the distribution will become spherical.

examples

zipfp file:
***********************************************************
10000 10000      1000 Number of correlation coefficients to print out
3 16            Number of symbols being simulated
p             t= tom's, else peter's
g             g = graph the symplex, otherwise not

zipfp:  parameters to control the zipf program.
***********************************************************

genhisp file for use with genhis
***********************************************************
x n 50
r -1 -0.5
***********************************************************

xyplop file for use with xyplo
***********************************************************
2 2       zerox zeroy         graph coordinate center
x -1 1 zx 0 25    zx min max (character, real, real) if zx='x' then set xaxis
y -1 1 zy 0 250   zy min max (character, real, real) if zy='y' then set yaxis
10 10     xinterval yinterval number of intervals on axes to plot
6 6       xwidth    ywidth    width of numbers in characters
1 1       xdecimal  ydecimal  number of decimal places
16.5 22.0 xsize     ysize     size of axes in cm
x
y
c         zc                  'c' crosshairs, axXyYnN
n 2       zxl base            if zxl='l' then make x axis log to the given base
n 2       zyl base            if zyl='l' then make y axis log to the given base
          *********************************************************************
1 2       xcolumn   ycolumn   columns of xyin that determine plot location
0         symbol column       the xyin column to read symbols from
0  0      xscolumn  yscolumn  columns of xyin that determine the symbol size
0 0 0     hue saturation brightness   columns for color manipulation
          *********************************************************************
p         symbol-to-plot      c(circle)bd(dotted box)x+Ifgpr(rectangle)
0         symbol-flag         character in xyin that indicates that this symbol
0.05      symbol sizex        side in inches on the x axis of the symbol.
0.05      symbol sizey        as for the x axis, get size from yscolumn
nl 0.05   no connection (example for connection is c- 0.05 for dashed 0.05 inch)
n  0.05   linetype  size      linetype l.-in and size of dashes or dots
          *********************************************************************
.
          *********************************************************************
***********************************************************

documentation

see also
   genhis.p, xyplo.p

author
   Thomas Dana Schneider

bugs

   The revision 2016 Jan 26 replaced the non-standard random(0) with a
   standard proceedure.  This will always give the same results.  For
   actual use, add parameters and the timeseed function to base the
   initial seed on the date and time.

technical notes
   This was replaced by a portable one, but with the danger
   of it not giving good results.

*)
(* end module describe.zipf *)

(* begin module zipf.const *)
maxsymbols = 25;
(* end module zipf.const *)

(* begin module vector.const *)
      maxvecpart = 64; (* the number of elements in a 'part' of a vector *)
      vpagewidth = 64; (* the width (in characters) of the output vector
         file from procedure writevector *)
(* end module vector.const version = 'matmod 2.05 88 dec 15 tds/gds'; *)

type
   position = 0..maxsymbols; (* position in the map *)
   (* a bunch of numbers to be sorted: *)
   maptype =  array[1..maxsymbols] of real;

(* begin module vector.type *)
(* vectors are useful for many applications, including the encoding of
   sequences.  this vector type is designed to be flexible enough to be
   used whenever one needs a vector.  it is designed as a linked list so
   it can contain as many elements as are ever needed.  the actual 'vector'
   is a record containing the length and a pointer to the first 'vectorpart'.
   that vectorpart is a record containing an array of the first 'maxvecpart'
   elements  (maxvecpart is a constant, from module vector.const, which must
   be included) and a pointer to the next 'vectorpart'.  the elements are
   of type real so that either integers or reals may be used.  *)
      partptr = ^vectorpart;
      vectorpart = record
         numbers : array[1..maxvecpart] of real;
         next : partptr;
      end;

      vector = record
         length : integer;
         part : partptr;
      end;
(* end module vector.type version = 'matmod 2.05 88 dec 15 tds/gds'; *)

var
   zipfp: text; (* parameter file used by this program *)
   data: text; (* data file produced by this program *)
   xyin: text; (* graphics data file produced by this program *)
   map: maptype; (* the numbers to be sorted *)
   seed: real; (* random number seed *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'matmod 2.05 88 dec 15 tds/gds'; *)

function lessthan(a, b: position): boolean;
(* see quicksort *)
begin
   (* this puts the largest one first rather than the smallest: *)
   lessthan:= (map[a] > map[b])
end;

procedure swap(a,b: position);
(* see quicksort *)
var
      hold: real;
begin
      hold:=map[a];
      map[a]:=map[b];
      map[b]:=hold
(*;write(output,'a=',a:1,', b=',b:1); print (@ for testing *)
end;

(* begin module quicksort *)

procedure quicksort(left, right: position);
(* quick sort a list between positions left and right, into ascending order.
   a position is simply a scalar of the form 0..max.
   the array to be sorted is dimensioned 1..max.
(the difference in the ranges is important to the correct operation
of the sort...)
   two external routines are used:
      function lessthan(a, b: position): boolean is a generalized test for
          value-at-a < value-at-b.
      procedure swap(a, b: position) switches the items at positions a and b.
   since these routines are external, the procedure is general.

   this procedure taken from the book
   'algorithms + data structures = programs' by niklaus wirth,
   prentice-hall, inc., englewood cliffs, n.j.(1976), pp. 76-82 *)
var
      lower, upper: position; (* the positions looked at currently *)
      center: position; (* the rough center of the region being sorted *)
begin
      lower  := left;
      center := (left + right) div 2;
      upper  := right;

      repeat
         while lessthan(lower, center) do lower := succ(lower);
         while lessthan(center, upper) do upper := pred(upper);

         if lower <= upper then begin
            (* keep track of the center through the map: *)
                 if lower = center then center:=upper
            else if upper = center then center:=lower;
            swap(lower, upper);
            lower := succ(lower);
            upper := pred(upper)
         end
      until lower > upper;

      if  left < upper then quicksort(left,  upper);
      if lower < right then quicksort(lower, right)
end;

(* end module quicksort version = 4.11; (@ of prgmod.p 1991 Apr 22 *)

(* begin module linear.regression *)
(* linear regression procedure
origin: 1980 august 5
      by thomas schneider
      copyright 1986 *)

procedure regress ( control: char; (* described below *)
                    x, y: real;    (* the data pairs *)
                var sumx, sumy, sumxsqd, sumysqd, sumxy, (* internal records *)
(* mean, variance and covariance of x and y: *)
                    ex, ey, varx, vary, covxy,
(* correlation coefficient: *)
                    r,
(* m = slope, b = y intercept as in y = mx + b *)
                    m, b
                        : real;
(* number of data pairs entered: *)
                var n: integer);
(* linear regression for variables x and y.
   the control variable has only three acceptable states:
      c clear all variables to zero
        (note that r is set to 2.0, so that a program can check whether
         or not the results have been calculated yet. the recommended test
         is not for equality, since round-off could cause problems.  use
              if r > 1.5 then (no calculations yet)
                         else (calculations done)    )
      e enter the x and y values into the internal sums
      r results are calculated from the sums
*)
begin
      if not( control in ['c','e','r']) then begin
         writeln(output,' linear regression control variable value, "',control,
                        '" is not acceptable.');
         writeln(output,' it must be in ["c","e","r"]');
         halt
      end

      else case control of
         'c': begin (* clear *)
            x:=0.0; y:=0.0;
            sumx:=0.0; sumy:=0.0;
            sumxsqd:=0.0; sumysqd:=0.0;
            sumxy:=0.0;
            ex:=0.0; ey:=0.0;
            varx:=0.0; vary:=0.0;
            covxy:=0.0;
            r:=2.0; (* this (impossible) value can act as a flag that
                       shows that the calculations have not been made. *)

            m:=0.0; b:=0.0;
            n:=0
         end;

         'e': begin (* enter data *)
            sumx:=sumx + x;
            sumy:=sumy + y;
            sumxsqd:=sumxsqd + x*x;
            sumysqd:=sumysqd + y*y;
            sumxy:=sumxy + x*y;
            n:=succ(n)
         end;

         'r': begin (* calculate results *)
            (* check for conditions that would bomb the program *)
            if n = 0 then begin
               writeln(output,'regress: n is 0; no samples recorded');
               halt
            end;
            if n = 1 then begin
               writeln(output,'regress: n is 1; regression impossible');
               halt
            end;

            ex:=sumx/n;
            ey:=sumy/n;

            varx:=(sumxsqd/n) - ex*ex;
            vary:=(sumysqd/n) - ey*ey;
            if varx=0.0 then begin
               writeln(output,'regress: variance of x is zero;',
                              ' regression impossible');
               halt
            end;
            if vary=0.0 then begin
               writeln(output,'regress: variance of y is zero;',
                              ' regression impossible');
               halt
            end;

            covxy:=(sumxy/n) - ex*ey;

            r:=covxy/(sqrt(varx * vary));

            m:=covxy/varx;
            b:=ey - m*ex
         end
      end
end;
(* end module linear.regression version = 'matmod 2.05 88 dec 15 tds/gds'; *)

(* begin module vector.primitives *)
(* these functions and procedures do manipulations on vectors. *)

function vget(v:vector; pos:integer): real;
(* this returns from vector 'v' the value of the element at position 'pos' *)
var i: integer;
begin
      if (pos > v.length) or (pos < 1) then begin
         writeln(output,' error in call to function vget:',
            ' position ',pos:1,' is beyond the end of the vector');
         halt;
      end;

      (* move to the correct 'vectorpart' *)
      for i := 1 to ((pos -1) div maxvecpart) do  v.part := v.part^.next;

      (* get the proper array element from this part *)
      vget := v.part^.numbers[(pos -1) mod maxvecpart + 1];
end;

procedure vput(var v:vector; pos:integer; number: real);
(* this puts into vector 'v' the value of 'number' at position 'pos' *)
var i: integer;
      firstpart: partptr;  (* of the vector *)
begin
      if (pos > v.length) or (pos < 1) then begin
         writeln(output,' error in call to function vput:',
            ' position ',pos:1,' is beyond the end of the vector');
         halt;
      end;

      (* move to the correct 'vectorpart' *)
      firstpart := v.part;
      for i := 1 to ((pos -1) div maxvecpart) do  v.part := v.part^.next;

      (* put the 'number' into the proper array element for this part *)
      v.part^.numbers[(pos -1) mod maxvecpart + 1] := number;
      v.part := firstpart;
end;

procedure makevector(var v: vector; l: integer);
(* create the linked list of vector-parts which are needed for a vector
   of length 'l' *)
var   numparts,  (* number of parts needed for the vector *)
      i: integer; (* index to the parts of the vector *)
      firstpart,  (* of the vector *)
      newpart: partptr;  (* of the vector *)
begin
      if l < 1 then begin
         writeln(output,' makevector: positive length required');
         halt
      end;

      v.length := l;
      new(v.part);
      firstpart := v.part;
      numparts := (v.length - 1) div  maxvecpart + 1;
      for i := 1 to (numparts - 1) do begin
         new(newpart);
         v.part^.next := newpart;
         v.part := newpart;
      end;
      v.part^.next := nil;
      v.part := firstpart;
end;
(* end module vector.primitives version = 'matmod 2.05 88 dec 15 tds/gds'; *)

(* begin module vector.vset *)
procedure vset(thevalue: real; var v: vector);
(* set the value thevalue into v *)
var   i: integer; (* an index *)
begin
      for i := 1 to v.length do vput(v,i,thevalue)
end;
(* end module vector.vset version = 'matmod 2.05 88 dec 15 tds/gds'; *)

(* begin module vector.vsub *)
procedure vsub(a: vector; var b: vector);
(* subtract the vector a from the vector b *)
var   i: integer; (* an index *)
begin
      if a.length <> b.length then begin
         writeln(output,' function vsub: vector lengths must be equal');
         halt
      end;

      for i := 1 to a.length
         do vput(b,i,vget(b,i)-vget(a,i))
end;
(* end module vector.vsub version = 'matmod 2.05 88 dec 15 tds/gds'; *)
 
(* begin module vector.writevector *)
procedure writevector(var thefile: text; v: vector; y,z: integer);
(* writes the elements of 'v' to 'thefile'.  the integers y and z are:
    y: the size of the field for printing the number;
    z: the size of the decimal part of the field,
    if z = 0 then integers will be printed. *)
var   pos,   (* posititon in the vector *)
      i,j: integer;  (* indecies *)
      numparts,  (* number of parts in the vector *)
      lastpart: integer; (* the number of elements in the last vector part *)
      firstpart: partptr;  (* pointer to the first vector part *)
      x: integer; (* the number of elements to write on a line *)
begin
      x := trunc(vpagewidth/(y+1));
      pos := 0;
      firstpart := v.part;
      numparts := (v.length - 1) div maxvecpart + 1;
      lastpart := (v.length - 1) mod maxvecpart + 1;
      if (z = 0) then begin
         for i := 1 to (numparts - 1) do begin
            for j := 1 to maxvecpart do begin
               write(thefile,' ',round(v.part^.numbers[j]):y);
               pos := succ(pos);
               if (pos mod x = 0) then writeln(thefile);
            end;
            v.part := v.part^.next;
         end;
         for j := 1 to lastpart do begin
            write(thefile,' ',round(v.part^.numbers[j]):y);
            pos := succ(pos);
            if ((pos < v.length) and (pos mod x = 0)) then writeln(thefile);
         end
      end
      else begin
         for i := 1 to (numparts - 1) do begin
            for j := 1 to maxvecpart do begin
               write(thefile,' ',v.part^.numbers[j]:y:z);
               pos := succ(pos);
               if (pos mod x = 0) then writeln(thefile);
            end;
            v.part := v.part^.next;
         end;
         for j := 1 to lastpart do begin
            write(thefile,' ',v.part^.numbers[j]:y:z);
            pos := succ(pos);
            if ((pos < v.length) and (pos mod x = 0)) then writeln(thefile);
         end;
      end;
      writeln(thefile);
      v.part := firstpart;
end;
(* end module vector.writevector version = 'matmod 2.05 88 dec 15 tds/gds'; *)

(* begin module vector.dotproduct *)
function dotproduct(vectora,vectorb: vector): real;
(* this returns the dotproduct of 'vectora' and 'vectorb' *)
var i: integer;  (* an index *)
    j: real;  (* partial products *)
begin
      if vectora.length <> vectorb.length then begin
         writeln(output,' function dotproduct: vector lengths must',
            ' be equal');
         halt
      end;
      j := 0;
      for i := 1 to vectora.length do
         j := j + vget(vectora,i) * vget(vectorb,i);
      dotproduct := j;
end;
(* end module vector.dotproduct version = 'matmod 2.05 88 dec 15 tds/gds'; *)

(* begin module vector.magnitude *)
function magnitude(var v: vector): real;
(* find the magnitude (length) of the vector v *)
begin
      magnitude := sqrt(dotproduct(v,v));
end;
(* end module vector.magnitude version = 'matmod 2.05 88 dec 15 tds/gds'; *)

(* begin module arccos *)
function arccos(x: real): real;
(* Calculate the arccosine.  Pascal only supplies arctan(angle),
so we have to calculate it ourselves.  Begin from

      2           2
   sin angle + cos angle = 1

             2
divide by cos angle to obtain a relationship between tan(angle) and
cos(angle).  Rearrange.
   The special case of x=0 must be dealt with, since this form
would give a division by zero.  The value that must be returned is
pi/2, since cos(pi/2)=0.  A nice way to find this is to note that at a
45 degree angle (pi/4) the tangent is 1, so that pi=4arctan(1),
and pi/2 = 2 * arctan(1).
   Angles larger than pi/2 can be had if x is negative. *)
var
   pi: real; (* 3.1415... *)
begin
   if (x < -1.0) or (x > +1.0) then begin
      writeln(output,' function arccos(x):',
         ' x was not -1<=x<=+1');
      halt
   end;
 
   if x=0.0
   then arccos := 2.0 * arctan(1.0)
   else if x > 0.0
        then arccos := arctan(sqrt( (1.0/(x*x)) - 1.0))
        else begin
           pi := 4.0 * arctan(1.0);
           arccos := pi - arctan(sqrt( (1.0/(x*x)) - 1.0))
        end
end;
(* end module arccos version = 'matmod 2.05 88 dec 15 tds/gds'; *)

(* begin module vector.angle *)
function angle(var a,b: vector): real;
(* find the angle between the vectors a and b.
to calculate the angle between the vectors we use
   dotproduct(a,b)=magnitude(a)*magnitude(b)*cos(angle)
which requires arccos(angle), supplied by module arccos. *)
begin
      angle := arccos( dotproduct(a,b)/
                     ( magnitude(a) * magnitude(b)) )
end;
(* end module vector.angle version = 'matmod 2.05 88 dec 15 tds/gds'; *)

(* begin module pic.polrec *)
procedure polrec(r,theta: real; var x,y: real);
(* convert polar to rectangular coordinates,
theta is in radians *)
begin
   x := r*cos(theta);
   y := r*sin(theta)
end;
(* end module pic.polrec version = 2.63; (@ of dops.p 1991 November 2 *)

(* begin module zipf.themain *)
procedure themain(var zipfp, data, xyin: text);
(* the main procedure of the program *)
const
   debug = false; (* if true, print trace values *)
var
    (* variables used for the linear regression *)
x, y: real;    (* the data pairs *)
sumx, sumy, sumxsqd, sumysqd, sumxy: real; (* internal records for regression *)
ex, ey, varx, vary, covxy: real; (* mean, variance and covariance of x and y *)
r: real; (* correlation coefficient: *)
m, b: real; (* m = slope, b = y intercept as in y = mx + b *)
n: integer; (* number of data pairs entered so far *)

(* avector: vector; the vector to the first dimension from the center
   of the simplex.  This is not actually used long, so it is done in
   the acvector first. *)
acvector: vector; (* vector between avector and cvector *)
actualsymbols: integer; (* actual number of symbols used *)
cvector: vector; (* the vector to the center of the simplex *)
delta: real; (* difference between current random value and previous one *)
count: integer; (* count of the number of correlation coefficients
                   done so far *)
graphing: boolean; (* if true, flatten the data into x and y in xyin *)
number: integer; (* number of correlation coefficients to calculate *)
previous: real; (* the previous random value *)
(* pvector: vector; the vector to the current point from the center
   of the simplex *)
pcvector: vector; (* vector between pvector and cvector *)
rank: integer; (* order of number *)
radius: real; (* distance from simplex center to point *)
theta: real; (* angle between first axis and central point and current point
                of simplex *)
tomsway: boolean; (* if true, use Tom's method, if false, use Peter's method *)
sum: real; (* sum of random numbers for normalizing them *)
showlow: boolean; (* show number - 1 if true *)

procedure show;
(* show the current list *)
var
   max: integer; (* how many to show *)
   sum: real; (* sum of numbers *)
   index: integer; (* to the list *)
begin
   sum := 0.0;
   if showlow then max := actualsymbols - 1
              else max := actualsymbols;
   for index := 1 to max do begin
      write(data,map[index]:10:5);
      if index > 1 then write(data,' +');
      writeln(data);
      sum := sum + map[index]
   end;
   writeln(data,'sum is: ',sum:15:13);
   writeln(data,' ');
end;

(* begin module random *)
(* random generator 1.  version = 1.00; 1986 December 31 *)
(* written by Thomas Schneider *)
procedure random(var seed: real);
(* WARNING: This random number generator is known to cycle on a Pyramid
90x computer.  It does not seem to cycle on a CDC Cyber.  The difference
is that the Pyramid has a 32 bit word, while the Cyber has 60 bits.
You should test the routine (use program tstrnd) before you use it.

   The seed value is used to generate a random number in the range 0 to 1.
The seed is then altered, to provide the seed of the next random number.
In other words, the seed is the random number.  To obtain the next random
number in the pseudo-random series, simply call random again with the same
seed.  Note that the series will not be truely random, since the previous
value determines the next value.  This is "pseudo-random".

An initial seed of 4.0 is suggested.  Note that this is not the first
random number.  You can use different seeds to get different pseudo-random
number series.

The expected mean and standard deviation of a flat distribution is
calculated using simple integrals:

       1
mean = / x dx = 0.5
       0
           1           2
variance = / (x - mean)  dx = 1/12;  stdev = sqrt(variance) = .2886751
           0

(The integral for the variance is solved easily by substituting y = x - mean.)

   The algorithm is derived from:
Pascal Programs For Scientists and Engineers, Alan R. Miller
Sybex, Berkeley CA (1981).
which in turn was an adaptation from HP-35 applications programs. *)
const
      pi = 3.141592653589; (* a value used to help mix up the result *)
var
      intermediate: real; (* an intermediate result *)
begin (* random *)
      intermediate := seed + pi;
      intermediate := exp(5.0*ln(intermediate));
      seed := intermediate - trunc(intermediate)
end; (* random *)
(* end module random *)

procedure graph;
(* show the points on the simplex to a graph *)
var
   rank: integer; (* index *)
begin
      if graphing then begin
         (* move the points into the P vector *)
         for rank := 1 to actualsymbols
         do vput(pcvector,rank,map[rank]);

{
      writeln(output);
      write(output,'p vector: ');
      writevector(output,pcvector,6,3);
}

         (* subtract the C vector from the P vector *)
         vsub(cvector,pcvector);

{
      write(output,'pc vector: ');
      writevector(output,pcvector,6,3);
}

         (* find the angle between the PC vector and the AC vector *)
         theta := angle(pcvector,acvector);

{
         (* replace PC by its distance to C *)
         vsub(cvector,pcvector);
}
         radius := magnitude(pcvector);

{
      writeln(output,'radius = ',radius:6:3);
      writeln(output,'theta = ',theta:6:3);
}

{zzz}
         polrec(radius, theta, x, y);
         writeln(xyin,x:15:12,' ',y:15:12);

{
         writeln(output,'degrees = ',theta*57.3:15:12);
         writeln(output,'dot pc c: ',dotproduct(pcvector,cvector):15:12);
         writeln(output,'dot pc ac: ',dotproduct(pcvector,acvector):15:12);
}

      end;
end;

begin
   writeln(output,'zipf ',version:4:2);
   seed := 0.0;

   reset(zipfp);
   readln(zipfp, number);
   if number < 3 then begin
      writeln(output,'at least 3 correlations are required in zipfp');
      halt
   end;
   readln(zipfp, actualsymbols);
   if actualsymbols > maxsymbols then begin
      writeln(output,'use fewer than ',maxsymbols:1,' symbols in zipfp');
      halt
   end;
   if zipfp^='t' then tomsway := true
                 else tomsway := false;
   readln(zipfp);
   if zipfp^='g' then graphing := true
                 else graphing := false;

   rewrite(data);
   writeln(data,'* zipf ',version:4:2);
   writeln(data,'* ',number:1,' number of correlation coefficients made');
   writeln(data,'* ',actualsymbols:1,' number of symbols');
   if tomsway then writeln(data,'* Tom''s way')
              else writeln(data,'* Peter''s way');
   if graphing then writeln(data,'* graphing simplex')
               else writeln(data,'* not graphing simplex');

   if graphing then begin
      rewrite(xyin);
      (* create the C vector, which goes to the center of the simplex *)
      makevector(cvector,actualsymbols);
      vset(1.0/actualsymbols,cvector);
      write(output,'c vector: ');
      writevector(output,cvector,6,3);

      (* To make the A minus C vector, start with the A vector,
         using acvector to hold it temporarily *)
      (* create the A vector *)
      makevector(acvector,actualsymbols);

      (* fill the A vector.  Start out with all zeros *)
      vset(0.0, acvector);
      (* now correct the first one, to get the a vector: *) 
      vput(acvector,1,1.0);

      (* create the A-C vector, by subtracting the c vector away: *)
      vsub(cvector,acvector);
      write(output,'ac vector: ');
      writevector(output,acvector,6,3);

      writeln(output,'dot ac c: ',dotproduct(acvector,cvector):15:12);

      (* create the PC vectors *)
      makevector(pcvector,actualsymbols);
{zzz}
   end;

   count := 0;
   showlow := false;
   repeat
 
      (* clear regression variables: *)
      regress ('c', 0.0, 0.0, sumx, sumy, sumxsqd, sumysqd, sumxy,
               ex, ey, varx, vary, covxy, r, m, b, n);

      (* chose actualsymbols random numbers *)
      for rank := 1 to actualsymbols
      do begin
         random(seed);
         map[rank] := seed;
      end;

      if debug then begin
         if not tomsway then showlow := true;
         writeln(data);
         writeln(data,'*****************');
         writeln(data,'* random numbers:'); show;
      end;

      if tomsway then begin  (* do Tom's way *)
         (* find their sum *)
         sum := 0.0;
         for rank := 1 to actualsymbols do sum := sum + map[rank];

         (* divide each number by the sum to produce actualsymbols
            random numbers which sum to 1. *)
         for rank := 1 to actualsymbols do map[rank] := map[rank]/sum;

         if debug then begin writeln(data,'* normalized:'); show; end;

graph;

         (* sort the numbers *)
         quicksort(1,actualsymbols);

         if debug then begin writeln(data,'* sorted :'); show; end;
      end
      else begin (* Peters way *)

         (* sort the numbers *)
         quicksort(1,actualsymbols-1);

         if debug then begin
            showlow := true;
            writeln(data,'* sorted :');
            show;
         end;

         (* replace numbers with differences: *)
         previous := 1.0;
         for rank := 1 to actualsymbols-1 do begin
            delta := previous - map[rank];
            previous := map[rank];
            map[rank] := delta
         end;
         map[actualsymbols] := previous;

         if debug then begin
            showlow := false;
            writeln(data,'* differences:');
            show;
         end;

graph;

         (* sort the numbers *)
         quicksort(1,actualsymbols);

         if debug then begin
            writeln(data,'* RE-sorted :');
            show;
         end;
      end;

      (* loop here through rank and log of rank *)
      for rank := 1 to actualsymbols do begin
         (* take the log versus the rank *)
         if debug then writeln(data,rank:1,
                       ' ',ln(map[rank]):10:5,' rank and ln(Pi)');

         (* determine the correlation coefficient *)
         (* enter regression variables: *)
         regress ('e', rank, ln(map[rank]),
                  sumx, sumy, sumxsqd, sumysqd, sumxy,
                  ex, ey, varx, vary, covxy, r, m, b, n);
      end;

      (* calculate regression variables: *)
      regress ('r', 0.0, 0.0, sumx, sumy, sumxsqd, sumysqd, sumxy,
                       ex, ey, varx, vary, covxy, r, m, b, n);

      if debug then write(data,'correlation coefficient: ');
      writeln(data,r:15:12);

      (* repeat to get distribution of correlation coefficients. *)
      count := count + 1
   until count >= number;
 
end;
(* end module zipf.themain *)

begin
   themain(zipfp, data, xyin);
1: end.
